*------------------------------------------------------------------------------*
*
*    File    :   MPEGA.h
*
*    Author  :   Stphane TAVENARD
*
*    $VER:   MPEGA.h  1.0  (25/10/1997)
*
*    (C) Copyright 1997-1997 Stphane TAVENARD
*        All Rights Reserved
*
*    #Rev|   Date   |                      Comment
*    ----|----------|--------------------------------------------------------
*    0   |25/10/1997| Initial revision                                     ST
*
*    ------------------------------------------------------------------------
*
*    MPEGA decoder library definitions
*
*------------------------------------------------------------------------------*

	ifnd	LIBRARIES_MPEGA_I
LIBRARIES_MPEGA_I	set	1

	ifnd	EXEC_TYPES_I
	include	exec/types.i
	endc

	ifnd	UTILITY_HOOKS_I
	include utility/hooks.i
	endc

* Controls for decoding

* Qualities
MPEGA_QUALITY_LOW    EQU	0
MPEGA_QUALITY_MEDIUM EQU	1
MPEGA_QUALITY_HIGH   EQU	2

*
*   Bitstream Hook function is called like (SAS/C syntax):
*
*
*   ULONG __saveds __asm HookFunc( register __a0 struct Hook  *hook,
*                                  register __a2 APTR          handle,
*                                  register __a1 MPEGA_ACCESS *access );
*
*   MPEGA_ACCESS struct specify bitstream access function & parameters
*
*   access->func == MPEGA_BSFUNC_OPEN
*      open the bitstream
*      access->data.open.buffer_size is the i/o block size your read function can use
*      access->data.open.stream_size is the total size of the current stream
*                                    (in bytes, set it to 0 if unknown)
*      return your file handle (or NULL if failed)
*   access->func == MPEGA_BSFUNC_CLOSE
*      close the bitstream
*      return 0 if ok
*   access->func == MPEGA_BSFUNC_READ
*      read bytes from bitstream.
*      access->data.read.buffer is the destination buffer.
*      access->data.read.num_bytes is the number of bytes requested for read.
*      return # of bytes read or 0 if EOF.
*   access->func == MPEGA_BSFUNC_SEEK
*      seek into the bitstream
*      access->data.seek.abs_byte_seek_pos is the absolute byte position to reach.
*      return 0 if ok
*

MPEGA_BSFUNC_OPEN  EQU	0
MPEGA_BSFUNC_CLOSE EQU	1
MPEGA_BSFUNC_READ  EQU	2
MPEGA_BSFUNC_SEEK  EQU	3

	structure	MPEGA_ACCESS,0
	LONG		MPA_func		* MPEGA_BSFUNC_xxx *
	LABEL		MPA_sizeof

	structure	MPEGA_ACCESSOPEN,MPA_sizeof
	APTR		MPAO_stream_name 	* in
	LONG		MPAO_buffer_size	* in
	LONG		MPAO_stream_size	* out
	LABEL		MPAO_sizeof

	structure	MPEGA_ACCESSREAD,MPA_sizeof
	APTR		MPAR_buffer		* in/out
	LONG		MPAR_num_bytes		* in
	LABEL		MPAR_sizeof

	structure	MPEGA_ACCESSSEEK,MPA_sizeof
	APTR		MPAS_abs_byte_seek_pos	* out
	LABEL		MPAS_sizeof
	
* Decoding output settings *
	structure	MPEGA_OUTPUT,0
	WORD		MPO_freq_div	;1, 2 or 4
	WORD 		MPO_quality	;0 (low) .. 2 (high)
	LONG		MPO_freq_max	;for automatic freq_div (if mono_freq_div == 0)
	LABEL		MPO_sizeof

* Decoding layer settings *
	structure	MPEGA_LAYER,0
	WORD		MPL_force_mono  	;1 to decode stereo stream in mono, 0 otherwise
	STRUCT		MPL_mono,MPO_sizeof	;mono settings
	STRUCT		MPL_stereo,MPO_sizeof	;stereo settings
	LABEL		MPL_sizeof

* Full control structure of MPEG Audio decoding *

	structure	MPEGA_CTRL,0
	APTR		MPC_bs_access		 ;NULL for default access (file I/O) or give your own bitstream access
	STRUCT		MPC_layer_1_2,MPL_sizeof ;Layer I & II settings
	STRUCT		MPC_layer_3,MPL_sizeof   ;Layer III settings
	WORD		MPC_check_mpeg           ;1 to check for mpeg audio validity at start of stream, 0 otherwise
	LONG		MPC_stream_buffer_size   ;size of bitstream buffer in bytes (0 -> default size)
                                                 ;NOTE: stream_buffer_size must be multiple of 4 bytes
	LABEL		MPC_sizeof
* MPEG Audio modes *

MPEGA_MODE_STEREO   EQU	0
MPEGA_MODE_J_STEREO EQU	1
MPEGA_MODE_DUAL     EQU	2
MPEGA_MODE_MONO     EQU	3

;	Public data (read only)
;	Stream info
	structure	MPEGA_STREAM,0
	WORD  		MPS_norm		;1 or 2
	WORD		MPS_layer		;1..3
	WORD		MPS_mode		;0..3  (MPEGA_MODE_xxx)
	WORD		MPS_bitrate		;in kbps
	LONG		MPS_frequency		;in Hz
	WORD		MPS_channels		;1 or 2
	ULONG		MPS_ms_duration		;stream duration in ms
	WORD		MPS_private_bit		;0 or 1
	WORD		MPS_copyright		;0 or 1
	WORD		MPS_original		;0 or 1

   ;Decoding info according to MPEG control
	WORD		MPS_dec_channels	;decoded channels 1 or 2
	WORD		MPS_dec_quality		;decoding quality 0..2
	LONG		MPS_dec_frequency	;decoding frequency in Hz
   ;Private data
	APTR	MPS_handle	;

MPEGA_MAX_CHANNELS 	EQU	2    ;Max channels
MPEGA_PCM_SIZE		EQU	1152 ;Max samples per frame


* Error codes *

MPEGA_ERR_NONE     EQU	0
MPEGA_ERR_BASE     EQU	0
MPEGA_ERR_EOF      EQU  MPEGA_ERR_BASE-1
MPEGA_ERR_BADFRAME EQU	MPEGA_ERR_BASE-2
MPEGA_ERR_MEM      EQU	MPEGA_ERR_BASE-3
MPEGA_ERR_NO_SYNC  EQU	MPEGA_ERR_BASE-4

	endc
