*****************************************************************************
** Engine: 14 Bit Amplifier II                                             **
**                                                                         **
** Project: Eagleplayer 2.05                                               **
** Authors: Jan Blumenthal & Henryk Richter                                **
** Start  : 1993/01/09                                                     **
** $Rev: 23 $ $Author: buggs $                                             **
** $Date: 2016-10-14 14:12:00 +0200 (Fr, 14. Okt 2016) $                   **
**                                                                         **
*****************************************************************************
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program (See the included file COPYING);
** if not, write to the Free Software Foundation, Inc.,
** 675 Mass Ave, Cambridge, MA 02139, USA.
**
**************************************************************************** 
; Status:
;  - abstracted Hardware output initialization/shutdown
;  - still references globals in 14 Bit Amplifier
;  - no actual check whether this runs on Amiga hardware


;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
;-------------------------------------- Paula 14 Bit Code --------------------------------------------
;-----------------------------------------------------------------------------------------------------
AmpOut_Paula14:
		dc.l	AmpOut_Paula14_Init
		dc.l	AmpOut_Paula14_End
		dc.l	AmpOut_Paula14_StartInt
		dc.l	AmpOut_Paula14_StopInt
		dc.l	AmpOut_Paula14_Check
		dc.l	AmpOut_Paula14_GetMixRates
		dc.l	AmpOut_Paula14_CalcnewmixingPeriod	;OHW_SetMixRate (optional)
		dc.l	AmpOut_Paula14_SetPlayMode		;OHW_SetPlayMode (optional)
		dc.l	AmpOut_Paula14_SetBoost			;OHW_SetBoost (optional)
		dc.l	0,0,0					;OHW_Reserved
	ifne	(*-AmpOut_Paula14)-OHW_Sizeof
	error	"Structure Size Mismatch!"
	endc

;--------------------- return whether this output hardware is available --------------------------------
;input:  A5 - datas
;output: D0 - 1=yes, 0=no
;regs:   all saved (except D0)
AmpOut_Paula14_Check:
;	push	d1-a6
	moveq	#1,d0
;	pop	d1-a6
	rts

;--------------------- return available mixing rates as long int array --------------------------------
;input:  A5 - datas (Amplifier)
;output: D0 - pointer to longword array of mixing frequencies in Hz
;regs:   all saved
AmpOut_Paula14_GetMixRates:
	push	a0

	lea	Paula_Mixfreqs_NTSC,a0
	move.l	#3546895,d0
	sub.l	AudioRate(pc),d0	;3546895 for PAL, 
	bne.s	.ntsc
	lea	Paula_Mixfreqs_PAL,a0
.ntsc
	move.l	a0,d0
	pop	a0
	rts

;--------------------- set surround mode mono-stereo-surround --------------------------------
;input:  A5 - datas (Amplifier)
;        D0 - mode (<0=mono, 0=stereo, >=0 surround), range is <64...64 
;output: D0 - ok = 0 / error <>0
;regs:   all saved
;note:   optional, set pointer to NULL if you won't implement this
AmpOut_Paula14_SetPlayMode:
	push	d1-a6

	move.l	d0,d1

	;* TODO: 0a,1a,2a,3a, copybuf14 *
	lea	Buffer0a,a0		; left
	lea	Buffer1a,a1
	lea	Buffer2a,a2
	lea	Buffer3a,a3

	lea	copybufStereo14,a4
	lea	copybufemptyStereo14,a6

	tst.b	d1			; Stereo ? -> use stereo routine (faster than the other)
	beq.s	.setit

;	lea	copybufSurround14_030,a4 ; table based surround
;	move.l	SurroundTabAddrMid(pc),d2
;	bne.w	.tablebasedsurround	; use table based surround when table is present

	lea	copybufSurround14,a4	; surround routine used in empty and non-empty cases

.tablebasedsurround:
	move.l	a4,a6			; 

	tst.b	d1
	bpl.s	.setit

	lea	copybufMono14,a4	; surround routine used in empty and non-empty cases
	move.l	a4,a6			; 
	neg.b	d1
.setit:
	move.l	a3,SpecialBuffer1	;left special: 4th channel (also left)
	move.l	a2,SpecialBuffer2	;right special: 3rd channel (also right)

	lea	buffer0,a5
	move.l	a0,(a5)+	;buffer0
	move.l	a1,(a5)+	;buffer1
	move.l	a2,(a5)+	;buffer2
	move.l	a3,(a5)+	;buffer3
	move.l	a4,(a5)+	;cbufad
	move.l	a6,(a5)+	;cbufempty

	and.w	#127,d1		;8 Bit -> 16 Bit
	cmp.w	#64,d1
	ble.s	.efvolok
	moveq	#64,d1
.efvolok
	move	d1,effectvol	

	; TODO: UPS buffer fill in 14 Bit mode
	lea	myupsstruct,a6
	lea	UPSBuffer0,a5

	lea	UPSBuffer0a,a0
	lea	UPSBuffer1a,a1

	move.l	a0,(a5)+
	move.l	a0,ups_voice1adr(A6)
	move.l	a0,ups_voice4adr(A6)

	move.l	a1,(a5)+
	move.l	a1,ups_voice2adr(A6)
	move.l	a1,ups_voice3adr(A6)

	move.l	a1,(a5)+		; irrelevant
	move.l	a0,(a5)+		; irrelevant

	move.w	#64,ups_voice1vol(a6)
	move.w	#64,ups_voice2vol(a6)
	move.w	#1,ups_voice3vol(a6)
	move.w	#1,ups_voice4vol(a6)

	clr.l	tmpvol
	clr.w	tmpvol2

	bsr	Makevoltable	; update volume table
	bsr	MakeSurroundTab ; update surround table (68030 and below)

	pop	d1-a6
	moveq	#0,d0
	rts


AmpOut_Paula14_SetBoost:
	push	all
	bsr	AmpOut_Paula14_do14tab
	pop	all
	rts

;---------------- Routines for the 14 Bit Player --------------------
; Init/End
;          - Initplayer -> A0: tag list
;                          A5: EP globals (not Amplifier globals!)
;                       -> return  D0: 0=OK, <>0=Error
;
; Start-Stop: Startint (no parameters, except A5=EP globals)
;
;             Stopint  (no parameters, except A5=EP globals)
;
AmpOut_Paula14_Init:
	move.l	#EPAMT_Flags,d0
	jsr	ENPP_FindTag(A5)
	beq	.err

	moveq	#1,d2
	move.l	d0,d1
	move.l	d0,d6
	and.l	#EPAMB_WaitForStruct,d6

	and.l	#EPAMB_8Bit,d0
	bne	.ok

	moveq	#2,d2
	btst	#EPAMF_8BitUnsigned,d1
	bne	.ok
	
	moveq	#3,d2
	btst	#EPAMF_16Bit,d1			;erstmal nur 16 Bit signed
	beq	.err
.ok
	move.b	d2,fformat

	move.l	#EPAMT_Audiostructs,d0
	jsr	ENPP_FindTag(A5)
	beq	.err
	move.l	d0,d7

	move.l	#EPAMT_NumStructs,d0
	jsr	ENPP_FindTag(A5)
	beq	.err
	move.l	d7,a0

	move	d0,numstructs

	subq.w	#1,d0			;Anzahl der Audio-Strukturen
	blt	.err

	moveq	#0,d1
	moveq	#0,d2
	lea	(a0),a2
.countvoc
	tst.w	AS_LeftRight(a2)
	bpl.s	.addL
	bmi.s	.addR
	bra.s	.noADD
.addL
	cmp.w	#2,AS_Leftright(a2)
	beq	.err
	addq.w	#1,d1
	bra.s	.noADD
.addR
	addq.w	#1,d2
.noadd
	lea	as_sizeof(a2),a2
	dbf	d0,.countvoc

	move.w	d1,numleft		;remember left/right voices -> may still lead to error (!)
	move.w	d2,numright

	move.l	#mix,mixad1		;sane default: 8 bit mixing
	move.l	#mix2,mixad2		;sane default: 8 bit mixing -> will be overwritten by setMixRoutine call
	st	onplay

	move.w	d2,d0
	add.w	d1,d0
	beq	.err			;no active voices ? none ? any ? Why, but why ?

	cmp	#32,d0			;more than 32 active voices ?
	bhi	.err			;bail out for now (though it would be easy to support more than that)

        cmp.b   #3,fformat
        bne.s   .normal8bit

        cmp     #1,d2			;currently, only two channels supported in 16 Bit mode -> subject to change
        bne     .err
        cmp     #1,d1
        bne     .err

.normal8bit
	subq.w	#2,d0
	bgt.s	.hi1
	moveq	#0,d0
.hi1
	mulu	#P_Sizeof,d0
	lea	Prefslist+datas,a1
	lea	(a1,d0.w),a1
	move.l	a1,usedmode
	
	move.w	P_Flags(A1),d0		;Flags.w (Bit0=disable)
	and.b	#PBB_Disable,d0
	bne	.err			;Modus disabled !!

	move.l	a0,epg_audiostruct(A5)	;Audiostruct setzen
	tst.l	d6			;Flag fuer "WaitforStruct" gesetzt ?
	beq.s	.normal
	lea	Extra_AS,a0		;ja, also eigene AS Struktur verwenden
.normal
	move.l	a0,myAudioStruct	;AudioStruct vom Replayer sichern

	bsr	setMixRoutine		;requires "usedmode"

	bsr	InitTables
	bne.s	.err2
	bsr	AmpOut_Paula14_do14tab	;initialize 14 bit table (128k memory provided by Inittables call)

	move.l	usedmode,a1
	move.b	P_Surround(a1),d0	;Playmode
	ext.w	d0
	ext.l	d0
	bsr	AmpOut_Paula14_SetPlayMode

	move.b	P_SurroundDelay(a1),d0
	bsr	setPrefetch

	move.w	numleft,d0
	add.w	numright,d0
	bsr	SelectEntry

	bsr	Copy_AS2

	moveq	#0,d0
	rts
.err
	moveq	#-1,d0
.err2
	sf	onplay
	rts

;----------------------------- "End" -----------------------------------
;
; Input:
;  A5 - EP Globals
; Output:
;  D0 - 0=ok, else error
AmpOut_Paula14_End:
	clr.b	onplay
	move	#-1,Currentnumvoc+datas		;kein Modul am Spielen

	bsr	AmpOut_Paula14_CH_DMAWait
	move.w	#15,$dff096

	clr.w	$dff0a8				;volume = 0
	clr.w	$dff0b8
	clr.w	$dff0c8
	clr.w	$dff0d8

	bsr	myFreemem			;clear allocated memory

	moveq	#0,d0
	rts

;----------------------------- "Play" -----------------------------------
;
; Input:
;  A5 - EP Globals
; Output:
;  D0 - 0=ok, else error
AmpOut_Paula14_StartInt:
	Push	d1-a6

	lea	AmpOut_Paula14_OLDINT0(pc),a0
	move.l	#-1,(a0)

;-------------------- hier alloc Audio -------------------------
	jsr	ENPP_AllocAmigaAudio(a5)
	tst.l	d0
	bne	fail$
;---------------------------------------------------------------
	bset	#1,$bfe001
	move.l	MyEPBase(pc),a0
	tst.b	DTG_LED(a0)
	beq.s	.off
	bclr	#1,$bfe001
.off
					;Audio Interrupt disable
	move.w	#$780,$dff09c		;Interrupt Request Hardwarehack

	movem.l	d0-a6,-(sp)		;Audio Interrupt handler durch
					;eigene ersetzen
	move.l	4.w,a6
	lea	AmpOut_Paula14_Aud0Interrupt(pc),a1
	moveq	#7,d0				;use first audio interrupt
	jsr	-162(a6)			;Setintvektor
	move.l	d0,AmpOut_Paula14_OLDINT0

	movem.l	(sp)+,d0-a6

	move.w	#$8080,$dff09a			;Interrupt enable (first audio)

	bsr	AmpOut_Paula14_StartDMA

	lea	myupsstruct,a1
	move.w	#UPSB_Adr!UPSB_LEN!UPSB_Per!UPSB_Vol!UPSB_DMACON,ups_flags(a1)
	move.w	#15,UPS_DmaCon(a1)
	clr.w	UPS_Enabled(a1)

	move.w	#64,ups_voice1vol(a1)
	move.w	#64,ups_voice2vol(a1)
	move.w	#1,ups_voice3vol(a1)
	move.w	#1,ups_voice4vol(a1)
	clr.l	tmpvol
	clr.w	tmpvol2

	move.l	newmixingperiod(pc),d0
	move.w	d0,ups_voice1per(a1)
	move.w	d0,ups_voice2per(a1)
	move.w	d0,ups_voice3per(a1)
	move.w	d0,ups_voice4per(a1)

	move.w	#SamLEN*NumChipBuffers,ups_voice1len(A1)
	move.w	#SamLEN*NumChipBuffers,ups_voice2len(A1)
	move.w	#SamLEN*NumChipBuffers,ups_voice3len(A1)
	move.w	#SamLEN*NumChipBuffers,ups_voice4len(A1)

	lea	myupsstruct,a6
	movem.l	UPSBuffer0(pc),d3-d4
	move.l	d3,ups_voice1adr(A6)
	move.l	d3,ups_voice4adr(A6)
	move.l	d4,ups_voice2adr(A6)
	move.l	d4,ups_voice3adr(A6)

	moveq	#0,d0
fail$
	Pop	d1-a6
	rts

;
; "Pause"
; Input:
;  A5 - EP Globals
; Output:
;  D0 - 0=ok, else error
AmpOut_Paula14_StopInt:
	lea	$dff0a8,a0
	clr.w	(a0)
	clr.w	$10(a0)
	clr.w	$20(a0)
	clr.w	$30(a0)

	jsr	ENPP_FreeAmigaAudio(a5)

	move.l	AmpOut_Paula14_OLDINT0(pc),d0
	cmp.l	#-1,d0
	beq.s	.empty

	move.w	#$780,$dff09a
	move.w	#$780,$dff09c

	move.w	#$f,$dff096		;stop DMA
	moveq	#16,d0			;sufficient for 31 kHz (normally, you wait about 5 lines)
	bsr	AmpOut_Paula14_CH_DMAWait
	move.w	#$f,$dff096		;yes, DMA stop (just to make sure...)
	
	move.l	4.w,a6
	move.l	AmpOut_Paula14_OLDINT0(pc),a1
	moveq	#7,d0
	jsr	-162(a6)			;Setintvektor

	move.w	#$780,$dff09a

	moveq	#-1,d0
	move.l	d0,AmpOut_Paula14_OLDINT0

	moveq	#0,d0
.empty
	rts

; User selected a new mixing frequency, apply it
;  
; Input:
;  A5 - Amplifier globals
;  D0 - Mixing Rate
; Output:
;  D0 - 0=ok, else error
AmpOut_Paula14_CalcnewmixingPeriod:
	push	d1-a6

	move.l	d0,d7
	;------- sanity check ---------
	cmp.l	#4000,d7
	blt.s	.def
	cmp.l	#57000,d7
	blt.s	.use
.def
	move.l	#28000,d7
	move.l	d7,playmixrate
.use
	;-------------------------------

	move.l	#14187580/4,clock		;Clock constant (reference clock constant is always the PAL constant)
	
	move.l	AudioRate(pc),d2		;3546895 for PAL

	move.l	d2,d0
	move.l	d7,d1				;mixingrate(pc),d1
	divu	d1,d0
	move.l	d2,d1
	divu	d0,d1
	swap	d1
	clr	d1
	swap	d1
	move.l	d1,mrate

	move.l	d2,d0
	lsl.l	#2,d0
	divu	d1,d0
	move.w	d0,mixingperiod4

	move.l	d2,d0
	divu	d1,d0
	swap	d0
	clr	d0
	swap	d0
	move.l	d0,newmixingperiod

	push	d0/d1/d4
	move.l	d0,d4			;playback period
	move.l	#264557,d1		;actual NTSC mixing period -> PAL reference period in AS Structure
	move.l	#3546895,d0
	sub.l	AudioRate(pc),d0
	bne.s	.ntsc
	move.l	#262144,d1		;no change in scaling for PAL
.ntsc
	mulu.l	d1,d4			;mixingperiod * 4 * 65536 -> same scaling as sample period plus room for fractional part

	move.l	d4,referenceperiod	;reference period for mixing step calculations (independent of underlying PAL/NTSC hardware)
	pop	d0/d1/d4

	Push	a1
	lea	myupsstruct,a1
	move.w	d0,ups_voice1per(a1)
	move.w	d0,ups_voice2per(a1)
	move.w	d0,ups_voice3per(a1)
	move.w	d0,ups_voice4per(a1)
	Pop	a1

	bsr.w	ResetASPeriods			;make sure, new interpolation factors are calculated

	pop	d1-a6
	move.l	mrate,d0
	rts


;============ Puffer l?schen und DMA starten =============================
AmpOut_Paula14_StartDMA:
	push	all

	move.w	#$f,$dff096		;DMA Stop

	lea	Buffer0a,a0
	move.w	#[ENDBUFFER-Buffer0a]/2-1,d0
.clr
	clr.w	(A0)+
	dbf	d0,.clr

	lea	StartBSSBuffer,a0
	move.w	#[ENDBSSBUFFER-StartBSSBuffer]/2-1,d0
.clr2
	clr.w	(a0)+
	dbf	d0,.clr2

	bsr	CalcValues
	clr.l	tmpvol
	clr.w	tmpvol2
	clr.l	calcvol
	clr.l	calcvol+4
	move.b	#2,waitstart
	
	move.l	newmixingperiod(pc),mixingperiod

					;Anfangsadressen setzen
	move.w	#$ff,$dff09e		;Modulierungen aller Art aus !
		
	lea	$dff0a0,a5
	move.w	#SamLEN,d0
	move.l	#SamLEN*2,Lastbuffer

	move.l	mixingperiod(pc),d1

	move.l	Buffer0,a0
	move.l	a0,(a5)
	move.w	d0,4(A5)
	move.w	d1,6(a5)
	clr.w	8(A5)			;volume 0

	lea	$10(a5),a5
	move.l	Buffer1,a0
	move.l	a0,(a5)
	move.w	d0,4(A5)
	move.w	d1,6(a5)
	clr.w	8(A5)

	lea	$10(a5),a5	
	move.l	Buffer2,a0
	move.l	a0,(a5)
	move.w	d0,4(A5)
	move.w	d1,6(a5)
	clr.w	8(A5)

	lea	$10(a5),a5	
	move.l	Buffer3,a0
	move.l	a0,(a5)
	move.w	d0,4(A5)
	move.w	d1,6(a5)
	clr.w	8(A5)

	moveq	#32,d0
	bsr	AmpOut_Paula14_CH_DMAWait	;warten bis DMA wirklich gestoppt
	move.w	#$800f,$dff096			;DMA Start
	moveq	#16,d0				;sufficient for 31 kHz (normally, you wait about 5 lines)
	bsr	AmpOut_Paula14_CH_DMAWait

	pop	all
	rts

;Input: d0 - number of lines to wait
AmpOut_Paula14_CH_DMAWait
	movem.l	d0/d1,-(SP)
	;moveq	#16,d0
.wart2
	move.b	$dff006,d1
.wart
	cmp.b	$dff006,d1
	beq	.wart
	dbf	d0,.wart2
	movem.l	(sp)+,d0/d1
	rts



;=========================================================================
AmpOut_Paula14_INT_AUD:
	move.w	#$780,$dff09c			;set interrupt(s) back
;	move.w	#1<<7,$dff09c			;set first audio interrupt back

	movem.l d0-a6,-(sp)			;TODO: all registers really needed to be saved ?
	move.l	MyEPBase(pc),a5

	move.l	Lastbuffer(pc),d0
	if	NumChipBuffers=2
	eor.l	#SamLEN*2,d0			;128 Worte
	else
	; TODO: AND instead of cmp if numchipbuffers == 4
	add.l	#SamLEN*2,d0
	cmp.l	#SamLEN*2*NumChipBuffers,d0
	blo.s	.lo
	moveq	#0,d0
.lo
	endc
	move.l	d0,LastBuffer

	move.w	#$780,$dff09c			;set interrupt(s) back (second time, some CPU boards are strange...)

	ifne	1
	lea	$dff0a0,a1
	movem.l	Buffer0(pc),d3-d6
	add.l	d0,d3
	move.l	d3,(a1)
	add.l	d0,d4
	move.l	d4,$10(a1)
	add.l	d0,d5
	move.l	d5,$20(a1)
	add.l	d0,d6
	move.l	d6,$30(a1)
	endc

.int
;	move.w	#$f0f,$dff180

	; low priority secondary interrupt for the mixing routine
	move.l	4.w,a6
	lea	AmpOut_Paula14_Int_loopStruct,a1
	cmp.b	#NT_Interrupt,ln_type(a1)
	bne.s	.skip

	jsr	-180(a6)	;_LVOcause

;	move.w	#$f00,$dff180
	movem.l	(sp)+,d0-a6

	rts
.skip
;-- Typ der Struktur ist NT_Softint, somit kommt der Prozessor beim Mixen nicht
;-- hinterher -> Mixrate verringern

	move.l	playmixrate(pc),d0
	sub.l	#6000,d0
	cmp.l	#8000,d0
	bge.s	.ok
	move.l	#8000,d0
.ok
	move.l	d0,playmixrate
	bsr	AmpOut_Paula14_CalcnewmixingPeriod
	move.l	newmixingperiod(pc),mixingperiod

;	move.w	#$f00,$dff180
	movem.l	(sp)+,d0-a6
	rts

;bla1	dc.l	bla

;----------------------------------------------------------------------------------------------------
;--------------------------- setzt Lautstaerke, Laenge, Period neu ------------------------------------
;----------------------------------------------------------------------------------------------------
; TODO: call only when necessary 
AmpOut_Paula14_SetHardWare:
		move	#SamLEN,d7
		lea	$dff0a0,a1

		move.w	#$800f,$dff096-$dff0a0(a1)	;DMA richtig setzen

	ifne	0
	push	all
	move.l	LastBuffer,d0
	lea	$dff0a0,a1
	movem.l	Buffer0(pc),d3-d6
	add.l	d0,d3
	move.l	d3,(a1)
	add.l	d0,d4
	move.l	d4,$10(a1)
	add.l	d0,d5
	move.l	d5,$20(a1)
	add.l	d0,d6
	move.l	d6,$30(a1)
	pop	all
	endc

		move	d7,4(a1)			;length
		move	d7,$10+4(a1)
		move	d7,$20+4(a1)
		move	d7,$30+4(a1)
	
		move.l	newmixingperiod(pc),d1		;period
		move.l	d1,mixingperiod
		move.w	d1,6(a1)
		move.w	d1,6+$10(a1)
		move.w	d1,6+$20(a1)
		move.w	d1,6+$30(a1)
	
		move.l	MyEPBase(pc),a5
		move.l	dtg_SndVol(A5),d0	; dtg_SndVol & dtg_SndLBal
		move.w	dtg_SndRBal(A5),d1	; dtg_SndVol & dtg_SndLBal
	
		cmp.w	tmpvol2(pc),d1
		bne.s	newvol$
		cmp.l	tmpvol(pc),d0
		beq.s	weiter$
newvol$
		move.l	d0,tmpvol
		move.w	d1,tmpvol2

		push	a1
		move.l	my2ndsignal+datas,d1
		moveq	#0,d0
		bset	d1,d0
		move.l	mytask+datas,a1
		move.l	4.w,a6
		jsr	_LVOSignal(a6)
		pop	a1
weiter$		
		move.w	#64,8(A1)		;keep hardware volume and adapt tables instead
		move.w	#64,8+$10(A1)		;
		move.w	#1, 8+$20(A1)		;!!
		move.w	#1, 8+$30(A1)		;!!
		rts

;----------------------------------------------------------------------------------------------------
;------------------------------- Mixroutine Main Loop -----------------------------------------------
;----------------------------------------------------------------------------------------------------
; main call from secondary interrupt (software-int to stay system-friendly)
AmpOut_Paula14_DoMix:
	movem.l	d0-a6,-(Sp)
	bsr	AmpOut_Paula14_SetHardware

;	move.w	#$ff0,$dff180

	move.l	myaudiostruct(pc),a4
;	move.l	a4,0

	;--------------------- mix left panned channels ----------------------
	move	numstructs,d7
	subq.w	#1,d7
	blt.w	.noleft

	move.l	mixad1,a1				; copy mix routine
;	lea	fm_mix,a1
	clr.w	.curchannelsl

.left_loop
		tst.w	AS_LeftRight(a4)
		ble.s	.left_noleft

			lea	TempBuffer,a2			; 16 Bit output buffer (temporary before scaling)
			move.w	#SamLEN*2,d0			;
			move.l	Vtabaddr(pc),a6

			movem.l	a1/a4/d7,-(sp)			; no registers saved by mixer, store needed pointers
			jsr	(a1)				; mix
			movem.l	(sp)+,a1/a4/d7
		
			tst.w	d0
			beq	.left_nochannel			;we expect that moveq is the last command

			add.w	#1,.curchannelsl
			move.l	mixad2,a1
.left_nochannel:
.left_noleft:
		lea	AS_Sizeof(a4),a4
		dbf	d7,.left_loop

		; if no channel was active, Tempbuffer will be "dirty", use pointer to 0-buffer in these cases
		lea	Tempbuffer,a0
		move	.curchannelsl(pc),d0
		bne.s	.leftok
		lea	ZeroSamplebuffer,a0
.leftok:	move.l	a0,.lbuffer
.noleft:

	;--------------------- mix right panned channels ----------------------
	move.l	myaudiostruct(pc),a4

	move	numstructs,d7
	subq.w	#1,d7
	blt.w	.noright
	
	move.l	mixad1,a1	;	
;	lea	fm_mix,a1
	clr.w	.curchannelsr

.right_loop
		tst.w	AS_LeftRight(a4)
		bge.s	.right_noright

			lea	TempBuffer2,a2			; 16 Bit output buffer (temporary before scaling)
			move.w	#SamLEN*2,d0			;
			move.l	VtabaddrR(pc),a6
			movem.l	a1/a4/d7,-(sp)			; no registers saved by mixer, store needed pointers
			jsr	(a1)				; mix
			movem.l	(sp)+,a1/a4/d7
			tst.w	d0
			beq	.right_nochannel		;we expect that moveq is the last command
			
			add.w	#1,.curchannelsr
			move.l	mixad2,a1	;	
.right_nochannel:
.right_noright:
		lea	AS_Sizeof(a4),a4
		dbf	d7,.right_loop

		; if no channel was active, Tempbuffer will be "dirty", use pointer to 0-buffer in these cases
		lea	Tempbuffer2,a0
		move	.curchannelsr(pc),d0
		bne.s	.rightok1
		lea	ZeroSamplebuffer,a0
.rightok1:	move.l	a0,.rbuffer
.noright:

	;--------------------- perform mono/surround, copy to ChipRAM  ----------------------
	move.l	.lbuffer(pc),a0			; 16 Bit mixing buffer (before scaling or zero buffer)
	move.l	.rbuffer(pc),a2			; 16 Bit mixing buffer (before scaling or zero buffer)

	move.l	LastBuffer(pc),d0		;double/multibuffered offset
	move.l	Buffer0,a1			;main output buffer
	move.l	UPSBuffer0,a5			;main UPS buffer
	move.l	SpecialBuffer1,a4		;14 bit low volume output buffer
	add.l	d0,a1
	add.l	d0,a5
	add.l	d0,a4

	move.l	cbufad,a6
	tst.w	.curchannelsl
	bne.s	.left_notempty
	move.l	cbufempty,a6
	bra.s	.left_nofilter
.left_notempty
	move.l	boostad(pc),d0
	beq.s	.left_nofilter
	push	all
	move.l	d0,a5
	lea	HB_DelayL,a1
	move	#SamLEN*2,d7
	jsr	(a5)				
	pop	all
.left_nofilter:
	lea	TBufLeft,a3
	move	prefetchL(pc),d0
	jsr	(a6)

	;--------------------- perform mono/surround, copy to ChipRAM  ----------------------
	move.l	.rbuffer(pc),a0			; 16 Bit mixing buffer (before scaling or zero buffer)
	move.l	.lbuffer(pc),a2			; 16 Bit mixing buffer (before scaling or zero buffer)

	move.l	LastBuffer(pc),d0		; offset for multi-buffering
	move.l	Buffer1,a1			;
	move.l	UPSBuffer1,a5			;
	move.l	SpecialBuffer2,a4		; 14 bit buffer
	add.l	d0,a1
	add.l	d0,a5
	add.l	d0,a4

	ifne	benchmark2
	lea	copybufsurround14old,a6
	else
	move.l	cbufad,a6
	tst.w	.curchannelsr
	bne.s	.right_notempty
	move.l	cbufempty,a6
	bra.w	.right_nofilter
.right_notempty
	move.l	boostad(pc),d0
	beq.s	.right_nofilter
	push	all
	move.l	d0,a5
	lea	HB_DelayR,a1
	move	#SamLEN*2,d7
	jsr	(a5)				
	pop	all
.right_nofilter:
	endc
	
	lea	TBufRight,a3
	move	prefetchR(pc),d0
	jsr	(a6)

	;-------------- Copy Audio Structure back to Player, if necessary -------------------------
	bsr	CopyASBack
	movem.l	(sp)+,d0-a6
	moveq	#0,d0
	rts
	
	; active left/right buffers, either Tempbuffer or Zerobuffer
.lbuffer	dc.l	0
.rbuffer	dc.l	0
	; num active channels left/right
.curchannelsl:	dc.w	0
.curchannelsr:	dc.w	0


*****************************************************************************
*              16 bit to 14 bit tables, including (optional) volume boost   *
* - Input:  no regs but requires                                            *
*            AdditiveArray, Buff14, Usedmode to be initialized              *
* - Output:                                                                 *
*****************************************************************************
AmpOut_Paula14_do14tab:	;initialize 14 bit table (128k memory provided by Inittables call)
	push	d0-d1/a0/a1
        move.l  buff14(pc),d0
        beq.s   .rts
        move.l  d0,a0
        lea     AdditiveArray(pc),a1
        bsr     _Create14Table

	; apply boost
	move.l	usedmode,d0
	beq.s	.rts
	move.l	d0,a0
	move.w	P_Vboost(a0),d0			;VBoost (auf xx.x bezogen,
	
	move.w	P_Flags(a0),d1			;Flags.w
	btst	#PBF_SincMixB,d1
	beq.s	.nohighboost

	move.l	P_Mixrate(A0),d1	;Flags.w
	cmp.l	#HighBoost28_Maxrate,d1
	bge.s	.high44
	mulu.w	#22,d0			;1.4
	bra.s	.highboost_apply
.high44:
	mulu.w	#34,d0			;2.1
.highboost_apply:
	lsr.l	#4,d0			;/16, normalize
.nohighboost:
					;z.B. 082 fuer 8.2) 
	ext.l	d0
        move.l  buff14(pc),a0
	bsr	_Boost14Table
.rts
	pop	d0-d1/a0/a1
        rts

	; Inputs:
	; A0   - 14 bit table, 65536 entries as 16 bit words
	; D0.l - boost factor * 10 (10....80)
_Boost14Table:
	push	d0-d3/d7/a0-a2

	adda.l	#$10000,a0	;middle
	move.l	a0,a1		;middle
	move.l	a0,a2		;middle

	divu	#10,d0		;upper word: 0-9, lower word: steps in samples
	move.l	#65536*10,d1	;threshold

	move.w	#32768-1,d7
	moveq	#0,d2
	moveq	#-1,d3
.negloop:
	move.w	(a2,d2.w*2),(a0)+ ;copy positive
	move.w	(a2,d3.w*2),-(a1) ;copy negative

	add.l	d0,d2		;next pos positive
	sub.l	d0,d3		;next pos negative
	cmp.l	d1,d2		;fractional part overflow ?
	blt.s	.noover
	sub.l	d1,d2
	addq.w	#1,d2
	subq.w	#1,d3
.noover
	cmp.w	#32767,d2
	bhs.s	.fill1

	dbf	d7,.negloop
	bra.s	.finish
.fill1:
	move	#32767,d2
	move	#-32768,d3
.fill:
	subq.w	#1,d7
	blt.s	.finish

	move.w	(a2,d2.w*2),(a0)+
	move.w	(a2,d3.w*2),-(a1)
	bra.s	.fill
.finish
	pop	d0-d3/d7/a0-a2
	rts
                ; Parameters

                ; a0 = Table address
                ; (MUST have enough space for 65536 UWORDS)
                ; a1 = Additive Array
                ; 256 UBYTEs
                ;
                ; the table is organized as follows:
                ; 32768 UWORDS positive range, ascending order
                ; 32768 UWORDS negative range, ascending order
                ; access: (a0,d0.l*2)
                ; where d0.w is signed word sample data
                ; and the upper word of d0.l is *cleared!*

		; note: statement above applies to original implementation
		; here the table is linear, zero point in the middle (!)
		;

_Create14Table  movem.l a2/d2-d6,-(sp)

                lea     128(a1),a2
                add.l   #2*32768,a0             ; place at the middle of the table

                move.l  a2,a1                   ; count the number of steps
                moveq   #128-1,d0               ; in the positive range
                moveq   #0,d5
.countpositive  move.b  (a1)+,d1
                ext.w   d1
                ext.l   d1
                add.l   d1,d5
                dbra    d0,.countpositive       ; d5=number of steps
                move.l  #32768,d6               ; reset stretch counter

                move.l  a2,a1                   ; middle value in calibdata
                move.w  #32768-1,d0             ; number of positive values -1
                moveq   #0,d1                   ; HI value
                moveq   #0,d2                   ; LO value
                moveq   #0,d3                   ; counter
.fetchnext2     move.b  (a1)+,d4                ; add calibtable to counter
                ext.w   d4
                add.w   d4,d3
.outerloop2     tst.w   d3
                bgt.s   .positive2
.negative2      addq.w  #1,d1                   ; increment HI value
                sub.w   d4,d2                   ; reset LO value
                bra.s   .fetchnext2
.positive2      move.b  d1,(a0)+                ; store HI and LO value
                move.b  d2,(a0)+
                sub.l   d5,d6                   ; stretch the table
                bpl.s   .repeat2                ; to 32768 entries
                add.l   #32768,d6
                addq.w  #1,d2                   ; increment LO value
                subq.w  #1,d3                   ; decrement counter
.repeat2        dbra    d0,.outerloop2

                move.l  a2,a1                   ; count the number of steps
                moveq   #128-1,d0               ; in the negative range
                moveq   #0,d5
.countnegative  move.b  -(a1),d1
                ext.w   d1
                ext.l   d1
                add.l   d1,d5
                dbra    d0,.countnegative       ; d5=number of steps
                move.l  #32768,d6               ; reset stretch counter

                sub.l   #2*32768,a0             ; again place at the middle of the table

                move.l  a2,a1                   ; middle value in calibdata
                move.w  #32768-1,d0             ; number of negative values -1
                moveq   #-1,d1                  ; HI value
                moveq   #-1,d2                  ; LO value
                moveq   #0,d3                   ; counter
.fetchnext1     move.b  -(a1),d4                ; add calibtable to counter
                ext.w   d4
                add.w   d4,d3
                add.w   d4,d2                   ; maximize LO value
.outerloop1     tst.w   d3
                bgt.s   .positive1
.negative1      subq.w  #1,d1
                bra.s   .fetchnext1
.positive1      move.b  d2,-(a0)                ; store LO and HI value
                move.b  d1,-(a0)
                sub.l   d5,d6                   ; stretch the table
                bpl.s   .repeat1                ; to 32768 entries
                add.l   #32768,d6
                subq.w  #1,d2                   ; decrement lo value
                subq.w  #1,d3                   ; decrement counter
.repeat1        dbra    d0,.outerloop1

                movem.l (sp)+,a2/d2-d6
                rts
;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
;-------------------------------------- Paula 14 Bit Code End ----------------------------------------
;-----------------------------------------------------------------------------------------------------

;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
;--------------------------------- 14 Bit Output Routines  -------------------------------------------
;-----------------------------------------------------------------------------------------------------
; Concept:   collect a variable number of channels in the mixing routines (unsigned), and scale it to
;            the relevant 14 bit dynamic range afterwards
; Workflow:

; There are several routines for this phase:
;  1. empty copy   - just zeroes the chipram buffer when no mixing was done
;  2. stereo copy  - fill one chipram buffer pair per channel
;  3. surround copy- apply the delayed opposite channel to the current channel
;                  - branch to stereo copy
;
; Input:
; Output:
;         -
;
; Used Registers: all, none saved
; 
; Notes/Caveats: 
;        number of samples still hardcoded (SamLEN*2)
;-----------------------------------------------------------------------------------------------------



;---------------------- no channel playing, write silence to two channels --
; Inputs:
;  D0 - prefetch
;  A0 - input samples main channel (left/right)
;  A1 - output samples high volume channel
;  A2 - input samples other channel (for mono/surround)
;  A3 - temp buffer for delayed samples in mono/surround
;  A4 - output samples low volume channel
;  A5 - UPS buffer in fastram, same contents as high volume channel
copybufemptyStereo14:
	ifne	benchmark
	move	#$3ff,$dff180
	endc

	move	#SamLEN*2/16-1,d7
	moveq.l	#0,d1
	moveq.l	#0,d2
	moveq.l	#0,d3
	moveq.l	#0,d4

	add.l	#SamLEN*2,A1
	add.l	#SamLEN*2,A4
	add.l	#SamLEN*2,A5
.ldd3
	movem.l	d1-d4,-(a1)	;main samples
	movem.l	d1-d4,-(a4)	;low volume (14 bit trick) samples
	movem.l	d1-d4,-(a5)	;UPS buffer
	dbf	d7,.ldd3

	ifne	benchmark
	move	#$888,$dff180
	endc
	rts

;---------------------- fuer Stereo 14 Bit ---------------------
; Inputs (*=used here):
;   D0 - prefetch
; * A0 - input samples main channel (left/right)
; * A1 - output samples high volume channel
;   A2 - input samples other channel (for mono/surround)
;   A3 - temp buffer for delayed samples in mono/surround
; * A4 - output samples low volume channel
; * A5 - UPS buffer in fastram, same contents as high volume channel
copybufStereo14:		;for 68020-68060, UPS Struktur in fast ram
	ifne	benchmark
	move	#$0ff,$dff180
	endc

	move.l	buff14_2(pc),a2 ;middle of the 14 bit table
	
	move	#SamLEN*2/4-1,d7	;mixed samples=SamLEN*2, 4 samples in one loop
	moveq	#8,d6
        movem.w (a0)+,d0/d2/d4/d5                       ;auto ext.l !!! SIGNED.l !! since 68000!
.ldd3
        move.w  (a2,d0.l*2),d1		;h0 l0
        move.w  (a2,d2.l*2),d3		;h1 l1

	move.w	d1,d0				;h0 l0
	ror.l	d6,d3				;l1 xx xx h1
	move.b	d3,d1				;h0 h1
	move.b	d0,d3				;l1 xx xx l0
	swap	d1				;h0 h1 xx xx
	ror.l	d6,d3				;l0 l1 xx xx

        move.w  (a2,d5.l*2),d3			;l0 l1 h3 l3
        move.w  (a2,d4.l*2),d1			;h0 h1 h2 l2
	ror.l	d6,d3				;l3 l0 l1 h3
	move.l	d1,d4				;h0 h2 h2 l2
	move.b	d3,d1				;h0 h1 h2 h3
	move.b	d4,d3				;l3 l0 l1 l2
	rol.l	d6,d3				;l0 l1 l2 l3
        movem.w (a0)+,d0/d2/d4/d5               ;auto ext.l !!! SIGNED.l !! since 68000!

        move.l  d3,(a4)+	;support low volume channel ( volume 1)
        move.l  d1,(a1)+	;main high volume channel ( volume 64 )
	move.l	d1,(a5)+	;UPS buffer

        dbf     d7,.ldd3

	ifne	benchmark
	move	#$0,$dff180
	endc

	rts

;---------------------- Surround 14 Bit ---------------------
; Inputs:
;  D0 - prefetch
;  A0 - input samples main channel (left/right)
;  A1 - output samples high volume channel
;  A2 - input samples other channel (for mono/surround)
;  A3 - temp buffer for delayed samples in mono/surround
;  A4 - output samples low volume channel
;  A5 - UPS buffer in fastram, same contents as high volume channel
;
; Notes: empty copy routine disabled in surround mode because of other channel possibly
;        contributing to sum channel
; Note:  branches to copybufStereo14 after completing surround calculations
copybufSurround14:	;68040/68060 targeted routine (multiplies instead of table accesses)
	ifne	benchmark
	move	#$4f4,$dff180
	endc

	; apply surround
	; 3 parts:
	; - use old samples from A3
	; - use new samples from A2 (new data)
	; - copy remaining samples from A2 to temporary buffer
	move	effectvol(pc),d6		; effect volume 0..64
	neg.w	d6
	lsl.w	#2,d6				; 0...256
	lea	SurroundTemp,A6			; temporary storage for surround signal combination

	push	a3
	push	d0
	move	d0,d7				; number of samples / 2 to write in first phase
	lsr.w	#1,d7				; we write four samples in each run
	subq	#1,d7
	
	; join new and delayed samples (from other channel)
.cbsr14_joindelayed:				
	movem.w (a3)+,d0/d1/d2/d3		; delayed samples from other channel D0,D1,D2,D3
	muls.w	d6,d0				; D0
	muls.w	d6,d1				; D1
	move.l	(a0)+,d4			; new samples from current channel (S0 S1)
	muls.w	d6,d2				; D2
	muls.w	d6,d3				; D3
	move.l	(a0)+,d5			; new samples from current channel (S2 S3)
	lsl.l	#8,d0				; (D0 xx)
	asr.l	#8,d1				; (xx D1)
	clr.w	d0				; (D0 00) -> avoid lower word overflow by +D4.l
	lsl.l	#8,d2				; (D2 xx)
	asr.l	#8,d3				; (xx D3)

	clr.w	d2				; (D2 00) -> avoid lower word overflow by +D5.l
	add.l	d4,d0				; (D0+S0 xx)
	add.w	d4,d1				; (xx    D1+S1)
	add.l	d5,d2				; (D2+S2 xx)
	add.w	d5,d3				; (xx    D3+S3)

	move.w	d1,d0				; (D0+S0 D1+S1)
	move.w	d3,d2				; (D2+S2 D3+S3)

	move.l	d0,(A6)+
	move.l	d2,(A6)+
	dbf	d7,.cbsr14_joindelayed

	; second run: join new samples
	move.w	#SamLEN,d7
	sub.w	2(a7),d7
	lsr.w	#1,d7
	subq.w	#1,d7
.cbsr14_joincurrent:				; join new and delayed samples (from other channel)
	movem.w (a2)+,d0/d1/d2/d3		; delayed samples from other channel D0,D1,D2,D3
	muls.w	d6,d0				; D0
	muls.w	d6,d1				; D1
	move.l	(a0)+,d4			; new samples from current channel (S0 S1)
	muls.w	d6,d2				; D2
	muls.w	d6,d3				; D3
	move.l	(a0)+,d5			; new samples from current channel (S2 S3)
	lsl.l	#8,d0				; (D0 xx)
	asr.l	#8,d1				; (xx D1)
	clr.w	d0				; (D0 00) -> avoid lower word overflow by +D4.l
	lsl.l	#8,d2				; (D2 xx)
	asr.l	#8,d3				; (xx D3)

	clr.w	d2				; (D2 00) -> avoid lower word overflow by +D5.l
	add.l	d4,d0				; (D0+S0 xx)
	add.w	d4,d1				; (xx    D1+S1)
	add.l	d5,d2				; (D2+S2 xx)
	add.w	d5,d3				; (xx    D3+S3)

	move.w	d1,d0				; (D0+S0 D1+S1)
	move.w	d3,d2				; (D2+S2 D3+S3)

	move.l	d0,(A6)+
	move.l	d2,(A6)+
	dbf	d7,.cbsr14_joincurrent

	pop	d0
	pop	a3

	; post part: store samples for next run
	; TODO: integrate into first loop (minor issue)
	subq	#1,d0
.cbsr14_post1:
	move.l	(a2)+,(a3)+
	dbf	d0,.cbsr14_post1

	; actual copy to chipram phase
	lea	SurroundTemp,A0			; temporary storage for surround signal combination
	bra	copybufStereo14			


	ifne	0
;---------------------- Surround 14 Bit ---------------------
; Inputs:
;  D0 - prefetch
;  A0 - input samples main channel (left/right)
;  A1 - output samples high volume channel
;  A2 - input samples other channel (for mono/surround)
;  A3 - temp buffer for delayed samples in mono/surround
;  A4 - output samples low volume channel
;  A5 - UPS buffer in fastram, same contents as high volume channel
;
; Notes: empty copy routine disabled in surround mode because of other channel possibly
;        contributing to sum channel
copybufSurround14_030: ; 
	ifne	benchmark
	move	#$0f0,$dff180
	endc

	;TODO - push A1, move data to (a5)+
	;     - use D4 as usual data register to enable superscalar execution
	;     - pop A1, copy UPSbuffer to ChipRAM
	;
;	move	effectvol(pc),d2		; effect volume 0..64 
;	lsl.w	#
		
	move.l	A4,d4
	sub.l	A1,d4		;difference between A1 and A4

	move.l	buff14_2(pc),A6 		;middle of the 14 bit table
	move.l	SurroundTabAddrMid(pc),a4	;middle of the surround table

	move	d0,d7
	lsr.w	#1,d7
	subq	#1,d7

	moveq	#8,d6

	push	a3
	push	d0
	; pre part: copy stored samples of other stream into current output
.cbsr14_pre1:
	movem.w	(a0)+,d1/d5			;new samples (S0 S1)
	movem.w (a3)+,d0/d2			;delayed samples from other channel (D0 D1)

	add	(A4,d0.w*2),d1			;scale old temporary value
	add	(A4,d2.w*2),d5			;scale old temporary value
	move.w	(A6,d1.w*2),d1			;h0 l0
	move.w	(A6,d5.w*2),d3			;h1 l1
	move.w	d1,d5				;h0 l0
	ror.l	d6,d3				;l1 xx xx h1
	move.b	d3,d1				;h0 h1
	move.b	d5,d3				;l1 xx xx l0

	ifne	0
	swap	d1				;h0 h1 xx xx
	ror.l	d6,d3				;l0 l1 xx xx

	move.l	(a0)+,d0		;next new samples (S2 S3)
	move.l	(a3)+,d2		;next delayed samples (D2 D3)
	add	(A4,d2.w*2),d0		;S2 S3+D3
	swap	d2			;D3 D2
	move.w	(A6,d0.w*2),d1		;h0 h1 h3 l3
	swap	d0			;S3+D3 S2
	move.l	d1,d5			;h0 h1 h3 l3
	lsr.w	d6,d1			;h0 h1 00 h3
	add	(A4,d2.w*2),d0		;S3+D3 S2+D3
	move	#$ff00,d2		;xx xx ff 00
	move.w	(A6,d0.w*2),d3		;l0 l1 h2 l2
	and.l	d3,d2			;xx xx h2 00
	lsl.w	d6,d3			;l0 l1 l2 00
	or.w	d2,d1			;h0 h1 h2 h3
	move.b	d5,d3			;l0 l1 l2 l3

	move.l	d3,(a1,d4.l)
	move.l	d1,(a1)+

	else
	rol.l	d6,d3				;xx xx l0 l1

	move.w	d3,(a1,d4.l)			;store lower bytes
	move.w	d1,(a1)+			;store upper bytes

	movem.w	(a0)+,d1/d5			;new samples
	movem.w (a3)+,d0/d2			;delayed samples from other channel

	add	(A4,d0.w*2),d1			;scale old temporary value
	add	(A4,d2.w*2),d5			;scale old temporary value
	move.w	(A6,d1.w*2),d1			;h0 l0
	move.w	(A6,d5.w*2),d3			;h1 l1
	move.w	d1,d5				;h0 l0
	ror.l	d6,d3				;l1 xx xx h1
	move.b	d3,d1				;h0 h1
	move.b	d5,d3				;l1 xx xx l0

;		swap	d1				;h0 h1 xx xx
;		ror.l	d6,d3				;l0 l1 xx xx
		rol.l	d6,d3				;xx xx l0 l1

	move.w	d3,(a1,d4.l)			;store lower bytes
	move.w	d1,(a1)+			;store upper bytes
	endc

	dbf	d7,.cbsr14_pre1

	; main part: copy samples from current stream and other stream
	move	#SamLEN-1,d7
	sub.l	(a7),d7
;	sub	d0,d7
.cbsr14:
	move	(a0)+,d1
	move	(a2)+,d2			;old temporary value
	add	(A4,d2.w*2),d1			;scale old temporary value
	move.w	(A6,d1.w*2),d1
	move.b	d1,d3				;lower byte = low volume
	lsl	d6,d3				;upper byte = low volume 1st sample

	move	(a0)+,d5
	move	(a2)+,d2			;old temporary value
	add	(A4,d2.w*2),d5			;scale old temporary value
	move.w	(A6,d5.w*2),d5
	move.b	d5,d3				;two lower bytes (low volumes)
	lsr	d6,d5				;upper byte -> lower byte
	move.b	d5,d1				;two upper bytes (high volumes)

	move.w	d3,(a1,d4.l)			;store lower bytes
	move.w	d1,(a1)+			;store upper bytes

	dbf	d7,.cbsr14
	
	pop	d0
	pop	a3				;
	; post part: store samples for next run
	subq	#1,d0
.cbsr14_post1:
	move.l	(a2)+,(a3)+
	dbf	d0,.cbsr14_post1

	ifne	benchmark
	move	#$0,$dff180
	endc

	rts
	endc

	ifne	benchmark2
;---------------------- Surround 14 Bit ---------------------
; Inputs:
;  D0 - prefetch
;  A0 - input samples main channel (left/right)
;  A1 - output samples high volume channel
;  A2 - input samples other channel (for mono/surround)
;  A3 - temp buffer for delayed samples in mono/surround
;  A4 - output samples low volume channel
;  A5 - UPS buffer in fastram, same contents as high volume channel
;
; Notes: empty copy routine disabled in surround mode because of other channel possibly
;        contributing to sum channel
copybufSurround14OLD:	; first working version, not fast (16 bit stores to ChipMem)
	ifne	benchmark
	move	#$f0f,$dff180
	endc

	move.l	A4,d4
	sub.l	A1,d4		;difference between A1 and A4

	move.l	buff14_2(pc),A6 		;middle of the 14 bit table
	move.l	SurroundTabAddrMid(pc),a4	;middle of the surround table

	move	d0,d7
	subq	#1,d7

	moveq	#8,d6

	push	a3
	; pre part: copy stored samples of other stream into current output
.cbsr14_pre1:
	move	(a0)+,d1
	move	(a3)+,d2			;old temporary value
	add	(A4,d2.w*2),d1			;scale old temporary value
	move.w	(A6,d1.w*2),d1
	move.b	d1,d3				;lower byte = low volume
	lsl	d6,d3				;upper byte = low volume 1st sample

	move	(a0)+,d5
	move	(a3)+,d2			;old temporary value
	add	(A4,d2.w*2),d5			;scale old temporary value
	move.w	(A6,d5.w*2),d5
	move.b	d5,d3				;two lower bytes (low volumes)
	lsr	d6,d5				;upper byte -> lower byte
	move.b	d5,d1				;two upper bytes (high volumes)

	move.w	d3,(a1,d4.l)			;store lower bytes
	move.w	d1,(a1)+			;store upper bytes

	dbf	d7,.cbsr14_pre1

	; main part: copy samples from current stream and other stream
	move	#SamLEN-1,d7
	sub	d0,d7
.cbsr14:
	move	(a0)+,d1
	move	(a2)+,d2			;old temporary value
	add	(A4,d2.w*2),d1			;scale old temporary value
	move.w	(A6,d1.w*2),d1
	move.b	d1,d3				;lower byte = low volume
	lsl	d6,d3				;upper byte = low volume 1st sample

	move	(a0)+,d5
	move	(a2)+,d2			;old temporary value
	add	(A4,d2.w*2),d5			;scale old temporary value
	move.w	(A6,d5.w*2),d5
	move.b	d5,d3				;two lower bytes (low volumes)
	lsr	d6,d5				;upper byte -> lower byte
	move.b	d5,d1				;two upper bytes (high volumes)

	move.w	d3,(a1,d4.l)			;store lower bytes
	move.w	d1,(a1)+			;store upper bytes

	dbf	d7,.cbsr14

	pop	a3				;
	; post part: store samples for next run
	subq	#1,d0
.cbsr14_post1:
	move.l	(a2)+,(a3)+
	dbf	d0,.cbsr14_post1

	ifne	benchmark
	move	#$0,$dff180
	endc

	rts
	endc	;benchmark2


;---------------------- Mono 14 Bit ---------------------
; Inputs:
;  D0 - prefetch
;  A0 - input samples main channel (left/right)
;  A1 - output samples high volume channel
;  A2 - input samples other channel (for mono/surround)
;  A3 - temp buffer for delayed samples in mono/surround
;  A4 - output samples low volume channel
;  A5 - UPS buffer in fastram, same contents as high volume channel
;
; Notes: empty copy routine disabled in surround mode because of other channel possibly
;        contributing to sum channel
copybufMono14:
	move.l	A4,d4
	sub.l	A1,d4		;difference between A1 and A4

	move.l	buff14_2(pc),A6 		;middle of the 14 bit table
	moveq	#8,d6
	move	#SamLEN-1,d7

	ifne	1

	move    effectvol(pc),d0                ; effect volume 0..64

.cbsr14:
	move	d0,d1				;effvol
	muls	(a2)+,d1			;effvol * other channel
	asr.l	#6,d1				;effvol * other channel normalized
	add.w	(a0)+,d1			;curchannel + other channel
	move.w	(A6,d1.w*2),d1			;x x Hv0 Lv0

	move.b	d1,d3				;x x x   Lv0
	lsl	d6,d3				;x x Lv0 0

	move	d0,d5				;effvol
	muls	(a2)+,d5			;effvol * other channel
	asr.l	#6,d5				;effvol * other channel normalized
	add.w	(a0)+,d5			;curchannel + other channel
	move.w	(A6,d5.w*2),d5			;x x Hv1 Lv1

	move.b	d5,d3				;x x Lv0 Lv1
	lsr	d6,d5				;0 x   x Hv1
	move.b	d5,d1				;x x Hv0 Hv1

	move.w	d3,(a1,d4.l)			;store lower bytes
	move.w	d1,(a1)+			;store upper bytes
	move.w	d1,(a5)+	;UPS buffer

	dbf	d7,.cbsr14

	else

	move.l	SurroundTabAddrMid(pc),a4	;middle of the surround table
.cbsr14:
	move	(a0)+,d1
	move	(a2)+,d2			;other channel
	add	(A4,d2.w*2),d1			;scale other channel input
	move.w	(A6,d1.w*2),d1


	move.b	d1,d3				;lower byte = low volume
	lsl	d6,d3				;upper byte = low volume 1st sample

	move	(a0)+,d5
	move	(a2)+,d2			;other channel
	add	(A4,d2.w*2),d5			;scale other channel input
	move.w	(A6,d5.w*2),d5

	move.b	d5,d3				;two lower bytes (low volumes)
	lsr	d6,d5				;upper byte -> lower byte
	move.b	d5,d1				;two upper bytes (high volumes)

	move.w	d3,(a1,d4.l)			;store lower bytes
	move.w	d1,(a1)+			;store upper bytes
	move.w	d1,(a5)+	;UPS buffer

	dbf	d7,.cbsr14

	endc

	rts

;----------------------------------- 14 Bit Amplifier Data ---------------------------------
AmpOut_Paula14_OLDINT0:	dc.l	0
;OLDINT1:	dc.l	0
;OLDINT2:	dc.l	0
;OLDINT3:	dc.l	0



AmpOut_Paula14_Int_loopStruct
	dc.l	0
	dc.l	0
	dc.b	2
	dc.b	32
	dc.l	INT_Name1		;Name
	dc.l	0
	dc.l	AmpOut_Paula14_Domix	;Interrupt Routine

AmpOut_Paula14_Aud0Interrupt:
	dc.l	0
	dc.l	0
	dc.b	2
	dc.b	127
	dc.l	INT_Name1		;Name
	dc.l	0
	dc.l	AmpOut_Paula14_INT_Aud	;Interrupt Routine

INT_Name1	dc.b	`Mixingroutine Audio Interrupt`,0

Paula_Mixfreqs_PAL:
;PAL mixing frequencies,3546895/n
;               160   124   123   80    74    65    64
	dc.l	22168,28603,28836,44336,47931,54567,55420,0
Paula_Mixfreqs_NTSC:
;NTSC mixing frequencies 3579545/n
;               162   124   81    74    65    64
	dc.l	22095,28867,44191,48372,55069,55930,0


