*****************************************************************************
** Engine: 14 Bit Amplifier II                                             **
**                                                                         **
** Project: Eagleplayer 2.05                                               **
** Authors: Jan Blumenthal & Henryk Richter                                **
** Start  : 1993/01/09                                                     **
** $Rev: 23 $ $Author: buggs $                                             **
** $Date: 2016-09-03 15:04:04 +0200 (Sa, 03. Sep 2016) $                   **
**                                                                         **
*****************************************************************************
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program (See the included file COPYING);
** if not, write to the Free Software Foundation, Inc.,
** 675 Mass Ave, Cambridge, MA 02139, USA.
**
**************************************************************************** 

;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
;-------------------------------------- Toccata Code -------------------------------------------------
;-----------------------------------------------------------------------------------------------------
AmpOut_Toccata:
		dc.l	AmpOut_Toccata_Init
		dc.l	AmpOut_Toccata_End
		dc.l	AmpOut_Toccata_StartInt
		dc.l	AmpOut_Toccata_StopInt
		dc.l	AmpOut_Toccata_Check
		dc.l	AmpOut_Toccata_GetMixRates
		dc.l	AmpOut_Toccata_SetNewMixingPeriod	;OHW_SetMixRate (optional)
		dc.l	AmpOut_Toccata_SetPlayMode		;OHW_SetPlayMode (optional)
		dc.l	AmpOut_Toccata_SetBoost			;OHW_SetBoost (optional)
		dc.l	0
		dc.l	AmpOut_Toccata_Exit			;OHW_Exit
		dc.l	0					;OHW_Reserved
	ifne	(*-AmpOut_Toccata)-OHW_Sizeof
	error	"Structure Size Mismatch!"
	endc

Toccata_SECONDINT	equ	0

TB	macro
	move.l	Toccata_base(pc),a6
	endm

;--------------------- return whether this output hardware is available --------------------------------
;input:  A5 - datas
;output: D0 - 1=yes, 0=no
;regs:   all saved (except D0)
AmpOut_Toccata_Exit:
	push	d1-a6
	move.l	Toccata_base(pc),d0
	beq.s	exit$
	move.l	d0,a1
	move.l	4.w,a6
	jsr	_LVOCloseLibrary(a6)
	clr.l	Toccata_base
exit$
	pop	d1-a6
	rts


;--------------------- return whether this output hardware is available --------------------------------
;input:  A5 - datas
;output: D0 - 1=yes, 0=no
;regs:   all saved (except D0)
AmpOut_Toccata_Check:
	push	d1-a6

	clr.l	toccata_nfreqs

	move.l	4.w,a6
	lea	toccata_name,a1
	moveq	#0,d0			;ab V6 wird TT_RawInt unterst?tzt
	CallSys	OpenLibrary
	move.l	d0,Toccata_base
	bne.s	libfound$

	lea     msg_ToccataLib,a0	;Infotext ausgeben (Requester)
        bsr     Merror

	moveq	#0,d0
	bra.s	ret$
libfound$

	bsr	AmpOut_Toccata_GetMixRates
	tst.l	d0
	bne.s	mixrates_found$		;no frequencies ? -> toccata is not present or not working

	lea     msg_ToccataHW,a0	;Infotext ausgeben (Requester)
        bsr     Merror

	moveq	#0,d0
	bra.s	ret$

mixrates_found$
	moveq	#1,d0			;yep, toccata.library is present and seems to work
ret$
	pop	d1-a6
	rts

;--------------------- return available mixing rates as long int array --------------------------------
;input:  A5 - datas (Amplifier)
;output: D0 - pointer to longword array of mixing frequencies in Hz
;regs:   all saved
;note:   this list is informational for the slider, the init routine calls FindFrequency() each time to make sure...
AmpOut_Toccata_GetMixRates:
	push	d1-a6
	move.l	Toccata_Base(pc),d0
	beq.s	nofreqs$

	lea	toccata_freqs(pc),a2
	moveq	#25-1,d2	;max. 25

	moveq	#0,d0
	moveq	#0,d1
loop$
	TB
	CallSys	T_NextFrequency
	move.l	d0,(a2)+
	beq	end$
	addq	#1,d1
	dbf	d2,loop$
	clr.l	(a2)+		;clean end of list
end$
	moveq	#0,d0
	move.l	d1,toccata_nfreqs
	beq.s	nofreqs$
	lea	toccata_freqs(pc),a2
	move.l	a2,d0
nofreqs$
	pop	d1-a6
	rts

******************************************************************************
*                   sucht geeignete Frequenz zum Abspielen raus              *
******************************************************************************
;Input:  D0 - desired frequency
;Output: D0 - found frequency or 0 for error
Ampout_Toccata_FindFrequency:
	push	d1-a6
	lea	datas,a5

;	move.l	#44100,d0		;sane default (handled elsewhere)
	TB
	move.l	a6,d1
	beq	notoc$
	CallSys	T_FindFrequency
notoc$
	pop	d1-a6
	rts

;--------------------- set surround mode mono-stereo-surround --------------------------------
;input:  A5 - datas (Amplifier)
;        D0 - mode (<0=mono, 0=stereo, >=0 surround), range is <64...64 
;output: D0 - ok = 0 / error <>0
;regs:   all saved
;note:   optional, set pointer to NULL if you won't implement this
AmpOut_Toccata_SetPlayMode:
	push	d1-a6

	move.l	d0,d1

	lea	copybufStereoT,a4
	lea	copybufemptyStereoT,a6

	tst.b	d1			; Stereo ? -> use stereo routine (faster than the other)
	beq.s	.setit

;	lea	copybufSurround14_030,a4 ; table based surround
;	move.l	SurroundTabAddrMid(pc),d2
;	bne.w	.tablebasedsurround	; use table based surround when table is present

	lea	copybufSurroundT,a4	; surround routine used in empty and non-empty cases

.tablebasedsurround:
	move.l	a4,a6			; 

	tst.b	d1
	bpl.s	.setit

	lea	copybufMonoT,a4		; surround routine used in empty and non-empty cases
	move.l	a4,a6			; 
	neg.b	d1
.setit:

	lea	cbufad,a5
	move.l	a4,(a5)+	;cbufad
	move.l	a6,(a5)+	;cbufempty

	and.w	#127,d1		;8 Bit -> 16 Bit
	cmp.w	#64,d1
	ble.s	.efvolok
	moveq	#64,d1
.efvolok
	move	d1,effectvol	

	; TODO: UPS buffer fill in 14 Bit mode
	lea	myupsstruct,a6
	lea	UPSBuffer0,a5

	lea	UPSBuffer0a,a0
	lea	UPSBuffer1a,a1

	move.l	a0,(a5)+
	move.l	a0,ups_voice1adr(A6)
	move.l	a0,ups_voice4adr(A6)

	move.l	a1,(a5)+
	move.l	a1,ups_voice2adr(A6)
	move.l	a1,ups_voice3adr(A6)

	move.l	a1,(a5)+		; irrelevant
	move.l	a0,(a5)+		; irrelevant

	move.w	#64,ups_voice1vol(a6)
	move.w	#64,ups_voice2vol(a6)
	move.w	#0,ups_voice3vol(a6)
	move.w	#0,ups_voice4vol(a6)

	clr.l	tmpvol
	clr.w	tmpvol2

	bsr	Makevoltable	; update volume table
	bsr	MakeSurroundTab ; update surround table (68030 and below)

	pop	d1-a6
	moveq	#0,d0
	rts


; D0: boost*10
AmpOut_Toccata_SetBoost:
	push	all
	move.l	buff14,d7		;128 kByte Buffer, allocated internally, available for the output HW to use
	beq.s	noboost$
	move.l	d7,a0

	; quick'n'easy boost table, linear boost with saturation
	ext.l	d0
	move.w	#$8000,d1		;-32768
	ext.l	d1
	move.l	d1,d3
negloop$
	move.l	d1,d2
	muls.l	d0,d2			;*(boost*10)
	divs.l	#10,d2			;/10
	cmp.l	d3,d2			;-32768.l
	bgt.s	negok$
	move.l	d3,d2			;
negok$
	move.w	d2,(a0)+

	addq.l	#1,d1
	bne.s	negloop$		;zero excluded

	moveq	#0,d1
	move.w	#32767,d3
	ext.l	d3
posloop$
	move.l	d1,d2
	muls.l	d0,d2			;*(boost*10)
	divs.l	#10,d2			;/10
	cmp.l	d3,d2			;32767.l
	ble.s	posok$
	move.l	d3,d2			;
posok$
	move.w	d2,(a0)+

	addq.l	#1,d1
	cmp.l	d3,d1
	ble.s	posloop$

noboost$
	pop	all
	rts


;---------------- Routines for the 14 Bit Player --------------------
; Init/End
;          - Initplayer -> A0: tag list
;                          A5: EP globals (not Amplifier globals!)
;                       -> return  D0: 0=OK, <>0=Error
;
; Start-Stop: Startint (no parameters, except A5=EP globals)
;
;             Stopint  (no parameters, except A5=EP globals)
;
AmpOut_Toccata_Init:
	move.l	Toccata_Base(pc),d0
	beq	.err

	move.l	#EPAMT_Flags,d0
	jsr	ENPP_FindTag(A5)
	beq	.err

	moveq	#1,d2
	move.l	d0,d1
	move.l	d0,d6
	and.l	#EPAMB_WaitForStruct,d6

	and.l	#EPAMB_8Bit,d0
	bne	.ok

	moveq	#2,d2
	btst	#EPAMF_8BitUnsigned,d1
	bne	.ok
	
	moveq	#3,d2
	btst	#EPAMF_16Bit,d1			;erstmal nur 16 Bit signed
	beq	.err
.ok
	move.b	d2,fformat

	move.l	#EPAMT_Audiostructs,d0
	jsr	ENPP_FindTag(A5)
	beq	.err
	move.l	d0,d7

	move.l	#EPAMT_NumStructs,d0
	jsr	ENPP_FindTag(A5)
	beq	.err
	move.l	d7,a0

	move	d0,numstructs

	subq.w	#1,d0			;Anzahl der Audio-Strukturen
	blt	.err

	moveq	#0,d1
	moveq	#0,d2
	lea	(a0),a2
.countvoc
	tst.w	AS_LeftRight(a2)
	bpl.s	.addL
	bmi.s	.addR
	bra.s	.noADD
.addL
	cmp.w	#2,AS_Leftright(a2)
	beq	.err
	addq.w	#1,d1
	bra.s	.noADD
.addR
	addq.w	#1,d2
.noadd
	lea	as_sizeof(a2),a2
	dbf	d0,.countvoc

	move.w	d1,numleft		;remember left/right voices -> may still lead to error (!)
	move.w	d2,numright

	move.l	#mix,mixad1		;sane default: 8 bit mixing
	move.l	#mix2,mixad2		;sane default: 8 bit mixing -> will be overwritten by setMixRoutine call
	st	onplay

	move.w	d2,d0
	add.w	d1,d0
	beq	.err			;no active voices ? none ? any ? Why, but why ?

	cmp	#32,d0			;more than 32 active voices ?
	bhi	.err			;bail out for now (though it would be easy to support more than that)

        cmp.b   #3,fformat
        bne.s   .normal8bit

        cmp     #1,d2			;currently, only two channels supported in 16 Bit mode -> subject to change
        bne     .err
        cmp     #1,d1
        bne     .err

.normal8bit
	subq.w	#2,d0
	bgt.s	.hi1
	moveq	#0,d0
.hi1
	mulu	#P_Sizeof,d0
	lea	Prefslist+datas,a1
	lea	(a1,d0.w),a1
	move.l	a1,usedmode
	
	move.w	P_Flags(A1),d0		;Flags.w (Bit0=disable)
	and.b	#PBB_Disable,d0
	bne	.err			;Modus disabled !!

	move.l	a0,epg_audiostruct(A5)	;Audiostruct setzen
	tst.l	d6			;Flag fuer "WaitforStruct" gesetzt ?
	beq.s	.normal
	lea	Extra_AS,a0		;ja, also eigene AS Struktur verwenden
.normal
	move.l	a0,myAudioStruct	;AudioStruct vom Replayer sichern

	bsr	setMixRoutine		;requires "usedmode"

	bsr	InitTables
	bne.s	.err2



	move.l	usedmode,a1
	move.b	6(a1),d0		;Playmode
	ext.w	d0
	ext.l	d0
	bsr	AmpOut_Toccata_SetPlayMode

	move.b	7(a1),d0
	bsr	setPrefetch

	moveq	#0,d0
	move.w	P_Vboost(a1),d0
	bsr	AmpOut_Toccata_SetBoost

	move.w	numleft,d0
	add.w	numright,d0
	bsr	SelectEntry

	bsr	Copy_AS2

	moveq	#0,d0
	rts
.err
	moveq	#-1,d0
.err2
	sf	onplay
	rts

;----------------------------- "End" -----------------------------------
;
; Input:
;  A5 - EP Globals
; Output:
;  D0 - 0=ok, else error
AmpOut_Toccata_End:
	clr.b	onplay
	move	#-1,Currentnumvoc+datas		;kein Modul am Spielen

	bsr	myFreemem			;clear allocated memory

	moveq	#0,d0
	rts

;----------------------------- "Play" -----------------------------------
;
; Input:
;  A5 - EP Globals
; Output:
;  D0 - 0=ok, else error
AmpOut_Toccata_StartInt:
	Push	d1-a6
	move.l	Toccata_Base(pc),d0
	beq	fail$

	clr.l	tmpvol
	clr.w	tmpvol2

;-------------------- Puffer l?schen -------------------------------------
        lea     TRawBuffer1,a0
        move.w  #[ENDTRawBuffer-TRawBuffer1]/2-1,d0
clr$
        clr.w   (A0)+
        dbf     d0,clr$

	lea	StartBSSBuffer,a0
	move.w	#[ENDBSSBUFFER-StartBSSBuffer]/2-1,d0
clr2$
	clr.w	(a0)+
	dbf	d0,clr2$

	bsr	AmpOut_Toccata_StartDMA
	bne	fail$

	lea	myupsstruct,a1
	move.w	#UPSB_Adr!UPSB_LEN!UPSB_Per!UPSB_Vol!UPSB_DMACON,ups_flags(a1)
	move.w	#15,UPS_DmaCon(a1)
	clr.w	UPS_Enabled(a1)

	move.w	#64,ups_voice1vol(a1)
	move.w	#64,ups_voice2vol(a1)
	move.w	#0,ups_voice3vol(a1)
	move.w	#0,ups_voice4vol(a1)

	move.l	newmixingperiod(pc),d0
	move.w	d0,ups_voice1per(a1)
	move.w	d0,ups_voice2per(a1)
	clr.w	ups_voice3per(a1)
	clr.w	ups_voice4per(a1)

	move.w	#SamLEN,ups_voice1len(A1)
	move.w	#SamLEN,ups_voice2len(A1)
	clr.w	ups_voice3len(A1)
	clr.w	ups_voice4len(A1)

	lea	myupsstruct,a6
	movem.l	UPSBuffer0(pc),d3-d4
	move.l	d3,ups_voice1adr(A6)
	move.l	d4,ups_voice2adr(A6)
	clr.l	ups_voice3adr(A6)
	clr.l	ups_voice4adr(A6)

	moveq	#0,d0
	Pop	d1-a6
	rts
fail$
	moveq	#-1,d0
	Pop	d1-a6
	rts


;
; "Pause"
; Input:
;  A5 - EP Globals
; Output:
;  D0 - 0=ok, else error
AmpOut_Toccata_StopInt:
	move.l	Toccata_Base(pc),d0
	beq	empty$

	bsr     AmpOut_Toccata_StopDMA
	moveq	#0,d0
empty$
	rts



; User selected a new mixing frequency, apply it
;  
; Input:
;  A5 - Amplifier globals
;  D0 - Mixing Rate
; Output:
;  D0 - 0=ok, else error
AmpOut_Toccata_SetNewMixingPeriod:
	move.l	Toccata_Base(pc),d0
	beq	return$

	tst.b	T_MixPerUpdate(pc)		;the call might recurse, when CalcValues is issued, hence check and return if necessary
	bne.s	return$
	st	T_MixPerUpdate

	;
	push	a0
	move.l	usedmode,d0
	beq.s	notplaying$
	move.l	d0,A0
	move.l	(a0),d0

	bsr	AmpOut_Toccata_CalcnewmixingPeriod
	pop	a0

	tst.b	T_Started
	beq.s	notplaying$

	bsr	AmpOut_Toccata_StopDMA
	bsr	AmpOut_Toccata_StartDMA

notplaying$
	sf	T_MixPerUpdate
return$
	rts


; Recalculate mixing parameters, find best frequency first
;  
; Input:
;  A5 - Amplifier globals
;  D0 - Mixing Rate
; Output:
;  D0 - 0=ok, else error
AmpOut_Toccata_CalcnewmixingPeriod:
	push	d1-a6

	;-------------------------------
	bsr	Ampout_Toccata_FindFrequency
	move.l	d0,playmixrate
	bne.s	sanefreq$
	move.l	#44100,d0			;no frequency found (possibly no soundcard present), use a default to avoid division by zero
sanefreq$
	move.l	#14187580/4,clock		;Clock constant (reference clock constant is always the PAL constant)

	ifne	1
	move.l	d0,d7
	
        move.l  clock(pc),d2
	move.l	d2,d0
        move.l  d7,d1                           ;mixingrate(pc),d1
        divu    d1,d0
        move.l  d2,d1
        divu    d0,d1
        swap    d1
        clr     d1
        swap    d1
        move.l  d1,mrate                        ;Abspielrate f?r Mixroutine setzen
	else
	move.l	d0,mrate			;no recalculation of mixing rate (used in "mix,mix2")
	move.l	d0,d1
	endc

;	move.l	AudioRate(pc),d2		;3546895 for PAL
	move.l	clock(pc),d2			;always PAL reference clock

	move.l	d2,d0
	lsl.l	#2,d0
	divu	d1,d0
	move.w	d0,mixingperiod4

	move.l	d2,d0
	divu	d1,d0
	swap	d0
	clr	d0
	swap	d0
	move.l	d0,newmixingperiod

	move.l	d0,d4
	move.l	#262144,d1		;no change in scaling for PAL
	mulu.l	d1,d4			;mixingperiod * 4 * 65536 -> same scaling as sample period plus room for fractional part
	move.l	d4,referenceperiod	;reference period for mixing step calculations (independent of underlying PAL/NTSC hardware)

	Push	a1
	lea	myupsstruct,a1
	move.w	d0,ups_voice1per(a1)
	move.w	d0,ups_voice2per(a1)
	move.w	d0,ups_voice3per(a1)
	move.w	d0,ups_voice4per(a1)
	Pop	a1

	bsr.w	ResetASPeriods			;make sure, new interpolation factors are calculated

	pop	d1-a6
	move.l	mrate,d0
	rts


;============ Puffer l?schen und DMA starten =============================
; Caution: start/stop DMA will be called while playing if the user selects a new mixrate
AmpOut_Toccata_StartDMA:
	push	all
	lea	datas,a5

	bsr	CalcValues	;TODO: find frequencys
	clr.l	tmpvol
	clr.w	tmpvol2
	clr.l	calcvol
	clr.l	calcvol+4
	move.b	#2,waitstart	;delay the actual start (if set >0)
	move.l	newmixingperiod(pc),mixingperiod

        lea     TRawBuffer1,a0   ;dieser Puffer wird zuerst abgespielt und kann beim ersten Int neu gef?llt werden
        move.l  a0,FillBuffer1
        lea     TRawBuffer2,a0
        move.l  a0,FillBuffer2

	move.l	playmixrate(pc),d0
	move.l	d0,Toc_PlayFreq

	move.l  my2ndsignal(a5),Toc_PlayMask
	move.l  mytask(a5),Toc_PlayTask

        lea     Toc_PlayTags(pc),a0
        TB
        CallSys T_RawPlayBack
	st	T_Started
        tst.l   d0
        bne     ok$

        lea     msg_Starterror,a0               ;Infotext ausgeben (Requester)
        bsr     Merror

	bsr	AmpOut_Toccata_StopDMA
ok$
	pop	all
	moveq	#0,d0
	rts

AmpOut_Toccata_StopDMA:
	push	all
	lea	datas,a5
	sf	T_Started

        moveq   #0,d0
        TB
        move.l  a6,d1
        beq     notoc$
        CallSys T_Stop
notoc$
	pop	all
	rts

;----------------------------------------------------------------------------------------------------
;--------------------------- setzt Lautstaerke, Laenge, Period neu ------------------------------------
;----------------------------------------------------------------------------------------------------
; TODO: call only when necessary 
AmpOut_Toccata_SetHardWare:
	ifne	userprogram
		move.l	MyEPBase(pc),a5
		move.l	dtg_SndVol(A5),d0	; dtg_SndVol & dtg_SndLBal
		move.w	dtg_SndRBal(A5),d1	; dtg_SndVol & dtg_SndLBal
	
		cmp.w	tmpvol2(pc),d1
		bne.s	newvol$
		cmp.l	tmpvol(pc),d0
		beq.s	weiter$
newvol$
		move.l	d0,tmpvol
		move.w	d1,tmpvol2

		push	a1			;signal main task to recalculate volume/balance
		move.l	my2ndsignal+datas,d1
		moveq	#0,d0
		bset	d1,d0
		move.l	mytask+datas,a1
		move.l	4.w,a6
		jsr	_LVOSignal(a6)
		pop	a1
weiter$		
	endc
		rts

;----------------------------------------------------------------------------------------------------
;------------------------------- Mixroutine Main Loop -----------------------------------------------
;----------------------------------------------------------------------------------------------------
; main call from secondary interrupt (software-int to stay system-friendly)
AmpOut_Toccata_DoMix:
	movem.l	d0-a6,-(Sp)

	bsr	AmpOut_Toccata_SetHardware

        move.l  FillBuffer1(pc),d0              ;double Buffering handling
        move.l  FillBuffer2(pc),FillBuffer1
        move.l  d0,FillBuffer2                  ;in FillBuffer2 steht der genau jetzt zu f?llende Puffer

	move.b	waitstart(pc),d0
	beq.s	.goon
	subq.b	#1,d0
	move.b	d0,waitstart
	bra.w	.end
.goon:
	move.l	myaudiostruct(pc),a4

	;--------------------- mix left panned channels ----------------------
	move	numstructs,d7
	subq.w	#1,d7
	blt.w	.noleft

	move.l	mixad1,a1				; copy mix routine
;	lea	fm_mix,a1
	clr.w	.curchannelsl

.left_loop
		tst.w	AS_LeftRight(a4)
		ble.s	.left_noleft

			lea	TempBuffer,a2			; 16 Bit output buffer (temporary before scaling)
			move.w	#SamLEN*2,d0			;
			move.l	Vtabaddr(pc),a6

			movem.l	a1/a4/d7,-(sp)			; no registers saved by mixer, store needed pointers
			jsr	(a1)				; mix
			movem.l	(sp)+,a1/a4/d7
		
			tst.w	d0
			beq	.left_nochannel			;we expect that moveq is the last command

			add.w	#1,.curchannelsl
			move.l	mixad2,a1
.left_nochannel:
.left_noleft:
		lea	AS_Sizeof(a4),a4
		dbf	d7,.left_loop

		; if no channel was active, Tempbuffer will be "dirty", use pointer to 0-buffer in these cases
		lea	Tempbuffer,a0
		move	.curchannelsl(pc),d0
		bne.s	.leftok
		lea	ZeroSamplebuffer,a0
.leftok:	move.l	a0,.lbuffer
.noleft:

	;--------------------- mix right panned channels ----------------------
	move.l	myaudiostruct(pc),a4

	move	numstructs,d7
	subq.w	#1,d7
	blt.w	.noright
	
	move.l	mixad1,a1	;	
;	lea	fm_mix,a1
	clr.w	.curchannelsr

.right_loop
		tst.w	AS_LeftRight(a4)
		bge.s	.right_noright

			lea	TempBuffer2,a2			; 16 Bit output buffer (temporary before scaling)
			move.w	#SamLEN*2,d0			;
			move.l	VtabaddrR(pc),a6
			movem.l	a1/a4/d7,-(sp)			; no registers saved by mixer, store needed pointers
			jsr	(a1)				; mix
			movem.l	(sp)+,a1/a4/d7
			tst.w	d0
			beq	.right_nochannel		;we expect that moveq is the last command
			
			add.w	#1,.curchannelsr
			move.l	mixad2,a1	;	
.right_nochannel:
.right_noright:
		lea	AS_Sizeof(a4),a4
		dbf	d7,.right_loop

		; if no channel was active, Tempbuffer will be "dirty", use pointer to 0-buffer in these cases
		lea	Tempbuffer2,a0
		move	.curchannelsr(pc),d0
		bne.s	.rightok1
		lea	ZeroSamplebuffer,a0
.rightok1:	move.l	a0,.rbuffer
.noright:

	;--------------------- perform mono/surround, copy to ChipRAM  ----------------------
	move.l	.lbuffer(pc),a0			; 16 Bit mixing buffer (before scaling or zero buffer)
	move.l	.rbuffer(pc),a2			; 16 Bit mixing buffer (before scaling or zero buffer)

	move.l	FillBuffer2,a1			;main output buffer
	move.l	UPSBuffer0,a5			;main UPS buffer
	move.l	UPSBuffer1,a4			;main UPS buffer

	move.l	cbufad,a6
	tst.w	.curchannelsl
	bne.s	.left_notempty
	tst.w	.curchannelsr
	bne.s	.right_notempty
	move.l	cbufempty,a6
	bra.s	.left_nofilter
.right_notempty:
.left_notempty:
	; highboost left/right (pointless with a good sound card, TODO: disable last entry in cycle gadget)
	move.l	boostad(pc),d0
	beq.s	.left_nofilter
	push	all
	move.l	d0,a5
	lea	HB_DelayL,a1
	move	#SamLEN*2,d7
	push	a2/a5
	jsr	(a5)	
	pop	a0/a5
	lea	HB_DelayR,a1
	move	#SamLEN*2,d7
	jsr	(a5)				
	pop	all

.left_nofilter:
	lea	TBufLeft,a3
	move	prefetchL(pc),d0
	jsr	(a6)

	ifne	0
	move.l	FillBuffer2,a1			;main output buffer
	moveq	#128-1,d0
.lo1
	move.l	#$7fff8000,(a1)+
	move.l	#$80007fff,(a1)+
	dbf	d0,.lo1
	endc

	;-------------- Copy Audio Structure back to Player, if necessary -------------------------
	bsr	CopyASBack
.end
	movem.l	(sp)+,d0-a6
;	moveq	#0,d0
	rts
	
	; active left/right buffers, either Tempbuffer or Zerobuffer
.lbuffer	dc.l	0
.rbuffer	dc.l	0
	; num active channels left/right
.curchannelsl:	dc.w	0
.curchannelsr:	dc.w	0

;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
;-------------------------------------- Toccata Code End ---------------------------------------------
;-----------------------------------------------------------------------------------------------------

;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
;--------------------------------- 16 Bit Output Routines  -------------------------------------------
;-----------------------------------------------------------------------------------------------------
; Concept:   collect a variable number of channels in the mixing routines (unsigned), and scale it to
;            the relevant 14 bit dynamic range afterwards
; Workflow:

; There are several routines for this phase:
;  1. empty copy   - just zeroes the chipram buffer when no mixing was done
;  2. stereo copy  - fill one chipram buffer pair per channel
;  3. surround copy- apply the delayed opposite channel to the current channel
;                  - branch to stereo copy
;
; Input:
; Output:
;         -
;
; Used Registers: all, none saved
; 
; Notes/Caveats: 
;        number of samples still hardcoded (SamLEN*2)
;-----------------------------------------------------------------------------------------------------



;---------------------- no channel playing, write silence to two channels --
; Inputs:
;  D0 - prefetch
;  A0 - input samples left channel, 16 bit signed
;  A1 - output samples 16 bit, interleaved 
;  A2 - input samples right channel, 16 bit signed
;  A3 - temp buffer for delayed samples in mono/surround
;  A4 - UPS buffer in fastram (right)
;  A5 - UPS buffer in fastram (left)
copybufemptyStereoT:
	ifne	benchmark
	move	#$3ff,$dff180
	endc

	move	#SamLEN*2/16-1,d7
	moveq.l	#0,d1
	moveq.l	#0,d2
	moveq.l	#0,d3
	moveq.l	#0,d4

	add.l	#SamLEN*2*4,A1	;Samlen*2 samples, *32 bit per entry
	add.l	#SamLEN*2,A4
	add.l	#SamLEN*2,A5
.ldd3
	movem.l	d1-d4,-(a1)	;main samples
	movem.l	d1-d4,-(a1)	;main samples
	movem.l	d1-d4,-(a1)	;main samples
	movem.l	d1-d4,-(a1)	;main samples

	movem.l	d1-d4,-(a4)	;UPS buffer
	movem.l	d1-d4,-(a5)	;UPS buffer
	dbf	d7,.ldd3

	ifne	benchmark
	move	#$888,$dff180
	endc
	rts

;---------------------- fuer Stereo 14 Bit ---------------------
; Inputs (*=used here):
;  D0 - prefetch
;  A0 - input samples left channel, 16 bit signed
;  A1 - output samples 16 bit, interleaved 
;  A2 - input samples right channel, 16 bit signed
;  A3 - temp buffer for delayed samples in mono/surround
;  A4 - UPS buffer in fastram (right)
;  A5 - UPS buffer in fastram (left)
copybufStereoT:		;for 68020-68060, UPS Struktur in fast ram
	ifne	benchmark
	move	#$0ff,$dff180
	endc

	move.l  buff14_2(pc),a6 	;middle of the boost table

	move	#SamLEN*2/2-1,d7	;mixed samples=SamLEN*2

.ldd3
        movem.w	(a0)+,d0/d1		;left
	movem.w	(a2)+,d2/d3		;right
        move    (a6,d0.l*2),d0		;l0
        move    (a6,d1.l*2),d1		;l1
	move	d0,d4		;UPS L0.w
	move	d1,d5		;UPS L1.w
	swap	d0
	move    (a6,d2.l*2),d2		;r0
	lsr.w	#8,d5		;UPS L1.b
	swap	d1
	move.b	d5,d4		;UPS L0.b L1.b
	move    (a6,d3.l*2),d3		;r1
	move.w	d4,(a5)+	;UPS L
	move	d2,d0			;l0 r0
	move	d3,d1			;l1 r1
	lsr.w	#8,d3		;l1.b
	move.l	d0,(a1)+
	move.b	d3,d2		;l0 l1
	move.l	d1,(a1)+
	move	d2,(a4)+	;UPS R
        dbf     d7,.ldd3

	ifne	benchmark
	move	#$0,$dff180
	endc

	rts

;---------------------- Surround 16 Bit ---------------------
; Inputs:
;  D0 - prefetch (ignored here, variables read directly)
;  A0 - input samples left channel, 16 bit signed
;  A1 - output samples 16 bit, interleaved 
;  A2 - input samples right channel, 16 bit signed
;  A3 - temp buffer for delayed samples in mono/surround
;  A4 - UPS buffer in fastram (right)
;  A5 - UPS buffer in fastram (left)
;
; Notes: empty copy routine disabled in surround mode because of other channel possibly
;        contributing to sum channel
; - prefetchL, prefetchR accessed directly in here
; - TBufLeft, TBufRight accessed directly in here
; Note:

; todo: change order of first / second word for better scheduling of UPS operation
surroundT	macro
	move	d6,d1		;volume multiplier
	move	d6,d2		;volume multiplier
	muls	(a6)+,d1	;other channel data * volume
	muls	(a6)+,d2	;
	addq.l	#8,a3		;next output position
	asr.l	#6,d1		;divide by 64
	asr.l	#6,d2		;divide by 64
	add.w	(a0),d1		;current channel input
	add.w	2(a0),d2	;current channel input
	addq.l	#4,a0		;input forward
	move    (a4,d1.w*2),(a3) ;with boost
	;move	d1,(a3)		;no boost
	move    (a4,d2.w*2),4(a3) ;with boost
	;move	d2,4(a3)	;no boost

		endm

ups_T	macro
	lsr.w	#8,d2
	move.b	d2,d1
	move	d1,(\1)+
	endm

copybufSurroundT:	;68040/68060 targeted routine (multiplies instead of table accesses)
	ifne	benchmark
	move	#$4f4,$dff180
	endc


	move	#SamLEN,d7	;mixed samples=SamLEN*2, two operations in each run, hence no * 2 here
	push	a0		;input left (see below pop a6)
	push	a4		;store UPS pointer for right channel (gets back to a5 later)

	move.l  buff14_2(pc),a4 	;middle of the boost table

	move	effectvol(pc),d6	; effect volume 0..64
	neg.w	d6			; negative: we want inverse additions

	;------------- left channel --------------------------
	lea	TBufRight,a6	;Temp buffer, stored from last run
	lea.l	-8(a1),a3	;ignore temp buffer in A3 -> load output array to A3 -8 (pre-increment in loop)

	move	prefetchL(pc),d0
	move	d7,d5
	sub.w	d0,d5
	subq	#1,d0
prefetchL1$
	surroundT		;modified d1,d2,a0,a3,a6, used d6
	ups_T	a5

	dbf	d0,prefetchL1$

	;fill rest with data from current right channel
	subq.w	#1,d5
	lea	(a2),a6		;current right channel
mainsurroundL1$
	surroundT		;modified d1,d2,a0,a3,a6, used d6
	ups_T	a5
	dbf	d5,mainsurroundL1$

	; post-processing: copy data to temp buffer for next run
	lea	TBufRight,a3	;Temp buffer, stored from last run
	move	prefetchL(pc),d0
	subq	#1,d0
left_post$
	move.l	(a6)+,(a3)+
	dbf	d0,left_post$

	pop	a5		;restore UPS pointer for right channel (from push a4)
	;------------- right channel --------------------------
	lea	TBufLeft,a6	;Temp buffer, stored from last run
	lea	-6(a1),a3	;output array right channel to A3 (-8+2 = first byte right channel -8)
	move.l	a2,a0		;input right channel
	move	prefetchR(pc),d0
	move	d7,d5
	sub.w	d0,d5
	subq	#1,d0
prefetchR1$
	surroundT		;modified d1,d2,a0,a3,a6, used d6
	ups_T	a5
	dbf	d0,prefetchR1$

	;fill rest with data from current left channel
	pop	a6		;input left (from push a0)
	subq.w	#1,d5
mainsurroundR1$
	surroundT		;modified d1,d2,a0,a3,a6, used d6
	ups_T	a5
	dbf	d5,mainsurroundR1$

	; post-processing: copy data to temp buffer for next run
	lea	TBufLeft,a3	;Temp buffer, stored from last run
	move	prefetchR(pc),d0
	subq	#1,d0
right_post$
	move.l	(a6)+,(a3)+
	dbf	d0,right_post$

	rts

;---------------------- Mono 16 Bit ---------------------
; Inputs:
;  D0 - prefetch
;  A0 - input samples main channel (left/right)
;  A1 - output samples high volume channel
;  A2 - input samples other channel (for mono/surround)
;  A3 - temp buffer for delayed samples in mono/surround
;  A4 - output samples low volume channel
;  A5 - UPS buffer in fastram, same contents as high volume channel
;
; Notes: - empty copy routine disabled in surround mode because of other channel possibly
;          contributing to sum channel
;        - slow...
;        - UPS output does not respect boost
copybufMonoT:
	ifne	benchmark
	move	#$0ff,$dff180
	endc

	move	effectvol(pc),d6	; effect volume 0..64

	move.l  buff14_2(pc),a6 	;middle of the boost table

	move	#SamLEN*2-1,d7		;mixed samples=SamLEN*2
.ldd3
        move.w	(a0)+,d0		;left
	move	d6,d4
	move	d6,d5
	move.w	(a2)+,d2		;right
	muls	d0,d4			;left*effect
	muls	d2,d5			;right*effect
	asr.l	#6,d4			;left*effect/64
	asr.l	#6,d5			;right*effect/64
	add	d5,d0
	add	d4,d2
        move    (a6,d0.w*2),(a1)+	;volume boost table (nonlinear)
	move    (a6,d2.w*2),(a1)+	;volume boost table (nonlinear)
	lsr.w	#8,d0
	lsr.w	#8,d2
	move.b	d0,(a5)+
	move.b	d2,(a4)+
        dbf     d7,.ldd3

	ifne	benchmark
	move	#$0,$dff180
	endc
	rts

	ifne	Toccata_SECONDINT
AmpOut_Toccata_Int1:
		push	all
		lea	Toc_SecondInt(pc),a1
		move.l	4.w,a6
		jsr	-180(a6)	;_LVOcause
		pop	all
		rts
	endc


;----------------- Toccata HW Data ------------------------------------------------------------------------

Toccata_base:	dc.l	0
toccata_freqs:	ds.l	26
toccata_nfreqs:	dc.l	0
FillBuffer1:	dc.l	0
FillBuffer2:	dc.l	0
T_Started:	dc.b	0
T_MixPerUpdate:	dc.b	0
T_unusedbytes	dc.b	0,0

TRawSize         =      SamLEN*8 ;*2 words *2 channels *2 (as usual for SamLEN)

toccata_name:		dc.b	`toccata.library`,0
TCallBackIntName1:	dc.b	`Toccata Amplifier Interrupt`,0
msg_Starterror          dc.b    'Unable to start playing.',10
                        dc.b    'Please check your library version',10
                        dc.b    'and your Toccata, Melody or Prelude ',10
                        dc.b    'soundcard if properly installed.',10
                        dc.b    'Make sure that no other Audio software',10
                        dc.b    'is already active.',10
                        dc.b    0

msg_ToccataHW           dc.b    'no Hardware found, please check your',10
                        dc.b    'Library and your Toccata, Melody or ',10
                        dc.b    'Prelude soundcard if properly installed.',10
                        dc.b    0
msg_ToccataLib          dc.b    'toccata.library not found, please make',10
                        dc.b    'sure that the library is in Libs:',10
                        dc.b    0

		cnop	0,4

Toc_PlayTags:
        dc.l    TT_Mode,TMODE_LINEAR_16_S
        dc.l    TT_Frequency
Toc_PlayFreq:                               ;wird vor dem Start richtig gesetzt
        dc.l    44100

        dc.l    TT_RawInt
        dc.l    Toc_CallBackInt

        dc.l    TT_RawBuffer1,TRawBuffer1	;actual buffers are in BSS section
        dc.l    TT_RawBuffer2,TRawBuffer2
        dc.l    TT_BufferSize,TRawSize		;TRawSize		;2048
        dc.l    TT_RawIRQSize,TRawSize/4	;512

        dc.l    TT_ErrorTask
Toc_PlayTask:
        dc.l    0

        dc.l    TT_ErrorMask
Toc_PlayMask:
        dc.l    0

        dc.l    TAG_DONE

Toc_CallBackInt:
        dc.l    0
        dc.l    0
        dc.b    NT_Interrupt
        dc.b    32                      ;32
        dc.l    TCallBackIntName1       ;Name
        dc.l    0
        ifne	Toccata_SECONDINT
		dc.l	AmpOut_Toccata_Int1
Toc_SecondInt:
	        dc.l    0
	        dc.l    0
	        dc.b    NT_Interrupt
	        dc.b    31                       ;pri
	        dc.l    TCallBackIntName1        ;Name
	        dc.l    0
        endc
	dc.l	AmpOut_Toccata_DoMix

