*****************************************************************************
** Engine: 14 Bit Amplifier II                                             **
**                                                                         **
** Project: Eagleplayer 2.05                                               **
** Authors: Jan Blumenthal & Henryk Richter                                **
** Start  : 1993/01/09                                                     **
** $Rev: 41 $ $Author: buggs $                                             **
** $Date: 2016-10-14 20:36:13 +0200 (Fr, 14. Okt 2016) $                   **
**                                                                         **
*****************************************************************************
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program (See the included file COPYING);
** if not, write to the Free Software Foundation, Inc.,
** 675 Mass Ave, Cambridge, MA 02139, USA.
**
**************************************************************************** 
; TODO:
;  - linear interpolation (oversampling) as alternative to sinc interpolator
;  - (rewrite to use 32 bit internal representation)
;  - (disable filtered mixing options on 68030 and below)
;  - (re-introduce table based surround for 68030 and below)
;  - mix16 for more than two channels
;  - mix16 for 68040/60
;  - store volume multiplier globally, get rid of volume table (caution: unsigned 8 Bit)
;  - downsampled filtered mixing for 8 Bit inputs
; -> AHIAMP
;    - write FAQ
; -> Toccata
;    - write FAQ
; locale (currently disabled)

; umgebaut auf neue AS_Struktur am 19.07.96
; 2.8.: Iconify Flag bei "Show" ausgetestet
; 6./7.8. - fix fuer Check, ob noch 8 Bytes zu mixen sind, addx eingebaut
;         - Unterst?tzung von EPAMF_WaitForStruct
; 23.4.98 - R?ckkopieren der AS Struktur wegen Scopes
; xx.5.98 - UnlockPubscreen fix
; 12.6.16 - first stage mixing routine rewritten
; 13.6.16 - fixed first stage mixing
;         - enabled 68020 support for faster copy buffer routines (there are 68020 cards with fastmem available)
;         - first stage mixing is faster now, mixing routines rewritten and cleaned up
;         - About requester includes some useful notes/hints now instead of an obsolete address
; 14.6.16 - removed redundant memory writes when channels are not playing, slight speed gains
;           without any additional burden, actually simplifies code in main DoMix routine
;         - added fast handling of 8 bit signed with max. volume (64)
; 15.6.16 - rewritten copybuffer routine for 68000, now 23.5 cycles/byte instead of 40
;         - 7 MHz 68000 works now (somewhat), with 16 kHz, 4 channel mixing is possible (still pointless, of course)
; 16.6.16 - UPS support fixed to be more responsive in FFT Analyzer and Scopes, even at 28 kHz
;         - Version bump to 10.01, reason: $VER was interpreted as 8.0 due to the version string layout,
;           the fix required something larger as version number
; 30.6.16 - fork 8 Bit Amplifier into (new) 14 Bit Amplifier
; 02.7.16 - Stereo 14 Bit working, with Boost
; 04.7.16 - Surround/Mono working, Balance working
; 06.7.16 - faster surround, still not satisfactory (needs twice the raster time compared to "unoptimized" stereo copy)
;         - increased delay buffers
; 07.7.16 - surround code rewritten, now just a little more demanding than pure stereo (overhead: about one mixed 8 bit channel)
; 17.7.16 - filtered mixing experiments (currently static choice in setplaymode)
;           - polyphase windowed sinc filter bank, 4 tap zero-phase filtering
; 19.7.16 - implemented selection between filtering modes
;         - "fadein" option (assemble-time, for now) which delays each new instrument by exactly 1 output sample,
;           -> "anti-click" approach (with correct phase)
;         - "fadeout" for one-shot instruments
; 25.7.16 - ported filtered mixing to 16 bit
;         - experimenting with higher polyphase resolution (1/32 instead of 1/16)
; 29.7.16 - first steps to implement high-boost to compensate for the amiga lowpass filter
;         - higher accuracy in mixing rate calculations (68020+)
;         - temporarily switched to Audio Interupt 3 (testing, trying to find out why FS-UAE does not like me)
; 30.7.16 - fixed race condition in fast mixing loop limit calculations
;         - changed high boost filter to 0dB midpoint at 11 kHz to keep higher dynamic range in mixing routines
; 03.8.16 - fixed UPS in mono mode
; 10.8.16 - finished Toccata integration, some cleanup still pending but working nontheless
; 14.8.16 - fixed routine for downsampled filtered mixing (16 Bit)
;         - activated 28kHz targeted highboost filter
; 17.8.16 - AHI support, first alpha
; 19.8.16 - AHI improvements, mode selection
;         - reduced fadeout which produced clicks -> investigate further (!)
; 22.8.16 - AHI modes included in main window
;         - error requester if AHI could not be opened
; 26.8.16 - Anti-Click filter (some tuning still pending)
; 28.8.16 - removed SurroundTab, rewrote copybufmono for 14 Bit Paula
; 10.9.16 - changed highboost 14 bit table extra boost to match 28/44 kHz volumes (Paula)
; 14.10.16- fixed default mode naming for AHI
;

	;
	;
	incdir	ep:include/
	include	exec/exec_lib.i
	include	exec/execbase.i
	include	exec/libraries.i
	include	exec/lists.i
	include	exec/memory.i
	include	exec/types.i
 	include	intuition/intuition.i
	include	intuition/intuition_LIB.i
	include	graphics/graphics_lib.i
	include	libraries/gadtools.i
	include	libraries/gadtools_lib.i
	include	dos/dos_lib.i
	include	libraries/dos.i
	include	misc/eagleplayer.i
	include	misc/eagleplayerengine.i
	include	misc/BuggsMacros.i
	include	misc/MixMacros.i
	include newmouse.i

	;
	section	0,code_p
	;
	; jump from version 0.9x to 10 -> $VER reported "8" with earlier version string, the fix requires something bigger
ver		macro
		dc.b	"10.09"
		endm
StructVer	macro
		dc.w	10,09
		endm
date		macro
		dc.b	"14-Oct-16"
		endm
StructDate	macro
		dc.b	14,10
		dc.w	2016
		endm

userprogram	=	1
debug		=	0
mixdebug	=	0	;debug mixing routine


	; output hardware selection -> currently just one enabled (!) - dynamic selection not yet implemented (UI issues)
PAULA14		EQU	0
TOCCATA		EQU	0
AHIOUT		EQU	1


test		=	0	;test without Eagleplayer
benchmark	=	0	;Benchmark = 1 => Vergleich zwischen alter wortweiser und neuer 
				;                 langwortweiser CopyBuffer Routine
benchmark2	=	0	; compare with old routines...

SamLEN		=	256	;should be a multiple of 16 (!), in 16 bit words (especially good for 45.1 kHz sampling rate)
fadein		=	1	;0/1: SincMix fade in by 1 sample -> same delay for all played samples (exactly 1 sample, 
				;regardless of sampling rate -> so no artificial phase offsets)
clickreduce	=	1	;0/1: SincMix reduce clicking by lowpass to the first loaded samples, if enabled
				;by "1", the fadein directive doesn't have an effect anymore
anticlick	=	1	;improved click reduction
fmix_accuracy	=	1 	;0=1/16 sample polyphase, 1=1/32 sample polyphase
V37Gadtools	=	0	;test: pretend to have V37 gadtools, even on Kick3.x to verify the Kick 2.x code path
NumChipBuffers	=	2	;Anzahl der Puffer im Chipram (4 statt 2 wg. FFT Analyzer)
PrefsVer	=	3	;increment this variable each time you change the preferences format (!)

Mixfreqlist	=	1	;use supplied list of frequencies instead of free slider
fastmix		=	1	;debug: compare old add,addx combination (0) against new approach (1), available in mix16
HighBoost28_Maxrate =	35000	;use 44 kHz highboost filter above 35kHz, else use 28 kHz filter
MaxVoices	EQU	32	;max. 32 voices supported (TODO: verify >32 but observe the code..., also don't forget prefsver)

	; anti-click definitions
AC_SHIFT	=	6	;time step resolution (max. 8, I use immediate shifts)
AC_REF		=	(1<<AC_SHIFT) ; reference value for multiplicators y[n] = y[n-1]*cur - (cur-AC_REF)*x[n]
AC_INITIAL	=	64	;anything >0 and <AC_REF, values AC_REF/2....4*AC_REF/5 should be interesting
AC_DELAY	=	1	;y[n-AC_DELAY]: sample distance for IIR (1...FM_TAPS)

nosurroundtab	=	1	;disable table based surround (no longer used, 68030 routine currently disabled for Paula)

;----------------------- Amplifier specific information --------------------

;extra window width in AHI mode to accomodate additional information in listview
	ifne	AHIOUT
XWidth	equ	48
	else
XWidth	equ	0
	endc


Project0Width = 309+XWidth		;Windowbreite (default bei Topaz 8)
Project0Height = 158+18
Project0Left = 41
Project0Top = 45
ZoomWidth = 200

GD_LISTVIEW             EQU	0
GD_Samplefreq           EQU	1
GD_MixingSlider         EQU	2
GD_VBoost               EQU	3
GD_Voices               EQU	4
GD_NumVBoost            EQU	5
GD_Playmode             EQU	6
GD_Surlevel             EQU	7
GD_Mixmode		EQU	8
GD_AHIMode		EQU	9
GD_AHIGlobal		EQU	10

GD_Disable		EQU	1

; Sample Formats (internal)
FMT_S8		EQU	1
FMT_U8		EQU	2
FMT_S16		EQU	3

CallSys	macro
	jsr	_LVO\1(a6)
	endm



Project0_CNT    EQU    11		;number of gadgets in the main window

; this list structure is used for both the Listview and the list of AHI modes (with names copied to listview)
ML_NameLength	EQU	96
 STRUCTURE	my_liststruct,0
	STRUCT	ML_Node,LN_SIZE
	Long	ML_AHIMode			;
	STRUCT	ML_Nodename,ML_NameLength	;room for settings + name
	LABEL	ML_Sizeof			;caution with size changes, sometimes ML_Sizeof is referenced by moveq

; TODO: finish inclusion here, not all functions are verified to actually use the structure as intented,
;       keep P_Mixrate as first entry for safety...
 structure	Prefsblock,0
	Long	P_Mixrate
	Word	P_Vboost
	Byte	P_Surround		;Mono/Surround
	Byte	P_SurroundDelay		;Delay
	Word	P_Flags			;(Bit0=disable)
	Long	P_AHIMode		;AHI Mode ID
	Label	P_Sizeof

PBF_Disable	EQU	0
PBF_SincMix	EQU	1		 ;use linear interpolation
PBF_SincMixB	EQU	2		 ;use sinc interpolation
PBB_Disable	EQU	1<<PBF_Disable
PBB_SincMix	EQU	1<<PBF_SincMix 	 ;use sinc interpolation
PBB_SincMixB	EQU	1<<PBF_SincMixB	 ;use sinc interpolation with HighBoost
PBB_MixMask	EQU	PBB_SincMix+PBB_SincMixB
PBB_MixMax	EQU	2		; current modes: 0-2

 structure	my_Prefs,0
	struct	P_Blocks,P_Sizeof*MaxVoices	;max. Stimmen
	Label	my_Prefs_Sizeof

	;/* FPOS+2: flags */
ASLoopF_BackLooping	EQU	0	;Ping-Pong: currently looping backwards
ASLoopF_Looping		EQU	1	;Ping-Pong: currently looping forwards
ASLoopF_FMInit		EQU	2	;Filtered Mixing initialized
ASLoopB_BackLooping	EQU	1<<ASLoopF_BackLooping
ASLoopB_Looping		EQU	1<<ASLoopF_Looping
ASLoopB_FMInit		EQU	1<<ASLoopF_FMInit



; precomputed periods (saves divisions and other computations), anticlick position
ASP_LastPeriod		EQU	AS_AmplifierPrivate1	; word size
ASP_Anticlickpos	EQU	AS_AmplifierPrivate1+2	; anticlick position (0=no anticlick)
ASP_LastInterpolate	EQU	AS_AmplifierPrivate2	; last interpolation constant

BUFCHECK		EQU	0


;
; Filtered Mixing definitions: 
;  - apply 4 tap polyphase windowed sinc filter (Lanczos-2 window in the first iteration)
FM_Preload		EQU	3			;load 3 samples at beginning/repeat of each instrument
FMB_Taps		EQU	2
FM_Taps			EQU	1<<FMB_Taps		;filter is 4 taps for now, adjust AS_FMSamples buffers in case of changes,
							;currently assumes 2^n as filter length, also adjust mixing macros
	ifne	fmix_accuracy			 	;0=1/16 sample polyphase, 1=1/32 sample polyphase
FMB_Accuracy		EQU	5
	else
FMB_Accuracy		EQU	4
	endc
FM_Accuracy		EQU	1<<FMB_Accuracy		;accuracy (2^n) for the polyphase coefficients

AS_FMPos		EQU	AS_AmplifierPrivate3	;1 Byte position in buffer (dictates write position and proper interpolation table)
AS_FMUnused1		EQU	AS_AmplifierPrivate3+1	;
AS_FMVol		EQU	AS_AmplifierPrivate3+2	;
AS_FMUnused3		EQU	AS_AmplifierPrivate3+3	;
AS_FMSamples1		EQU	AS_AmplifierPrivate4	;2 samples in 16 bit (already volume corrected)
AS_FMSamples2		EQU	AS_AmplifierPrivate5	;2 samples in 16 bit



MIXSL_SliderUp		EQU	0	;slider knob released (d5)
MIXSL_SliderDown	EQU	1	;slider knob down (d5)
MIXSL_ExplicitFreq	EQU	2	;explicit frequency in Hz (d5)
MIXSL_ListSize		EQU	3	;number of available mixing frequencies in list (d5)
MIXSL_SetLevel		EQU	4	;get current value from prefs and adjust slider level (no argument)

;--------------- Output hardware jump list ---------------------------------
		rsreset
OHW_Init	rs.l	1	;Initialize after replayer
OHW_Shutdown	rs.l	1	;Shutdown at the end of the module
OHW_StartInt	rs.l	1	;Start mixing (start/stop will be also called with "pause")
OHW_StopInt	rs.l	1	;Stop mixing
OHW_Check	rs.l	1	;check whether this output is available
OHW_GetMixRates	rs.l	1	;available mixing rates as long int array
OHW_SetMixRate	rs.l	1	;user selected new mixing rate, apply
OHW_SetPlayMode	rs.l	1
OHW_SetBoost	rs.l	1
OHW_Setup	rs.l	1	;called before any other functions: setup private structures/libraries etc.
OHW_Exit	rs.l	1	;called on quit: free resources
OHW_Reserved	rs.l	1	;for future additions
OHW_SizeOf	rs.l	0

	;status flags for output hardware
OHWS_NOINIT	equ	0	;no hardware init (keep this define 0 !)
OHWS_INIT	equ	1	;output initialized
OHWS_ACTIVE	equ	2	;output currently active



;------------------------------------------------------------------------
begin:
	ifne	0
	;movep test
	lea	data,a0
	moveq	#-1,d0
	movep.l	d0,0(a0)
	move.l	#$55667788,d0
	movep.l	d0,1(a0)

	movep.l	0(a0),d0
	movep.l	1(a0),d1
	movep.w	0(a0),d2
	movep.w	1(a0),d3

	illegal
data:	dcb.l	20,0

	endc

	
	ifne	mixdebug

	move.l	4.w,a6
	move.l	ex_EClockFrequency(a6),d0
	move.l	d0,d1
	add.l	d0,d0
	add.l	d0,d0
	add.l	d1,d0
	move.l	d0,AudioRate
	;move.l	d0,clock

	move.l	#14187580/4,clock		;reference Clock constant (PAL)
;	move.l	#28614,d0			;28603,d7
	move.l	#44336,d0			;28603,d7
;	move.l	#55400,d0
	bsr	AmpOut_Paula14_CalcnewmixingPeriod
	move.b	#FMT_S8,fformat	 ; signed 8 Bit
;	move.b	#FMT_S16,fformat ; signed 16 Bit
	lea.l	testtab,a0
	move.l	a0,Vtabaddr
	move.l	a0,VtabaddrR
	bsr	makevoltable

	lea	Extra_AS,A4
	move.w	#-1,AS_LeftRight(a4)	; left
	move.w	#-1,AS_DMABit(A4)
	clr.b	AS_NoLoop(a4)
	move	#$40,AS_LeftVolume(a4)  ;2 b changed 4 panning support

	clr.w	ASP_LastPeriod(A4)
	clr.w	ASP_Anticlickpos(A4)   ; no Anticlick (yet)
	move.w	#321,AS_Period(A4);    ; 321=44100Hz, 124*4,AS_Period(a4) ; 124*8,AS_Period(a4)

	clr.l	AS_Currentadr(a4)	; pos (addr)
	lea	testsample+4,a0
	move.l	#testsamplesize,d0
	move.l	a0,AS_RepeatAdr(A4)
	move.l	d0,AS_RepeatSize(a4)
;	move.b	#ASLoopB_PingPong,AS_LoopFlag(a4)	;4=PingPong
	move.b	#ASLoopB_Forwards,AS_LoopFlag(a4)	;4=PingPong

	ifne	0
	move.w	AS_Period(a4),d0 ; 124*8,AS_Period(a4)
	; mixing offset calculation test
	move.l	clock(pc),d4		;reference clock: PAL timing
	divu	d0,d4			;/period (*4) -> 0.25*mixrate in d4
	swap	d4			;0.25*mixrate * 65536
	clr	d4			;clear lower word
	lsr.l	#2,d4			;mixrate * 16384
	move.l	mrate(pc),d0		;
	divu	d0,d4
	swap	d4
	clr	d4
	rol.l	#4,d4			;
	;
	move.l	newmixingperiod,d4	;
	move.l	#264557,d1		;actual NTSC mixing period -> PAL reference period in AS Structure
	move.l	#3546895,d0
	sub.l	AudioRate(pc),d0
	bne.s	.ntsc
	move.l	#262144,d1		;no change in scaling for PAL
.ntsc
	mulu.l	d1,d4			;mixingperiod * 4 * 65536 -> same scaling as sample period plus room for fractional part

	move.l	referenceperiod(pc),d4
	moveq	#0,d0
	move.w	AS_Period(a4),d0 ; 124*8,AS_Period(a4)
	divu.l	d0,d4			;mixing_period/sample_period*65536
	subq.l	#1,d4			;workaround for fm_mix which requires that the integer step size is zero
	swap	d4			;lower word: byte offset, upper word: fractional part of offset
	endc

.lo
	lea	TempBuffer,a2
	move.l	#SamLEN*2,d0
	move.l	Vtabaddr,A6
	
;	bsr	fm16down_mix
;	bsr	mix16
	bsr	fm_mix
	add.l	#1,.count

	; verify zeroes
	lea	TempBuffer,a2
	move.l	#SamLEN*2-1,d0
.chk
	tst.w	(a2)+
	beq.s	.ok
	move.l	.count,d1
	nop
.ok	
	dbf	d0,.chk


	; scramble data
	lea	TempBuffer,a2
	move.l	#SamLEN*2-1,d0
.scram
	move	d0,d1
	eor.w	#$808f,d1
	move.w	d1,(a2)+
	dbf	d0,.scram


	btst	#6,$bfe001
	bne.s	.lo

	illegal
.count	dc.l	0
	endc


	ifne	0

	move.l	#$F3000000,d0
	add.l	#$03000000,d0
	bcc.s	.ok

	nop
	nop
.ok

	lea	TBufLeft,a2
	lea	TBufLeft,a0
	lea	TBufLeft,a6

	lea	Extra_AS,A4
	move.b	#3,AS_FMPos(a4)
	move.w	#$ffff,AS_FMSamples1(a4)
	move.w	#$7fff,AS_FMSamples1+2(a4)
	move.w	#$8000,AS_FMSamples1+4(a4)
	move.w	#$3fff,AS_FMSamples1+6(a4)



	move.l	#$13000000,d0
	move.l	#$f0000000,d4

	moveq	#0,d3
	move.b	AS_FMPos(a4),d3			;ringbuffer index
	lsl.w	#FMB_Accuracy+FMB_Taps+1,d3	;accuracy bits + taps bits + word size
	lea	FM_tab_100,a3
	add.l	d3,a3		;table for current ringbuffer position

	; Ringbuffer work
	move.l	d0,d3				;upper 16 bit: fractional position
	rol.l	#FMB_Accuracy+FMB_Taps+1,d3	;
	and.l	#$10000-(1<<(FMB_Taps+1)),d3	;currently relevant 4 coefficients

	fm_mix_one	99
	illegal
.str
.fmtab	
	endc

	ifne	0
	; Boost table test
		lea	testtab+$10000,a0
		move.l	a0,a1
		move.w	#32768-1,d0
		moveq	#0,d1
		moveq	#-1,d2
.tst
		move	d1,(a0)+
		move	d2,-(a1)
		addq	#1,d1
		subq	#1,d2
		dbf	d0,.tst

		lea	testtab,a0
		moveq	#15,d0
		bsr	_Boost14Table
	endc

	bra.w	fang
	EUSN_Identifier
	dc.l	0		;eus_next

	ifne	debug
	dc.w	-2
	else
	dc.w	0		;eus_Usernr
	endc
MyEPBase
	dc.l	0		;eus_EPBase
	dc.l	0		;eus_Freetable
	dc.l	0		;eus_Taskadr
	dc.l	0		;eus_unused1
	dc.l	0		;eus_unused2
	dc.l	SpecialJumps	;EUS_SpecialJumpTab
	dc.l	TagListe	;EUS_TagList
	dc.w	0		;eus_ticks
	dc.w	0		;eus_tickcounter
	dc.l	0		;eus_tickflags
	dc.l	USMB_Configuration!USMB_KillModule!USMB_Zipwin!USMB_ChangeConfig;eus_msgflags
	dc.l	MyAnalyzername	;eus_PName
	dc.l	myBuggsname
	dc.l	myinfo
	dc.w	37		;eus_Kickstart
	dc.l	EAGLEVERSION	;eus_EPVersion
	StructVer		;eus_version,revision
	dc.l	MyUsername	;eus_username
	dc.w	Project0Left	;eus_winx
	dc.w	Project0Top	;eus_winy
	dc.w	EUSB_Openwin	;window opened
	dc.w	1		;pri
	dc.w	0		;flags

	dc.w	0,0		;EUS_Special2
	dc.b	0,$10,0,0	;EUS_Special3

ZoomPosX	equ	EUS_Special2
ZoomPosY	equ	EUS_Special2+2
ZoomState	equ	EUS_Special3	;Zoom or not zoom
OnMonoMode	equ	EUS_Special3+1	;Mono Modules Mode (untersch. Delay)

	StructDate		;EUS_Creatordate
	dc.b	0		;EUS_Priority

	dc.b	EUTY_Amplifier	;EUS_Type

	dc.w	0		;EUS_Reserved2

	ifne	PAULA14
	 dc.l	`14B2`		;EUS_Reserved3;EUS_AMIDNr
	else
	 ifne	AHIOUT
	  dc.l	`AHI0`
	 else
	  dc.l	`TOC2`		;EUS_Reserved3;EUS_AMIDNr
	 endc
	endc
	dc.l	0		;EUS_Reserved4
	dc.l	MyUPSStruct	;EUS_Reserved5
	dc.l	0		;EUS_Reserved6

TagListe:
	dc.l	EUT_AttnFlags,AFF_68020
	dc.l	EUT_EPSubItems,EP_Menu
;	dc.l	EUT_SpecialinfoLNr,LNrE_SpecialInfo_14BitAmplifier
	dc.l	TAG_DONE

	ifne	PAULA14
	 dc.b	`$VER: EP_14Bit_Amplifier ` ; Eagleplayer 14 Bit 32 Voices Amplifier `
	else
	 ifne   AHIOUT
	  dc.b	`$VER: EP_AHI_Amplifier ` ;
	 else
	  dc.b	`$VER: EP_Toccata_Amplifier ` ;
	 endc
	endc
	ver
	dc.b	` `
	date
	dc.b	0
	even
SpecialJumps
	dc.l	Checkfeatures	;AMJ_CheckFeatures, Test ob Anforderungen des Replays erfüllt werden können
	dc.l	Amp_Init	;AMJ_Init	;?bergabe der Struktur/test&Init
	dc.l	Amp_StartInt 	;AMJ_StartInt	;Alloc Audio & Start Int
	dc.l	Amp_StopInt	;AMJ_StopInt	;Free Audio & Stop Int
	dc.l	Amp_End		;AMJ_End		;Mem freigeben, etc.

	dc.l	Copy_AS			;AMJ_Amplifier	;Einsprung, wenn neu gef?llte Amplifier-Struktur
					;vom Replayer ?bergeben wurde (meist nach DTP_Interrupt-
					;Abarbeitung)
	dc.l	0	;AMJ_PokeAdr	;nur fuer Noteplayer initialisieren,
	dc.l	0	;AMJ_PokeLen	;die auf die Hardware poken, wie
	dc.l	0	;AMJ_PokePer	;z.B. der Chipram Player, sonst
	dc.l	0	;AMJ_PokeVol	;auf 0 setzen, diese Jumps werden mit
	dc.l	0	;AMJ_DMAMask	;den selben Parametern aufgerufen, wie
					;die ENPP_Poke...
	dc.l	0	;AMJ_Command	;z.B. Filter
fang:
	move.l	4.w,a6

	ifeq	V37Gadtools
	cmp.w	#39,LIB_VERSION(A6)
	bge.s	.do39
	endc

	subq	#6,NG_Listview+6
.do39
	move.l	ex_EClockFrequency(a6),d0
	move.l	d0,d1
	add.l	d0,d0
	add.l	d0,d0
	add.l	d1,d0
	move.l	d0,AudioRate			;clock constant PAL/NTSC
	;move.l	d0,clock			;one of these is redundant

	lea	datas,a5

	move.l	a5,a0
	move.l	#DatasLen,d0
	bsr	ClearBuf
	
	lea	msgport1(a5),a0
	moveq	#-1,d0
	move.b	d0,mp_SigBit(a0)
	move.l	d0,mysignal(a5)
	move.l	d0,My2ndsignal(a5)


	;lea	ZeroSamplebuffer,a0		;moved into BSS Buffer range, always cleared before starting to play
	;move.l	#SamLEN*2*2+2,d0		;used in case that no mixing took place, always zero-filled
	;bsr	ClearBuf

	ifne	BUFCHECK
		push	all

		lea	Checkbuffers,A6
.bufcheckinit
		move.l	(A6)+,D0 ; check PTR
		beq.s	.nc
		move.l	d0,A0
		move.l	(A6)+,D0 ; check Length
		subq.l	#1,D0
		bcs.s	.nc
.bufcheckinit1
		move.l	#$DEADBEEF,(A0)+
		dbf	d0,.bufcheckinit1
		bra.s	.bufcheckinit
.nc
		Pop	All
	endc


	move	#-1,LastBoost

	move.w	AttnFlags(A6),d0
	and.w	#AFF_68020,d0		;020+ is sufficient, reason for older version requiring 68030 was fastram, which a plain A1200 doesn`t have
	move	d0,MC68020on(A5)	;wenn 030, dann UPS ins FastRam
	move.w	AttnFlags(A6),d0
	and.w	#AFF_68040,d0		;020+ is sufficient, reason for older version requiring 68030 was fastram, which a plain A1200 doesn`t have
	move	d0,MC68040on(A5)	;wenn 030, dann UPS ins FastRam

;	clr	MC68020on(a5)

	move	#-1,Currentnumvoc(a5)		;kein Modul am Spielen

	bsr	OpenLibs		;open required libraries
	tst.l	d0
	beq	NoLibs

	;---------------- config init --------------------------------------
	suba.l	a0,a0				;no specific output hardware(yet), use default
	bsr	SetOutputHW
	bsr	Makedefaults			;Default prefs setzen
	bsr	LoadCalibration			;load cybersound calibration table (if present)
	bsr	Loadconfig

	; the amplifier has the same name as the old one but a new
	; preferences structure: make sure the old preferences do not
	; collide with ours
	move	PrefsVersion(A5),d0
	cmp.w	#`EP`,d0			; EP was not in the old prefs
	bne	.def
	move	PrefsVersion+2(A5),d0
	cmp	#PrefsVer,d0
	beq	.useprefs
.def
;	move	#Project0Left,WinPosX(a5)
;	move	#Project0Top,WinPosY(a5)
	clr.l	ZoomPosX(A5)
	clr.b	Zoomstate(A5)
	move	#`EP`,Prefsversion(A5)
	move	#Prefsver,Prefsversion+2(A5)
	bsr	Makedefaults			;Default prefs setzen
.useprefs
	;---------------------------------------------------------------------
	st	onsysfont(A5)

	ifne	test
		Push	All
		bsr	Openwin
		Pop	All
		tst.l	Winhandle(A5)
		beq.s	endme
		bra.s	.noep
	endc

	moveq	#0,d7
	bsr	sendfirstmessage	;Message an Eagleplayer senden,da?
					;der Analyzer gern mal die Struktur-
					;Adresse haben m?chte
	blt	endme	;Fehler aufgetreten
.noep:
	bsr	HandleActions		;Auf Usereingaben reagieren

	tst.l	d7
	blt.s	endme
	bsr	sendlastmessage
endme:
	bsr	closedownwin
	bsr	closedownscreen

	move.l	Amp_OHWStatus,d0
	and.l	#OHWS_INIT,d0
	beq.s	.noOHWinit
	bsr	Amp_Exit
.noOHWinit:

NoLibs:
nowin:
	move.l	gadsave(a5),d0
	beq.s	.nogad
	move.l	d0,a1
	move.l	4.w,a6
	jsr	_LVOCloseLibrary(A6)
	clr.l	gadsave(A5)
.nogad:
	move.l	grafsave(a5),d0
	beq.s	.nograf
	move.l	d0,a1
	move.l	4.w,a6
	jsr	_LVOCloseLibrary(a6)
.nograf:
	move.l	intsave(a5),d0
	beq.s	.noint
	move.l	d0,a1
	move.l	4.w,a6
	jsr	_LVOCloseLibrary(a6)
.noint
	move.l	4.w,a6
	lea	msgport1(a5),a2
	move.b	mp_sigbit(a2),d0
	ext.w	d0
	ext.l	d0
	jsr	_LVOFreeSignal(a6)
NoPort:
	move.l	4.w,a6
	move.l	_dos(a5),d0
	beq.s	.nodos
	move.l	d0,a1
	jsr	_LVOCloseLibrary(a6)
.nodos:
Endprog:
	move.l	mysignal(a5),d0
	move.l	4.w,a6
	jsr	_LVOFreeSignal(a6)

	move.l	My2ndsignal(a5),d0
	move.l	4.w,a6
	jsr	_LVOFreeSignal(a6)
NoSignal:
	moveq	#0,d0
	rts

MakeMsgPort:
					;Message Port initialisieren
	move.l	4.w,a6
	moveq	#-1,d0
	jsr	_LVOAllocSignal(a6)	;MSG Signal zuweisen lassen
	move.b	d0,mp_sigbit(a4)
	tst.l	d0
	blt	.nosignal2		;Fehler

	move.b	#nt_msgport,ln_type(a4)
	move.b	#0,ln_pri(a4)
	move.b	#pa_signal,mp_flags(a4)

	move.l	mytask(a5),mp_sigtask(a4);mp_SigTask

	lea	mp_msglist(a4),a1
	NEWLIST	a1			;macro in Lists.i

	moveq	#0,d0
.nosignal2:				;Fehler
	rts

OpenLibs:
	move.l	4.w,a6
	moveq	#-1,d0
	move.l	d0,My2ndsignal(a5)
	jsr	_LVOAllocSignal(a6)	;Signal allokieren
	move.l	d0,mysignal(a5)
	blt	.NoSignal

	moveq	#-1,d0
	jsr	_LVOallocsignal(a6)	;2. Signal allokieren (f?r Interrupt)
	move.l	d0,My2ndsignal(a5)
	blt	.NoSignal

	suba.l	A1,A1
	JSR	_LVOFindTask(A6)
	move.l	d0,mytask(a5)

	lea	dos(pc),a1
	moveq	#37,d0			;Brauchen DOS V1.2,fuer Message,falls
	jsr	_LVOOpenLibrary(a6)	;Start von 1.2/1.3 aus
	move.l	d0,_dos(a5)
	beq	.NoDos

	lea	MsgPort1(a5),a4		;MSG Port Struktur initialisieren
	bsr	MakeMsgPort
	tst.l	d0
	blt	.NoPort
	move.l	a4,MsgPort(A5)

	lea	int(pc),a1
	moveq	#37,d0
	jsr	_LVOOpenLibrary(a6)
	move.l	d0,intsave(a5)
	beq	.NoInt

	lea	graf(pc),a1
	moveq	#37,d0
	jsr	_LVOOpenLibrary(a6)
	move.l	d0,grafsave(a5)
	beq	.NoGraf

	lea	gad(pc),a1
	ifeq	V37Gadtools
	moveq	#37,d0
	else
	moveq	#34,d0
	endc
	jsr	_LVOOpenLibrary(a6)
	move.l	d0,gadsave(a5)
	beq	.NoGadTo

	moveq	#1,d0
	rts

.NoSignal:
.NoDos:
.NoPort:
.NoGadTo:
.NoGraf:
.NoInt:
	moveq	#0,d0
	rts

;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
;--------------------------------- Output hardware dependent code ------------------------------------
;-----------------------------------------------------------------------------------------------------
;
; Redirectors for Init/End/Start/Stop

;---------------- just return -1 ----------------------------------------------
Amp_Ret_Error:
		moveq	#-1,d0
		rts
Amp_Ret_Error0:	; pointer-routines return 0 for error cases
		moveq	#0,d0
		rts

Amp_Func_Call	macro
		move.l	Amp_OHWStr,d1
		beq.w	Amp_Ret_Error
		move.l	d1,a1
		move.l	\1(A1),d1
		beq.w	Amp_Ret_Error
		move.l	d1,a1
		jmp	(a1)
		endm
Amp_Func_CallP	macro
		move.l	Amp_OHWStr,d1
		beq.w	Amp_Ret_Error0
		move.l	d1,a1
		move.l	\1(A1),d1
		beq.w	Amp_Ret_Error0
		move.l	d1,a1
		jmp	(a1)
		endm

;------------------ Amplifier init/end/start play/stop play (HW dependent) ---------------------
; Init
;          - Initplayer (once per module)
; in: 
;  A0: tag list
;  A5: EP Globals
; out:
;  D0: 0=OK, <>0=Error
; registers:
;  don't care, EP saves registers
Amp_Init:	Amp_Func_Call	OHW_Init
; Init
;          - Endplayer (once per module)
; in: 
;  A5: EP Globals
; out:
;  D0: 0=OK, <>0=Error
; registers:
;  don't care, EP saves registers
Amp_End:	Amp_Func_Call	OHW_Shutdown
; Start playing 
;
; in: 
;  A5: EP Globals
; out:
;  D0: 0=OK, <>0=Error
; registers:
;  don't care, EP saves registers
Amp_StartInt:	Amp_Func_Call	OHW_StartInt
; Stop playing 
;
; in: 
;  A5: EP Globals
; out:
;  D0: 0=OK, <>0=Error
; registers:
;  don't care, EP saves registers
Amp_StopInt:	Amp_Func_Call	OHW_StopInt
; Internal: check whether this output hardware is available
; in:
;  A5 - internal structures ("datas")
; out:
;  D0 - 1=yes, 0=no
; registers:
;  save all registers but D0 (!)
Amp_Check:	 Amp_Func_CallP	OHW_Check
; Internal: on change of the output hardware, update available output rates
; in:
;  A5 - internal structures ("datas")
; out:
;  D0 - pointer to 0-terminated long int array of mixing rates in Hz
;     - 0 in error case
; registers:
;  save all registers but D0 (!)
Amp_GetMixRates: Amp_Func_CallP	OHW_GetMixRates
; Internal: called when user selects a new desired mixing rate
; in:
;  A5 - internal structures ("datas")
;  D0 - new mixing rate
; out:
;  D0 - actual mixing rate (0 for error)
;
Amp_SetMixRate:	 Amp_Func_CallP	OHW_SetMixRate
; Internal: called to set mono-stereo-surround panning
; in:
;  A5 - internal structures ("datas")
;  D0 - panning
; out:
;  D0 - ok=0, error <> 0 
;       (also returns ok, when this function is not implemented, 
;        i.e. NULL pointer)
Amp_SetPlayMode:	 Amp_Func_CallP	OHW_SetPlayMode

; Internal: called when volume boost changed
; in:
;  A5 - internal structures ("datas")
;  D0 - boost value
; out:
;  D0 - ok=0, error <> 0 
;       (also returns ok, when this function is not implemented, 
;        i.e. NULL pointer)
Amp_SetBoost:		 Amp_Func_CallP	OHW_SetBoost

; Internal: setup output instance (libs etc.), called before any other functions
; in:
;  A5 - internal structures ("datas")
; out:
;  D0 - 0=ok, error <> 0
; registers:
;  save all registers but D0 (!)
Amp_Setup:	 Amp_Func_CallP	OHW_Setup

; Internal: cleanup output instance (libs etc.) on exit
; in:
;  A5 - internal structures ("datas")
; out:
;  D0 - 0=ok, error <> 0
; registers:
;  save all registers but D0 (!)
Amp_Exit:	 Amp_Func_CallP	OHW_Exit


*********************************************************************
* Clear a buffer (not fast, re-implement if performance is crucial) *
* Input: A0   - pointer                                             *
*        D0.l - number of bytes                                     *
* Regs:  none                                                       *
*********************************************************************
ClearBuf:
	push	d0/a0

.clr	move.b	#0,(a0)+
	subq.l	#1,d0
	bne.s	.clr

	pop	d0/a0
	rts
********************************************************************
*                  Integrity Checks                                *
********************************************************************
	ifne	BUFCHECK
DoCheckBuffers:
	movem.l	d0-a6,-(sp)

	eor.l	d1,d1
	lea	Checkbuffers,A6

.bufcheckinit
	move.l	(A6)+,D0 ; check PTR
	beq.s	.nc
	move.l	d0,A2
	
	move.l	(A6)+,D0 ; check Length
	subq.l	#1,D0
	bcs.s	.nc

	eor.l	d2,d2
	moveq.l	#-1,d3

.bufcheckinit1
	cmp.l	#$DEADBEEF,(A2)+
	beq.s	.ok

	move.l	d2,d3

.ok

	addq.l	#1,D2
	dbf	d0,.bufcheckinit1

	tst.l	d3
	blt.s	.ok2

	moveq	#1,d4
	lea	Internalpos+2,a3
.lo1	
	move.b	d3,d2
	and.b	#$f,d2
	cmp.b	#10,d2
	blt.s	.n1
	add.b	#`A`-`0`-10,d2
.n1
	add.b	#`0`,d2
	move.b	d2,-(a3)
	lsr	#4,d3
	dbf	d4,.lo1

	move.b	d1,Internalidx
	add.b	#`0`,(Internalidx)


	lea	Internalmessage(pc),a0
	bsr	Merror
.ok2
	
	addq	#1,d1
	bra.s	.bufcheckinit
.nc

	movem.l	(sp)+,d0-a6
	rts
	
	else	

DoCheckBuffers:
		rts

	endc



********************************************************************
*                  Init & Open Window                              *
********************************************************************
Openwin:
	tst.l	winhandle(A5)
	bne.w	.endme

	clr.l	port(A5)

	move.l	intsave(a5),a6
	bsr	Setupscreen		;Public Screen holen
	bne	.Endme
	bsr	SetUpWin		;Fenster ?ffnen
	tst.l	d0
	bne	.Endme

	bsr	drawbevel

	move.l	Winhandle(a5),a0
	move.l	wd_userport(a0),a0	;User Port
	move.l	a0,port(a5)
.endme
rtsA
	rts

;----------------------------------------------------------------------------------
; draw bevel boxes
;----------------------------------------------------------------------------------
DrawBevel
	push	all

	move.l	gadsave(A5),a6
	move.l	winhandle(A5),a0
	move.w	wd_height(a0),d0
	cmp	offy(a5),d0
	beq.s	.nobevel

	lea	Project0Bevel(pc),a4
	move.w	(a4)+,d7		;first word: count
	subq.w	#1,d7
	blt.s	.nobevel
.bevel_loop:
	move.l	winhandle(A5),a0
	move.l  wd_RPort(a0),a2
	lea.l   IR(pc),a1
	move.l  Vi(a5),4(a1)

	move.w	2(a4),d0		;move.w  #116,d0
	bsr	ComputeY
	move.w  offy(a5),d1
	add.w	d0,d1

	move	4(a4),d0		;move.w	#300,d0
	bsr	ComputeX
	move.w	d0,d2

	move	6(a4),d0		;move.w	#38,d0
	bsr	ComputeY
	move.w	d0,d3

	move	(a4),d0			;move.w  #4,d0
	bsr	ComputeX
	add.w   offx(A5),d0

	move.l  a2,a0
	jsr     _LVODrawBevelBoxA(a6)

	lea	8(a4),a4
	dbf	d7,.bevel_loop
.nobevel
	pop	all
	rts
ClearWin:
	move.l	winhandle(A5),a1
	move.w	wd_height(a1),d0
	cmp	offy(a5),d0
	beq.w	rtsA

	move.l	grafsave(a5),a6
	move.l	wd_rport(a1),a1
	moveq	#0,d0
	jsr	_LVOsetapen(A6)

	move.l	winhandle(A5),a1
	move.l	wd_rport(a1),a1
	moveq	#0,d0
	jsr	_LVOsetbpen(A6)

	move.l	winhandle(A5),a1
	move.l	wd_rport(a1),a1
	moveq	#1,d0			;RP_Jam2
	jsr	_LVOsetdrmd(A6)

	move.b	rp_areaptsz(a1),d4
	move.l	rp_areaptrn(a1),d5

	move.l	winhandle(A5),a1
	move.w	wd_width(A1),d2		;x2
	moveq	#0,d0
	move.b	wd_borderright(a1),d0
	sub.w	d0,d2

	move.w	wd_height(a1),d3
	moveq	#0,d0
	move.b	wd_borderbottom(a1),d0
	sub.w	d0,d3

	move.l	wd_rport(a1),a1

	move.w	offx(a5),d0		;x1
	move.w	offy(A5),d1		;y1
	jsr	_LVORectFill(a6)

	move.l	winhandle(A5),a0
	move.l	intsave(a5),a6
	jsr	_LVORefreshWindowFrame(A6)
	rts
*************** Window & Screen freigeben *************************
Closewin:
	bsr	closedownwin
	bra	closedownscreen
********************************************************************
*                  Init Message Struktur                           *
********************************************************************
Sendfirstmessage
	move.l	4.w,a6
	lea	Portname(pc),a1
	jsr	_LVOFindPort(a6)
	tst.l	d0
	beq.s	.error

	move.l	d0,a0			;Portadresse

	move.l	msgport(a5),d3
	move.l	mytask(a5),d4
	move.l	#-1,d5
	move.w	#USclass_Newuserprg,d6

	ifne	debug

	bsr	allocmsg
	beq.s	.error
	move.l	d0,a1			;Messageadresse

	move.w	#UM_sizeof-20,mn_length(a1)
	move.b	#nt_message,ln_type(a1) ;Message-Typ
	lea	begin(pc),a2
	move.l	a2,um_result(a1)
	move.w	#-2,UM_UserNr(a1)
	move.l	#USM_Userprg,UM_Type(a1)

	move.l	d3,mn_replyport(a1)	;Portadresse,an die
						;zur?ckgesendet wird
	move.l	d3,UM_Userport(a1)
	move.l	d4,UM_TaskAdr(a1)
	move.l	d5,UM_Signal(a1)
	move.w	d6,UM_class(a1)

	move.l	4.w,a6
	jsr	_LVOputmsg(a6)

	else
	bsr	sendmess
	endc

	moveq	#0,d7
	rts
.error
	moveq	#-1,d7
	rts
*******************************************************************************
sendmess:
	bsr	allocmsg
	bne.s	.ok
	rts
.ok
	move.l	d0,a1			;Messageadresse

	move.w	#UM_sizeof-20,mn_length(a1)
	move.b	#nt_message,ln_type(a1) ;Message-Typ

	ifne	debug
	move.w	saveusernr(a5),UM_UserNr(a1)
	else
	lea	begin(pc),a2
	move.w	EUS_Usernr(A2),UM_UserNr(a1)
	endc

	move.l	#USM_Userprg,UM_Type(a1)

	move.l	d3,mn_replyport(a1)	;Portadresse,an die
						;zur?ckgesendet wird
	move.l	d3,UM_Userport(a1)
	move.l	d4,UM_TaskAdr(a1)
	move.l	d5,UM_Signal(a1)
	move.w	d6,UM_class(a1)

	move.l	4.w,a6
	jmp	_LVOputmsg(a6)
******************************************************************************
* Abschiedsgru? an Eagleplayer,der dann wei?,da? der Analyzer geendet hat    *
******************************************************************************
Sendlastmessage:
	move.l	4.w,a6
	lea	Portname(pc),a1	;EaglePlayer finden
	jsr	_LVOfindport(a6)
	tst.l	d0
	beq.w	.ok	;Eagleplayer hat sich schon verabschiedet

	move.l	d0,a0			;Portadresse
	move.l	msgport(a5),d3
	moveq	#0,d4
	moveq	#0,d5
	move.w	#USclass_exit,d6

	bsr	sendmess
.wait
	move.l	MSGport(a5),a0
	jsr	_LVOwaitport(a6)
	move.l	MSGport(a5),a0
	jsr	_LVOgetmsg(a6)
	tst.l	d0
	beq.s	.wait
	move.l	d0,a1
	cmp.l	#USM_Userprg,UM_Type(a1)
	beq.s	.meine
.nichtmeine
	jsr	_LVOreplymsg(a6)
	bra.s	.wait
.meine
.ok
	rts
*******************************************************************
*      Public Screen ranholen und Visual Info anfordern           *
*******************************************************************
SetupScreen
	move.l	Intsave(a5),a6
	lea	begin(pc),a0
	move.l	eus_epbase(a0),d0
	beq.s	.noep
	move.l	d0,a0
	move.l	EPG_Pubscreen(A0),d0
.noep
	move.l	d0,a0
	jsr	_LVOLockPubScreen(a6)
	move.l	d0,MyPubscreen(A5)
	beq.s	.error

	move.l	mypubscreen(A5),a0
	move.l	Gadsave(A5),a6
	suba.l	a1,a1
	jsr	_LVOGetVisualInfoA(a6)
	move.l	d0,Vi(A5)
	beq.s	.error     
	moveq	#0,d0
	rts
.error
	moveq	#-1,d0
	rts
*******************************************************************
*      Public Screen ranholen und Visual Info anfordern           *
*******************************************************************
CloseDownScreen:
	Push	All
	
	move.l	Gadsave(a5),a6
	move.l	Vi(A5),d0
	beq.s	.novi
	move.l	d0,a0
	jsr	_LVOFreeVisualInfo(a6)
	clr.l	Vi(A5)
.novi
	move.l	Intsave(A5),a6
	suba.l	a0,a0
	move.l	Mypubscreen(A5),d0
	beq.s	.nopub
	move.l	d0,a1
	jsr	_LVOUnlockPubScreen(a6)
	clr.l	Mypubscreen(A5)
.nopub
	Pop	All
	rts
GetFont
	tst.b	onsysfont(A5)
	beq.s	.stand

	move.l	grafsave(A5),a6
	lea.l	MyRastPort(a5),a3
	move.l	a3,a1
	jsr	_LVOInitrastport(A6)
	move.l	a3,a1
	lea.l	Attr(a5),a0
	jsr	_LVOAskFont(A6)
	lea.l	Attr(a5),a0
	jsr	_LVOOpenfont(A6)
	move.l	d0,fontadr(A5)
	beq.s	.stand
	move.l	d0,a1
	rts
.stand
	move.l	mypubscreen(A5),a1
	lea.l	sc_RastPort(a1),a1
	move.l	rp_Font(a1),a1
	rts
ComputeFont:
	bsr	GetFont

	lea.l	Attr(a5),a0
	move.l	a0,Font(A5)

	move.l  LN_NAME(a1),ta_Name(a0)
	move.w  tf_YSize(a1),ta_YSize(a0)
	move.w  tf_XSize(a1),FontX(A5)
	move.w  tf_YSize(a1),FontY(A5)

	move.l	fontadr(A5),d0
	beq.s	.nofree
	move.l	d0,a1
	move.l	grafsave(A5),a6
	jsr	_LVOClosefont(A6)
.nofree
	move.l	mypubscreen(A5),a0
	move.b	sc_WBorLeft(a0),d0
	ext.w	d0
	move.w	d0,OffX(A5)

	move.l	sc_font(A0),a1
	move.w	ta_YSize(a1),d0

	addq.w	#1,d0
	move.b	sc_WBorTop(a0),d1
	ext.w	d1
	add.w	d1,d0
	move.w	d0,OffY(A5)

	tst.w	d2
	beq.s	CompDone
	tst.w	d3
	beq.s	CompDone
	move.w	d2,d0
	bsr	ComputeX

	add.w   OffX(A5),d0
	move.b	sc_WBorRight(a0),d1
	ext.w	d1
	add.w	d1,d0
	cmp.w	sc_Width(a0),d0
;	ifne	test
;	bra.s	Usetopaz
;	endc
	bhi.s	UseTopaz
	move.w	d0,d2

	move.w	d3,d0
	bsr	ComputeY
	add.w	OffY(a5),d0
	move.b	sc_WBorBottom(a0),d1
	ext.w	d1
	add.w	d1,d0
	cmp.w	sc_Height(a0),d0
	bhi.s	UseTopaz
	move.w	d0,d3
CompDone:
	ext.l	d2
	ext.l	d3
	rts
UseTopaz:
	add.w   OffX(A5),d2
	move.b	sc_WBorRight(a0),d1
	ext.w	d1
	add.w	d1,d2

	add.w	OffY(a5),d3
	move.b	sc_WBorBottom(a0),d1
	ext.w	d1
	add.w	d1,d3

	move.l	Font(A5),a0
	move.l	#TopazName,ta_Name(a0)
	move.w	#8,FontY(A5)
	move.w	#8,FontX(a5)
	move.w	#8,ta_YSize(a0)
	rts
ComputeX:
	move.l	d1,-(sp)
	move.w	FontX(a5),d1
	bra.s	ComputeItX
ComputeY:
	move.l	d1,-(sp)
	move.w	FontY(A5),d1
	mulu	d1,d0
	addq.w	#4,d0
	lsr.w	#3,d0
	move.l	(sp)+,d1
	rts
ComputeItX:
	mulu	d1,d0
	addq.w	#4,d0
	lsr.w	#3,d0
	move.l	(sp)+,d1
	rts

*------------ Hookfunktion, um aktuellen Modus farblich hervorzuheben ------*
*---------------- derzeit noch kein Support fuer "disabled"-Gadgets ---------*
*benutzte Globals:
* - grafsave+datas = Graphics Base
*
*Erkennung, ob Zeile hervorzuheben oder nicht:
* -  erstes Zeichen im darzustellenden Text ist ein "|"
	;a0:Hook
	;a1:Message
	;a2:Node
my_hook
	Push	d1-a6

	move.l	lvdm_MethodID(a1),d1
	cmp.l	#LV_Draw,d1
	bne.w	.hi
	move.l	a1,a4
	move.l	lvdm_RastPort(a4),d7
	beq.w	.hi

	move.l	grafsave+datas,a6			;Graphics Base

	move.l	lvdm_DrawInfo(A4),d0
	beq.w	.hi
	move.l	d0,a1
	move.l	dri_Pens(a1),a1

	sf	d2
	move.l	ln_name(A2),a5		;
	cmp.b	#`|`,(a5)		; Test ob hervorzuheben oder nicht
	bne.s	.textpen
	st	d2

	addq.l	#1,a5
	move	highlighttextpen*2(a1),d0
	bra.s	.high
.textpen
	move	textpen*2(a1),d0
.high
	move	backgroundpen*2(a1),d6

	cmp.l	#LVR_SELECTED,lvdm_State(a4)
	bne.s	.normalBG
	move	FILLPEN*2(a1),d6
	move	textpen*2(a1),d0
	tst.b	d2
	beq.s	.normalBG
	move	highlighttextpen*2(a1),d0
.normalBG
	move	d6,d1
	moveq	#RP_Jam1,d2
	move.l	d7,a1
	jsr	_LVOsetabpendrmd(A6)

	move.w	lvdm_bounds+ra_minx(a4),d0
	move.w	lvdm_bounds+ra_miny(a4),d1
	move.l	d7,a1
	add	rp_TxBaseline(a1),d1
	jsr	_LVOMove(A6)

	move.l	d7,a1
	moveq	#RP_Jam2,d0
	jsr	_LVOsetdrmd(A6)

	move.w	lvdm_bounds+ra_maxx(a4),d5		;Begrenzung des Textes
	sub	lvdm_bounds+ra_minx(a4),d5		;auf Breite des
	blt	.hi					;Listviews, solange

	move.l	a5,a0					;Text
	moveq	#-1,d3					;bis der Text passt
.lop
	addq	#1,d3
	tst.b	(a0)+
	bne.s	.lop
.charloop
	move.l	d7,a1
	move.l	a5,a0					;Text
	move.w	d3,d0
	jsr	_LVOtextlength(A6)
	cmp.w	d5,d0
	blt.s	.ok
	subq.w	#1,d3
	bra.s	.charloop
.ok
	move	d0,d4
	move.w	d3,d0
	move.l	a5,a0					;Text
	move.l	d7,a1
	jsr	_LVOtext(A6)

	move.l	d7,a1
	move	d6,d0
	move	d6,d1
	moveq	#0,d2
	jsr	_LVOsetabpendrmd(A6)

	move	lvdm_bounds+ra_minx(a4),d0
	add	d4,d0
	move	lvdm_bounds+ra_miny(a4),d1
	move	lvdm_bounds+ra_maxx(a4),d2
	move	lvdm_bounds+ra_maxy(a4),d3
	move.l	d7,a1
	jsr	_LVORectfill(a6)	;freien Platz nach dem Text l?schen
	
	Pop	d1-a6
	moveq	#LVCB_OK,d0
	rts
.hi
	Pop	d1-a6
	moveq	#LVCB_Unknown,d0
	rts
;------------------------------------------------------------------------------------------
CreateGadgets
	move.l	Gadsave(A5),a6
	lea.l	Project0GList(a5),a0
	jsr	_LVOCreateContext(a6)
	move.l	d0,a3
	tst.l	d0
	beq.w	.Project0CError

	moveq	#0,d3
	lea.l	Project0GTags(pc),a4
.Project0GL:
	move.l	4.w,a6
	lea.l	Project0NGads(pc),a0
	move.l	d3,d0
	mulu	#gng_SIZEOF,d0
	add.l	d0,a0
	move.l	gng_UserData(a0),d0
	cmp.l	#GD_Disable,d0
	beq.s	.Project0SkipDisabled

	lea.l	BufNewGad(a5),a1
	moveq	#gng_SIZEOF,d0
	jsr	_LVOCopyMem(a6)
	lea.l	BufNewGad(a5),a0
	move.l	Vi(a5),gng_VisualInfo(a0)
	move.l	Font(a5),gng_TextAttr(a0)

	move.w	gng_LeftEdge(a0),d0
	bsr	computex
	add.w	OffX(a5),d0
	move.w	d0,gng_LeftEdge(a0)

	move.w	gng_TopEdge(a0),d0
	bsr	computey
	add.w	OffY(a5),d0
	move.w	d0,gng_TopEdge(a0)

	move.w	gng_Width(a0),d0
	bsr	computex
	move.w	d0,gng_Width(a0)

	move.w	gng_Height(a0),d0
	bsr	computey
	move.w	d0,gng_Height(a0)

	move.l	Gadsave(a5),a6
	lea.l	Project0GTypes(pc),a0
	moveq	#0,d0
	move.l	d3,d1
	asl.l	#1,d1
	add.l	d1,a0
	move.w	(a0),d0
	move.l	a3,a0
	lea.l	BufNewGad(a5),a1
	move.l	a4,a2
	jsr	_LVOCreateGadgetA(a6)
	tst.l	d0
	bne.s	.Project0COK
	bra	.Project0CError
.Project0SkipDisabled:
	;
	bra.s	.Project0DN
.Project0COK:
	move.l	d0,a3
	move.l	d3,d0
	asl.l	#2,d0
	lea.l	Project0Gadgets(a5),a0
	add.l	d0,a0
	move.l	a3,(a0)
.Project0TL:
	tst.l	(a4)
	beq.s	.Project0DN
	addq.w	#8,a4
	bra.s	.Project0TL
.Project0DN:
	addq.w	#4,a4
	addq.w	#1,d3
	cmp.w	#Project0_CNT,d3
	bmi	.Project0GL
.Project0CError
	rts
*************************************************************************
*            Fenster ?ffnen, Gadgets und Men?s init                     *
*************************************************************************
SetUpWin:
	movem.l d1-d4/a0-a4/a6,-(sp)

	move.l	mypubscreen(a5),a0
	move.l  sc_Font(a0),a1
	move.w  ta_YSize(a1),d3
	addq.w  #1,d3
	move.b  sc_WBorTop(a0),d0
	ext.w   d0
	add.w	d3,d0
	move.w	d0,offy(a5)

	move.b  sc_WBorLeft(a0),d0
	ext.w   d0
	move.w	d0,offx(a5)

	move.w	#Project0Width,d2
	move.w	#Project0Height,d3

	bsr	ComputeFont
	move.l	mypubscreen(A5),a0

	move.l	d2,ahawidth
	move.l	d3,ahaheight

	move.l	d2,d0
	move.l	d0,d4

	moveq	#0,d2
	add.w	d2,d0
	add.w	OffX(a5),d0
	move.b	sc_WBorRight(a0),d1
	ext.w	d1
	add.w	d1,d0
	cmp.w	sc_Width(a0),d0
	bls.s	.Project0WOk
	move.w	sc_Width(a0),d0
	sub.w	d4,d0
	move.w	d0,d2
.Project0WOk:
	move.l	d3,d0
	move.l	d0,d4

	lea	begin(pc),a1
	move.w	eus_winy(a1),d3
	add.w	d3,d0

	add.w	OffY(A5),d0
	move.b	sc_WBorBottom(a0),d1
	ext.w	d1
	add.w	d1,d0
	cmp.w	sc_Height(a0),d0
	bls.s	Project0HOk
	move.w	sc_Height(a0),d0
	sub.w	d4,d0
	move.w	d0,d3
Project0HOk:
	move.l	Gadsave(A5),a6
	lea.l	Project0NewMenu0(pc),a0

	move.w	Hiercheck_Mono(pc),d0		;Men?eintrag: Mono aus den prefsssetzen
	and.w	#~Checked,d0
	move.b	Begin+OnMonoMode(pc),d1
	beq.s	.noset
	or.w	#Checked,d0
.noset	move	d0,HierCheck_Mono	

	lea.l	MTags0(pc),a1
	jsr	_LVOCreateMenusA(a6)

	move.l	d0,Project0Menus(A5)
	tst.l	d0
	beq	Project0MError
	move.l	d0,a0
	move.l	Vi(A5),a1
	lea.l	mtags0(pc),a2
	jsr	_LVOLayoutMenusA(a6)

	moveq	#0,d0
	moveq	#0,d1
	moveq	#0,d2
	moveq	#0,d3
	moveq	#0,d7
	lea	begin(Pc),a1
	move	eus_Winx(a1),d0
	move	eus_Winy(a1),d1
	move	ZoomPosX(a1),d2
	move	ZoomPosY(a1),d3

	move.l	ahawidth(pc),d4
	move.l	ahaheight(pc),d5
	move.l	#ZoomWidth,d6
	move	offy(A5),d7

	tst.b	ZoomState(a1)
	beq.s	.Normal
	exg.l	d0,d2
	exg.l	d1,d3
	exg.l	d4,d6
	exg.l	d5,d7
.normal
	lea.l	Project0WindowTags(pc),a1
	move.l	d0,Project0L+4-Project0WindowTags(A1)
	move.l	d1,Project0T+4-Project0WindowTags(A1)
	move.l	d4,Project0W+4-Project0WindowTags(A1)
	move.l	d5,Project0H+4-Project0WindowTags(A1)
	move.w	d2,ZoomStruct-Project0WindowTags(A1)
	move.w	d3,ZoomStruct+2-Project0WindowTags(A1)
	move	d6,ZoomStruct+4-Project0WindowTags(A1)
	move	d7,ZoomStruct+6-Project0WindowTags(A1)

	move.l	Intsave(A5),a6
	move.l	mypubscreen(a5),Pubadr-Project0WindowTags(a1)
	suba.l	a0,a0
	jsr	_LVOOpenWindowTagList(a6)
	move.l	d0,Winhandle(a5)
	tst.l	d0
	beq	Project0WError
.open13
	move.l	Winhandle(a5),a0
	move.l	Project0Menus(A5),a1
	jsr	_LVOSetMenuStrip(a6)

	bsr	createGadgets

	bsr	AddGads

	moveq	#0,d0
Project0Done:
	movem.l	(sp)+,d1-d4/a0-a4/a6
	rts
Project0CError:
	moveq	#1,d0
	bra.s	Project0Done
Project0GError:
	moveq	#2,d0
	bra.s	Project0Done
Project0MError:
	moveq	#3,d0
	bra.s	Project0Done
Project0WError:
	moveq	#4,d0
	bra.s	Project0Done

AddGads:
	movem.l	d0-a6,-(sp)
	move.l	Project0GList(a5),d0
	beq.s	.nogads
	move.l	d0,a1

	moveq	#0,d1
.count
	move.l	d0,a0

	addq.l	#1,d1

	move.l	(A0),d0
	bne.s	.count

	move.w	d1,RealNumGadgets(A5)

	move.l	intsave(a5),a6
	move.l	winhandle(a5),a0
	moveq	#-1,d0
	moveq	#-1,d1
	suba.l	a2,a2
	jsr	_LVOaddglist(a6)

	move.l	Project0GList(a5),a0
	move.l	winhandle(a5),a1
	suba.l	a2,a2
	moveq	#-1,d0
	jsr	_LVOrefreshglist(a6)

	move.l	GadSave(a5),a6
	move.l	Winhandle(a5),a0
	suba.l	a1,a1
	jsr	_LVOGT_RefreshWindow(a6)

	bsr	UpdateList
.nogads
	movem.l	(sp)+,d0-a6
	rts
*************************************************************************
*            Fenster schlie?en, Gadgets und Men?s free                  *
*************************************************************************
Closedownwin:
	Push	All
	move.l	Intsave(A5),a6
	move.l	Project0Menus(A5),a0
	cmpa.l	#0,a0
	beq	Project0NMenu
	move.l	Winhandle(a5),a0
	jsr	_LVOClearMenuStrip(a6)
	move.l	GadSave(a5),a6
	move.l	Project0Menus(A5),a0
	jsr	_LVOFreeMenus(a6)
	move.l	#0,Project0Menus(A5)
Project0NMenu:
	move.l	Intsave(A5),a6
	move.l	Winhandle(a5),d0
	beq.s	.nowin
	move.l	d0,a0
	jsr	_LVOCloseWindow(a6)
	clr.l	Winhandle(a5)
	clr.l	port(a5)
.nowin
	bsr	Freegadgets
	Pop	All
	rts
FreeGadgets
	movem.l	d0-a6,-(sp)

	move.l	GadSave(a5),a6
	move.l	Project0GList(A5),a0
	cmpa.l	#0,a0
	beq	.nogad
	jsr	_LVOFreeGadgets(a6)
	clr.l	Project0GList(A5)
.nogad
	movem.l	(sp)+,d0-a6
	rts
*************************************************************
*    Gadgetadresse ausrechnen                               *
* Input D0 - Gadgetnummer (0 < X < MAX)                     *
*************************************************************
_Gadr
	lea	Gadgetadrs(a5),a0
	asl.l	#2,d0
	move.l	(a0,d0),a0
	rts
*************************************************************
*           Nachrichtenbearbeitung                          *
*************************************************************
HandleActions
.ma
;	bra.s	notestvolume$
	; Signal from Interrupt: new volume or balance
	btst	d3,d0
	beq.s	notestvolume$

	tst.b	onplay
	beq.s	notestvolume$

;	move.l	MyEPBase(pc),a1
;	move.l	dtg_SndVol(A1),d0	; dtg_SndVol & dtg_SndLBal
;	move.w	dtg_SndRBal(A1),d1	; dtg_SndVol & dtg_SndLBal
	
;	cmp.w	tmpvol2(pc),d1
;	bne.s	newvol$
;	cmp.l	tmpvol(pc),d0
;	beq.s	notestvolume$
;newvol$
	bsr	Makevoltable
	bclr	d3,d0			;unsure about the flags...
	tst.l	d0
	beq.s	.gowait
	;bra.s	.gowait
notestvolume$
	move.l	gadsave(a5),a6

	move.l	port(a5),d0
	beq.s	.nowinny
	move.l	d0,a0
	jsr	_LVOGT_getimsg(a6)
	tst.l	d0
	bne	.msg
.nowinny
	move.l	4.w,a6
	
	move.l	msgport(a5),a0
	jsr	_LVOgetmsg(a6)
	tst.l	d0
	bne	.AN_msg

	bsr	UpdateDownGadgets
	bne	.ma
.gowait
	moveq	#0,d0
	move.l	4.w,a6

	ifne	userprogram
	move.l	msgport(a5),a0
	move.b	mp_sigbit(a0),d1
	bset	d1,d0
	endc

	move.l	Port(a5),d2
	beq.s	.nomain
	move.l	d2,a0
	move.b	mp_sigbit(a0),d1
	bset	d1,d0
.nomain
	move.l	Mysignal(a5),d3	;Signal fuer Analyzer
	bset	d3,d0		;Signalbit fuer Abarbeitung des Analyzers selbst
				;(Auswertungsroutine namens "ANALYZE1")
	move.l	My2ndsignal(a5),d3	;Signal fuer Volume Change
	bset	d3,d0		;Neuberechnung der Tabellen bei Lautstaerke/Balance Aenderung
			
	jsr	_LVOwait(a6)
	bra	.ma
;======================= Message an Hauptwindow ============================
.msg
	move.l	d0,a1
	move.l	im_class(a1),d7
	move.w	im_code(a1),d5 		<- fuer Tastatur / Selnummer Listview
	move.l	im_Iaddress(a1),d6 	<- fuer Gadgetabfrage
	move.l	im_seconds(a1),d2	<- fuer Doppelclick
	move.l	im_micros(a1),d3	<- auch f. Doppelclick
	jsr	_LVOGT_replyimsg(a6)

	cmp.l	#IDCMP_Changewindow,d7	;Refresh
	bne.s	.nochangewin
	lea	begin(pc),a1
	
	ifne	BUFCHECK
		bsr	DoCheckBuffers
	endc

	move.l	winhandle(a5),a0
	move	OffY(A5),d0
	cmp	wd_height(a0),d0
	bne.s	.setposun
	move.w	wd_leftedge(a0),ZoomPosX(a1)
	move.w	wd_topedge(a0),ZoomPosY(a1)
	st	ZoomState(a1)
	bra	.ma
.setposun
	move.w	wd_leftedge(a0),EUS_WinX(a1)
	move.w	wd_topedge(a0),EUS_Winy(A1)
	sf	ZoomState(a1)
	bra	.ma
.nochangewin
	cmp.l	#refreshwindow,d7	;Refresh
	bne.s	.noref
	pea	.ma(pc)
.ref
	bsr	clearwin

	move.l	gadsave(a5),a6
	move.l	winhandle(a5),a0	;*Window
	jsr	_LVOGT_beginrefresh(A6)

	move.l	winhandle(a5),a0	;*Window
	moveq	#1,d0
	jsr	_LVOGT_Endrefresh(A6)	

	bsr	drawbevel
	rts
	;------------------ Gadgets abfragen ------------------------
.noref
	cmp.l	#Gadgetup,d7
	bne.w	.notup

	move.l	d6,a1

	move.w	gg_gadgetid(A1),d4	;ID fuer Gadgets

	bsr	TestGadgets
	bra	.ma
.notup
	cmp.l	#Gadgetdown,d7
	bne.w	.notgad

	move.l	d6,a1

	move.w	gg_gadgetid(A1),d4	;ID fuer Gadgets

	bsr	TestDOWNGadgets
	bra	.ma

	;------------------------- Men?s abfragen ----------------------------
.menuloop
	move.l	temp(A5),a1
	move.w	mi_nextselect(A1),d5
	bra.s	.menuwei
.notgad:
	cmp.l	#menupick,d7
	bne.w	.nomenu
.menuwei
	move.w	d5,d0
	cmp.w	#menunull,d0
	beq	.ma

	move.l	Project0Menus(A5),a0
	move.l	intsave(a5),a6
	jsr	_LVOitemaddress(a6)
	move.l	d0,a1
	tst.l	d0
	beq	.ma
	move.l	d0,temp(A5)

	move.w	d5,d0
	and.w	#$1f,d0
	bne.s	.2ndmenu
	move.w	d5,d0
	and.w	#$7e0,d0
	bne.s	.noMonoMode

	move.w	mi_flags(a1),d0
	and.w	#checked,d0
	move.w	d0,d1
	lsr.w	#4,d0
	move.b	d0,begin+OnMonoMode	;TODO: change prefetch

	move.w	Hiercheck_Mono(pc),d0
	and.w	#~Checked,d0
	or.w	d1,d0
	move.w	d0,Hiercheck_Mono

	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist+P_SurroundDelay(a5),a0
	move.b	(A0,d0.w),d0
	bsr	setPrefetch

	bra	.menuloop

.noMonoMode
	cmp.w	#64,d0
	bne.s	.noabout		;1. Item (About) ?
	lea	Aboutmessage(pc),a0
	bsr	Merror
	bra	.menuloop
.noabout
	cmp.w	#96,d0			;2. Item (Hide)
	beq.w	.quitit
	cmp.w	#128,d0			;3. Item (Quit)
	beq	.killit
	bra	.menuloop
.2ndmenu				;2. Men?
	move.w	d5,d0
	and.w	#$7e0,d0
	bne.s	.noglobal

	move.w	mi_flags(a1),d0
	and.w	#checked,d0
	move.w	d0,d1
	lsr.w	#4,d0
	move.b	d0,Onglobal(A5)

	move.w	Hiercheck(pc),d0
	and.w	#~Checked,d0
	or.w	d1,d0
	move.w	d0,hiercheck

	moveq	#GD_AHIGlobal,d0
	bsr	_GADR
	tst.l	a0
	beq.s	.nobutton
	lea     mytags(a5),a3           ;tags
	move.l	#GTCB_Checked,(a3)+
	move.l	#Checked,d0
	and.w	Hiercheck(pc),d0
	beq.s	.nobuttoncheck
	moveq	#1,d0
.nobuttoncheck
	move.l	d0,(a3)+
	move.l	#TAG_DONE,(a3)+
	bsr	SetGAD
.nobutton

	bsr	SetVoic

	bra	.menuloop
.noglobal
	cmp	#64,d0
	bne.s	.nodefaults

	bsr	Makedefaults			;Default prefs setzen
	bsr	Updatelist

	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist+P_Mixrate(a5),a0
	move.l	(A0,d0.w),d0
	bsr	CalcValues

	moveq	#0,d0
	bsr	Amp_SetPlayMode

	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist+P_SurroundDelay(a5),a0
	move.b	(A0,d0.w),d0
	bsr	setPrefetch

	bra	.menuloop
.nodefaults
	cmp.w	#96,d0
	bne	.nosavepref
	bsr	SaveConfig
.nosavepref
	bra	.menuloop
.nomenu
; -> doesn`t seem to work, at least not with my PIC based adapter and WMD
;	cmp.l	#IECLASS_NEWMOUSE,d7
;	bne.s	.nowheel
	; handle wheel mice
;	nop
;.nowheel:
	cmp.l	#rawkey,d7		;Mouse wheel: both RAWKEY and NEWMOUSE
	bne	.nokey

	cmp.b	#NM_WHEEL_UP,d5
	bne.s	.nowheelup
	nop
.nowheelup:
	
	cmp.b	#NM_WHEEL_DOWN,d5
	bne.s	.nowheeldown
	nop
.nowheeldown:

	cmp.b	#$21,d5
	bne.s	.noS

	lea	Prefslist+P_Surround(a5),a0

	tst.b	onglobal(A5)
	bne.s	.sliPMglobal

	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	clr.b	(A0,d0.w)
	bra.s	.sliPMup
.sliPMglobal
	moveq	#MaxVoices-2,d0
.sliPMall
	clr.b	(A0)			;Playmode
	add.w	#P_Sizeof,a0
	dbf	d0,.sliPMall
.sliPMup
	bsr	UpdateList		;Sp?ter nur Listview update !!!!
	moveq	#0,d0
	bsr	Amp_SetPlayMode
	bra	.ma
.noS
	cmp.b	#$45,d5			;ESC
	beq	.quitit
	bra	.ma
.nokey
	cmp.l	#CloseWindow,d7
	bne	.ma
.quitit
	ifne	userprogram

	bsr	closewin

	move.l	4.w,a6
	lea	Portname(pc),a1
	jsr	_LVOfindport(a6)
	tst.l	d0
	beq.s	.endme		;kein Eagleplayerport->Ende des Programmes
	move.l	d0,a0			;Portadresse
	move.l	msgport(a5),d3
	move.l	mytask(a5),d4
;	move.l	mysignal(a5),d5
	move.l	#-1,d5
	move.w	#USclass_Hide,d6		;DeActivate,d6
	bsr	sendmess
	bra	.ma

	endc
.endme
	moveq	#-1,d7
	rts
.killit
	moveq	#0,d7
	RTS
;===================== Message an Userprogramm Port =====================
	ifne	userprogram
.AN_msg
	move.l	d0,a1

	move.l	MSGPort(a5),UM_Userport(a1)

	lea	begin(pc),a2

	ifne	debug
	move.w	um_usernr(a1),saveusernr(A5)
	else
	move.w	EUS_Usernr(A2),UM_UserNr(a1)
	endc

	move.l	mytask(a5),UM_TaskAdr(a1)
;	move.l	mysignal(a5),d1
	move.l	#-1,d1
	move.l	d1,UM_Signal(a1)

	move.w	UM_class(a1),d3		<-
	cmp.w	#USclass_exit,d3
	bne.w	.noex1

	clr.l	um_signal(a1)
	clr.l	um_taskadr(a1)
	clr.l	um_userport(a1)
.noex1
	move.l	UM_Type(a1),d7
	move.l	UM_structadr(a1),d6 	<-

	clr.w	UM_class(A1)

	cmp.w	#USclass_Saveconfig,d3
	bne	.nosavepref1
	bsr	SaveConfig
.nosavepref1
	cmp.w	#USclass_getconfig,d3		;when this command comes,set
						;the actual preferences into
						;the Eagleuser-Structure
						;(EUS_-structure at the be-
						;ginning of the program)
	bne	.noprev

	move.w	eus_flags(a2),d0
	and.w	#~EUSB_Show,d0

	move.l	winhandle(A5),d1		;Window
	beq.s	.prnowin			;no Window opened,no change
						;of preferences
	or.w	#eusb_Show,d0			;set Flag: window is open

	move.l	d1,a0				;Window
	move.w	wd_leftedge(a0),eus_winx(a2)	;Save Windowpostion
	move.w	wd_topedge(a0),eus_winy(a2)

;	move.w	wd_width(a0),d2
;	moveq	#0,d1
;	move.b	wd_borderleft(A0),d1
;	sub.w	d1,d2
;	move.b	wd_borderright(A0),d1
;	sub.w	d1,d2
.prnowin
	move.w	d0,eus_flags(a2)		;Save Flags
.noprev

	cmp.w	#USCLASS_Show,d3
	bne.s	.noshow

	move.l	MYEpBase(pc),a0
	move.l	EPG_SomePrefs(a0),d0
	btst	#EGPRF_Iconify,d0
	bne.w	.noshow

	Push	All
	bsr	Openwin
	Pop	All
	tst.l	Winhandle(A5)
	beq.s	.noshow
	move	d3,UM_Class(a1)
.noshow

	cmp.w	#USclass_Hide,d3
	bne.s	.noclosewin
	bsr	Closewin
.noclosewin

	move.l	mn_replyport(a1),d0
	cmp.l	MSGport(a5),d0		 ;eine von uns abgeschickte Message ?
	beq.s	.noreply
	jsr	_LVOreplymsg(a6)
.noreply
	
	cmp.l	#USM_Userprg,d7	;Message zur?ckbekommen ?
	bne.s	.nomy

	move.b	#0,messagesent(a5)	;merken,da? Message-Struktur frei
	bra	.mess	;eventl. Message vom Esagle auswerten

.nomy
	cmp.l	#USM_Eagleplayer,d7	;Hat EaglePlayer was zu melden ?
	bne.w	.noeagle		;n?
.mess

;	lea	begin(pc),a0
;	move.l	eus_epbase(a0),a0
;	move.b	EPG_Wordreserved7(a0),d0
;	beq	.killit

;	move.l	d6,an_structadr(a5)
.refre
	cmp.w	#USclass_Dummy,d3	;leere Message,nur neue Analyzer-
	beq.w	.ma			;adresse bekommen
	cmp.w	#USclass_killmodule,d3
	bne.s	.nokill

	move	#-1,Currentnumvoc(a5)
	bsr	MakeLV

.nokill
	cmp.w	#USclass_zipwin,d3
	bne.s	.nozip
.setzip
	move.l	4.w,a6
	move.l	winhandle(a5),d0
	beq	.ma
	move.l	d0,a0
	move.w	wd_height(a0),d0
	cmp.w	myhoehe1(a5),d0
	beq.w	.ma		;Window schon klein
.zip
	move.w	d0,-(sp)

	move.l	intsave(a5),a6
	jsr	_LVOzipwindow(a6)

	move.w	(sp)+,d4

	move.l	winhandle(a5),a2
.Zipwait:
	moveq	#2,d1
	move.l	_Dos(a5),a6
	jsr	_LVODelay(a6)

	cmp.w	wd_height(a2),d4
	beq.s	.zipwait

	bra	.ma
.nozip
	cmp.w	#USclass_unzipwin,d3
	bne.s	.nounzip

.clearzip
	move.l	4.w,a6
	move.l	winhandle(a5),d0
	beq	.ma
	move.l	d0,a0
	move.w	wd_height(a0),d0
	cmp.w	myhoehe1(a5),d0
	beq.s	.zip
	bra	.ma
.nounzip
	cmp.w	#USclass_exit,d3
	beq	.endme

;	cmp.w	#USclass_Show,d3
;	beq.s	.win
;	cmp.w	#USclass_Activate,d3
;	bne.w	.nowin
;.win
;	bsr	Openwin
;	bra	.ma
;.nowin
	cmp.w	#USclass_NewConfig,d3	;is there a new config
	bne.w	.nonewcon		;no

	lea	begin(pc),a0		;start of program (EUS_-Structure)
	move.w	eus_flags(a0),d0	;get Flags
	btst	#eusf_openwin,d0	;Window open ?
	beq.s	.nowino
	bsr	openwin			;open it
	bra.s	.weio
.nowino
	bsr	closewin		;else close it
.weio
	lea	begin(pc),a0		;start of program (EUS_-Structure)
	move.l	winhandle(a5),d0	;windowaddress
	beq.s	.nothingtomove		;no Window , no movement

	move.l	d0,a1			;Window
	move.l	wd_wscreen(a1),a2	;Parent Screen

	move.w	eus_winx(a0),d0		;new X-position
	move.w	d0,d1			;save
	blt.s	.oldx			;less than 0 -> stop

	add.w	wd_width(a1),d1		;new position + Window-width
	cmp.w	sc_width(a2),d1		;bigger than Screenwidth ?
	bhi.s	.oldx			;then old X-position (Delta = 0)

	sub.w	wd_leftedge(a1),d0	;new - old X-position = Delta X
	bra.s	.gety
.oldx
	moveq	#0,d0			;Delta X = 0
.gety
	move.w	eus_winy(a0),d1		;new Y-position

	move.w	d1,d2			;save
	blt.s	.oldy			;less than 0 -> stop and take old

	add.w	wd_height(a1),d2	;new position + window-height
	cmp.w	sc_height(a2),d2	;bigger than Screenheight ?
	bhi.s	.oldy

	sub.w	wd_topedge(a1),d1	;new - old Y-position = Delta Y
	bra.s	.setdeltas
.oldy	moveq	#0,d1

.setdeltas
	move.l	intsave(a5),a6		;Intuition
	move.l	a1,a0			;Winhandle
	jsr	_LVOmovewindow(A6)
.nothingtomove
	lea	begin(pc),a0
	move.w	eus_flags(a0),d0
	btst	#eusf_zipwin,d0		;Zipwindow ?
	bne	.setzip			;yes
	bra	.clearzip		;no
.nonewcon
	bra	.ma
.noeagle
	bra	.ma
	else
	* USERPROGRAM
.AN_Msg
	bra	.ma
	endc

*********************** Finde Tag ?ber EP Funktion ***************************
FindTag
	movem.l	a1/a5,-(sp)
	move.l	myEPBase(pc),a5
	move.l	EPG_FindTag(a5),a1
	jsr	(a1)
	movem.l	(sp)+,a1/a5
	rts
*****************************************************************
allocmsg
	movem.l	d1-a6,-(sp)
	move.l	4.w,a6
	moveq	#UM_Sizeof,d0
	move.l	#$10001,d1
	jsr	_LVOallocmem(A6)
	movem.l	(sp)+,d1-a6
	tst.l	d0
	rts
*****************************************************************
freemsg
;	moveq	#UM_sizeof,d0

	moveq	#0,d0
	move.w	mn_length(a1),d0
	add.w	#20,d0
	move.l	4.w,a6
	jsr	_LVOfreemem(A6)
	rts
******************************************************************************
*                       Ausgabe von Error-Messages                           *
*                 Input: a0 -> Zeiger auf Ausgabestring                      *
******************************************************************************
MERROR:
	movem.l	d0/a2-a3,-(sp)
	suba.l	a3,a3			;*Arguments
	lea	my_easygadget(pc),a2
	bsr	Errorout
	movem.l	(sp)+,d0/a2-a3
	rts
Errorout
	movem.l	d1-a6,-(sp)
	move.l	Intsave(a5),a6
	lea	my_easystruct(a5),a1

	move.l	a0,es_textformat(a1)

	move.l	a2,es_gadgetformat(a1)

	lea	my_easytitle(pc),a0
	move.l	a0,es_Title(a1)
	move.l	winhandle(a5),a0	;*Window
	suba.l	a2,a2			;*IDCMP_ptr
	jsr	_LVOEasyRequestArgs(a6)			;!! durch Reqtools
							;ersetzen
	movem.l	(sp)+,d1-a6	
	rts

;======================= Load 14 Bit Calibration Table ==============================================
LoadCalibration:
	movem.l	d0-a6,-(sp)
	moveq	#-1,d7
	
	move.l	4.w,a6
	lea	dos(pc),a1
	moveq	#33,d0
	jsr	_LVOopenlibrary(A6)
	tst.l	d0
	beq	.error
	move.l	d0,a6
	lea	.loadname(pc),a1
	move.l	a1,d1
	move.l	#1005,d2			;mode_oldfile
	jsr	_LVOopen(a6)
	move.l	d0,d1				;Handle
	beq.s	.closelib
	move.l	d0,d6

	lea	AdditiveArray(pc),a0
	move.l	a0,d2				;Adr
	move.l	#256,d3				;len
	jsr	_LVOread(A6)
	move.l	d0,d7

	move.l	d6,d1
	jsr	_LVOclose(a6)
.closelib
	move.l	a6,a1
	move.l	4.w,a6
	jsr	_LVOcloselibrary(A6)
	cmp.l	#256,d7
	beq.s	.quit
.error
	lea	AdditiveArray(pc),a0
	move	#255-1,d0
.cop	move.b	#$5f,(a0)+
	dbf	d0,.cop
	move.b	#$7f,(a0)
.quit
	movem.l	(sp)+,d0-a6
	moveq	#0,d0
	rts
.loadname:	dc.b	"ENV:Cybersound/Sounddrivers/14Bit_Calibration",0

;============= Save Player Config Routine ===========================
SaveConfig:
	movem.l	d0-a6,-(sp)
	
	move.l	4.w,a6
	lea	dos(pc),a1
	moveq	#33,d0
	jsr	_LVOOpenLibrary(A6)	;Open DOS
	tst.l	d0
	beq	.error
	move.l	d0,a6

	move.l	myepbase(pc),a0
	move.l	EPG_ConfigDirArrayPtr(a0),d1
	moveq	#-2,d2			;access_read
	jsr	_LVOLock(a6)
	tst.l	d0
	beq	.nolock			;Path not found
	move.l	d0,d1
	jsr	_LVOCurrentDir(A6)
	move.l	d0,d7

	lea	savename(pc),a1
	move.l	a1,d1
	move.l	#1006,d2		;MODE_NEWFILE
	jsr	_LVOOpen(A6)
	move.l	d0,d6
	beq.s	.fail

	move.l	d0,d1				;Handle
	lea	ConfigDatas(A5),a1
	move.l	a1,d2				;Adr
	move.l	#ConfigDatasEND-ConfigDatas,d3	;len
	jsr	_LVOWrite(a6)

	move.l	d6,d1
	jsr	_LVOClose(a6)
.fail
	move.l	d7,d1
	jsr	_LVOCurrentdir(A6)
	move.l	d0,d1
	jsr	_LVOUnLock(A6)
.nolock
	move.l	a6,a1
	move.l	4.w,a6
	jsr	_LVOCloseLibrary(A6)
.error
	movem.l	(Sp)+,d0-a6
	rts
;============= Load Player Config Routine ===========================
Loadconfig:
	movem.l	d0-a6,-(sp)

	move.l	4.w,a6
	lea	dos(pc),a1
	moveq	#33,d0
	jsr	_LVOOpenLibrary(A6)	;Open DOS
	tst.l	d0
	beq	.error
	move.l	d0,a6

	move.l	myepbase(pc),a0
	move.l	EPG_ConfigDirArrayPtr(a0),d1
	moveq	#-2,d2			;access_read
	jsr	_LVOLock(a6)
	tst.l	d0
	beq	.nolock			;Path not found
	move.l	d0,d1
	jsr	_LVOCurrentDir(A6)
	move.l	d0,d7

	lea	savename(pc),a1
	move.l	a1,d1
	move.l	#MODE_Oldfile,d2		;MODE_NEWFILE
	jsr	_LVOOpen(A6)
	move.l	d0,d6
	beq.s	.fail

	move.l	d0,d1				;Handle
	lea	ConfigDatas(A5),a1
	move.l	a1,d2				;Adr
	move.l	#ConfigDatasEND-ConfigDatas,d3	;len
	jsr	_LVORead(a6)

	move.l	d6,d1
	jsr	_LVOClose(a6)
.fail
	move.l	d7,d1
	jsr	_LVOCurrentdir(A6)
	move.l	d0,d1
	jsr	_LVOUnlock(A6)
.nolock
	move.l	a6,a1
	move.l	4.w,a6
	jsr	_LVOCloseLibrary(A6)
.error
	movem.l	(sp)+,d0-a6
	rts
	cnop	0,4
********************************************************************
*               initialize new output hardware                     *
* input: A0 - pointer to new OHW structure                         *
*        A5 - "datas"                                              *
********************************************************************
SetOutputHW:
	;TODO: shutdown old output HW, as soon as multiple instances are implemented
	;
	push	all
	move.l	a0,D0			;new supplied output ?
	bne.s	.new
	tst.l	Amp_OHWStr		;already initialized and no new one given ? -> keep last setting
	bne.s	.keep
	lea	Ampout_DEFAULT,A0	;def: use Paula directly (TODO: check for Draco)

.new:
	move.l	OHW_Setup(a0),d0	
	beq.s	.nosetup
	move.l	d0,a1
	jsr	(a1)			;call setup
.nosetup	
	move.l	OHW_Check(a0),d0
	beq.s	.alwaysavail		;no check -> always available
	move.l	d0,a1
	jsr	(a1)
	moveq	#1,d1
	cmp.l	d0,d1
	beq.w	.alwaysavail		;

	;error -> shutdown
	move.l  OHW_Shutdown(a0),d0
	beq.s	.noshutdown
	move.l	d0,a1
	jsr	(a1)
.noshutdown

	; ERROR REQUESTER HERE

	; TODO: branch to error and keep (hopefully working) old one
.alwaysavail:
	move.l	a0,Amp_OHWStr
	move.l	#OHWS_INIT,Amp_OHWStatus	;totally new status, only the init flag set
.keep:

	;list of mixing frequencies
	bsr	Amp_GetMixRates

	moveq	#0,d5
	move.l	d0,Amp_OHWMixrates	;available mixing rates
	beq.s	.nomixrates

	move.l	d0,a0
	moveq	#-1,d5
.mixrates_count:
	addq.l	#1,d5
	tst.l	(a0)+
	bne.s	.mixrates_count

.nomixrates:

	move.l	d5,Amp_OHWMixrateCount
	move.l	d1,d5			;mixing frequency
	moveq	#MIXSL_ListSize,d1
	bsr	HandleMixingSlider	;update mixing slider with supplied frequency
	;
	pop	all
	rts
********************************************************************
*               initialize default preferences                     *
********************************************************************
Makedefaults:
	movem.l	d0-a6,-(sp)


	lea	Prefslist(a5),a0
	tst.w	MC68040on(a5)
	bne.s	.def68040
	; 68020+ defaults
	move.l	#28603,d0
	moveq	#10,d1
	move.w	#$0000+SamLEN/10,d2		;Effect (Stereo+Delay 350bytes)
	moveq	#0,d4				;Flags (not disabled, use nearest neighbor resampling)
	move.w	#500,d6				;subtract this every 4 additional channels
	bra.s	.setdefs

.def68040:
	; 68040+ defaults
	move.l	#28603,d0			;Rate
	moveq	#10,d1				;VBoost "1.0"
	move.w	#$0000+SamLEN/10,d2		;Effect (Stereo+Delay 350bytes)
	moveq	#0,d4				;Flags
	moveq	#0,d6				;all channel counts same speed
.setdefs:
	moveq	#MaxVoices-1,d3
.defaults:
	move.l	d0,P_Mixrate(a0)		;mixrate.l
	move.w	d1,P_Vboost(a0)                 ;vboost.w
	move.w	d2,P_Surround(a0)               ;effect.w
	move.w	d4,P_Flags(a0)                  ;flags.w
	ifne	AHIOUT
	move.l	#AHI_DEFAULT_ID,P_AHIMode(a0)	;
	endc

	lea	P_Sizeof(a0),a0
	addq.w	#1,d5
	and.w	#3,d5
	bne.s	.no
	addq.w	#1,d1
	sub.l	d6,d0				;500 Hz per 4 steps, fast CPUs should be able to handle this
.no
	dbf	d3,.defaults

	movem.l	(sp)+,d0-a6
	rts
*****************************************************************************
*                        Gadget up routine                                  *
* A1 - current gadget structure                                             *
* D4 - Gadget ID                                                            *
* D5 - argument (IM_Code)                                                   *
*****************************************************************************
TestGadgets:
	cmp.w	#GD_Listview,d4		;Listview-Gadget
	beq.w	.id_listview
	cmp.w	#GD_VBOOST,d4		;Volume Boost Slider-Gadget
	beq.w	.id_vboost
	cmp.w	#GD_Samplefreq,d4
	beq.w	.id_samplefreq
	cmp.w	#GD_MixingSlider,d4
	beq.w	.id_mixingslider
	cmp.w	#GD_Playmode,d4		;Playmode Slider
	beq.w	.id_playmode
	cmp.w	#GD_Surlevel,d4
	beq.w	.id_surrounddelay
	cmp.w	#GD_MixMode,d4
	beq.w	.id_mixmode
	cmp.w	#GD_AHIMode,d4
	beq.w	.id_ahimode
	cmp.w	#GD_AHIGlobal,d4
	beq.w	.id_ahiglobal
	bra.w	.endit

;-------------------- global button in AHI mode -----
.id_ahiglobal:
	move.l	d4,d0
	bsr	_GADR
	move.l	a0,d0
	beq.w	.endit

	move.l	gadsave(A5),a6		;lib
	sub.l	a2,a2			;req

	lea	temp(a5),a1
	lea	mytags(a5),a3		;tags
	move.l	#GTCB_Checked,(A3)+
	move.l	a1,(a3)+
	clr.l	(A3)+
	lea	mytags(a5),a3		;tags

	move.l	winhandle(A5),a1	;win
	jsr	_LVOGT_Getgadgetattrsa(A6)

	move.l	temp(a5),d0
	move.b	d0,onglobal(a5)
	bsr	SetVoic

	;TODO: update menu

	bra	.endit
;-------------------- AHI mode button ---------------
.id_ahimode:
	ifne	AHIOUT
	bsr	AHIAMP_Configure
	endc
	bra	.endit
;-------------------- listview ----------------------
.id_listview:
	move.w	SelectedW(A5),d6	;aktuell angeclickter Modus
	move.w	d5,SelectedW(A5)

	cmp.w	d5,d6			;kein Doppelklick
	beq	.testdoub

	move.l	d2,Seconds(a5)
	move.l	d3,Micros(a5)

	bsr	UpdateList
	bra	.endit
.testdoub:
	move.l	Seconds(a5),d0
	move.l	Micros(a5),d1
	move.l	d2,Seconds(a5)
	move.l	d3,Micros(a5)
	move.l	intsave(a5),a6
	jsr	_LVOdoubleclick(a6)
	cmp.l	#1,d0
	bne.w	.disp

	clr.l	Seconds(a5)
	
	mulu	#P_Sizeof,d6
	lea	Prefslist(a5),a0
	move.w	P_Flags(A0,d6.w),d0		;Flags.w (Bit0=disable)
	eor.w	#PBB_Disable,d0
	move.w	d0,P_Flags(a0,d6.w)
.disp:
	bsr	UpdateList
	bra	.endit
;-------------------- vboost ----------------------
.id_vboost:
	sf	onvboostslider(A5)

	tst.b	onglobal(A5)
	bne.s	.sli2global

	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist+P_Vboost(a5),a0
	move.w	d5,(A0,d0.w)
	bra.s	.sli2up
.sli2global:
	lea	Prefslist+P_Vboost(a5),a0
	moveq	#MaxVoices-2,d0
.sli2all:
	move.w	d5,(A0)
	add.w	#P_Sizeof,a0
	dbf	d0,.sli2all
.sli2up:
	; update current boost only when the correct mode is selected
	tst.b	onplay
	beq.s	.boost_nosele
	move.w	numleft,d0
	add.w	numright,d0
	subq.w	#2,d0
	cmp.w	SelectedW(a5),d0
	bne.s	.boost_nosele

	move	d5,d0			;boost * 10
	bsr	Amp_SetBoost		;bsr	do14tab
.boost_nosele
;	bsr	UpdateList		;Sp?ter nur Listview update !!!!

	bsr	SetVBoost
	bsr	MakeLV			;Listview in Ordnung bringen
	bra	.endit
;-------------------- sample frequency/mixing slider ----------------------
.id_samplefreq:
	move.l	gg_specialinfo(a1),a0	;Stringinfo Struktur
	move.l	si_Longint(A0),d1	;Wert des Gadgets
	cmp.l	#8000,d1
	bgt.s	.ok1
	move.l	#8000,d1
.ok1:
	cmp.l	#57000,d1		;more just doesn't make sense, at least with amiga hardware
	ble.s	.ok2			;any 96 kHz hardware for Amiga ?
	move.l	#57000,d1		;
.ok2:
	move.l	d1,d5			;mixing frequency
	moveq	#MIXSL_ExplicitFreq,d1
	bsr	HandleMixingSlider	;update mixing slider with supplied frequency
	bra	.endit

.id_mixingslider:
	sf	onmixingslider(A5)

	moveq	#MIXSL_SliderUp,d1	;knob released
	bsr	HandleMixingSlider	;D5: position
	bra	.endit

;-------------------- playmode mono-stereo-surround  ----------------------
.id_playmode:
	sf	onPlaymodeslider(A5)
					;d5: -64 <-> 64
	tst.b	onglobal(A5)
	bne.s	.slipmodeG

	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist(a5),a0
	move.b	d5,P_Surround(A0,d0.w)
	bra.s	.sliPMODEup
.slipmodeG:
	lea	Prefslist+P_Surround(a5),a0
	moveq	#MaxVoices-2,d0
.sliPMall:
	move.b	d5,(A0)			;Playmode
	add.w	#P_Sizeof,a0
	dbf	d0,.sliPMall
.sliPMODEup:
;	bsr	UpdateList		;Sp?ter nur Listview update !!!!
	bsr	MakeLV

	move.b	d5,d0
	ext.w	d0
	ext.l	d0
	bsr	Amp_SetPlayMode

	bra	.endit
;-------------------- surround delay  ----------------------
.id_surrounddelay:
	sf	onSurlevslider(A5)
					;d5: -64 <-> 64
	tst.b	onglobal(A5)
	bne.s	.slisurleG

	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist(a5),a0
	move.b	d5,P_SurroundDelay(A0,d0.w)
	bra.s	.slisurlevUP
.slisurleG:
	lea	Prefslist+P_SurroundDelay(a5),a0
	moveq	#MaxVoices-2,d0
.sliSLall:
	move.b	d5,(A0)			;Playmode
	add.w	#P_Sizeof,a0
	dbf	d0,.sliSLall
.slisurlevUP:
;	bsr	UpdateList		;Sp?ter nur Listview update !!!!
	bsr	MakeLV

	move.w	d5,d0
	bsr	setPrefetch
	bra	.endit

;-------------------- surround delay  ----------------------
.id_mixmode:
	cmp.w	#PBB_MixMax,d5		;ignore wrong values
	bgt.s	.endit			;

	moveq	#0,d6
	tst.w	d5
	beq.s	.idmixmode_f
	moveq	#PBB_SincMix,d6
	cmp.w	#1,d5
	beq.s	.idmixmode_f
	moveq	#PBB_SincMixB,d6
.idmixmode_f:				; calculate actual mode

	tst.b	onglobal(A5)
	bne.s	.idmixmode_global

	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist(a5),a0

	move	P_Flags(a0,d0.w),d5
	and.w	#$ffff-PBB_MixMask,d5
	or.w	d6,d5
	move	d5,P_Flags(a0,d0.w)
	bra.s	.idmixmode_update

.idmixmode_global:
	lea	Prefslist(a5),a0
	moveq	#MaxVoices-2,d0
.idmixmode_updateloop:
	move	P_Flags(a0),d5
	and.w	#$ffff-PBB_MixMask,d5
	or.w	d6,d5
	move	d5,P_Flags(a0)
	add.w	#P_Sizeof,a0
	dbf	d0,.idmixmode_updateloop

.idmixmode_update:
	;update requires 
	; - new volume table(s)
	; - mixing address(es)
	bsr	MakeLV
	;
	bsr	setMixRoutine

	bsr	Makevoltable

	move.l	usedmode,d0
	beq.w	.idmixmode_inactive
	move.l	d0,A1
	move.w	P_Vboost(a1),d0
	bsr	Amp_SetBoost		;bsr	do14tab
.idmixmode_inactive:
	bra	.endit
.endit
.ma
	rts
*****************************************************************************
*                       SliderGadgets auswerten                             *
*****************************************************************************
TestDOWNGadgets
	move.l	4.w,a6

	ifne	V37Gadtools
	moveq	#-1,d0
	else
	cmp.w	#39,lib_version(A6)
	endc

	blt	.endit

	cmp.w	#GD_MixingSlider,d4
	bne.s	.nosli

	st	onmixingslider(A5)
.nosli
	cmp.w	#GD_VBoost,d4
	bne.s	.nosli2

	st	onvboostslider(A5)
.nosli2
	cmp.b	#GD_Playmode,d4
	bne.s	.nosli3

	st	onPlaymodeslider(A5)
.nosli3

	cmp.w	#GD_Surlevel,d4
	bne.s	.nosurlev

	st	onSurlevslider(A5)
.nosurlev

.endit
	rts

*****************************************************************************
*                 Slider Down Updates                                       *
*****************************************************************************
UpdateDownGadgets
	move.l	4.w,a6
	cmp.w	#39,lib_version(A6)
	blt	.endit

	tst.b	onmixingslider(A5)
	beq.s	.nosli1
	
	moveq	#GD_MixingSlider,d0
	bsr	_GADR			;Gadgetadresse in A0
	move.l	a0,d0
	beq.s	.nosli1
	move.l	a0,a1
	bsr	.testup

	moveq	#MIXSL_SliderDown,d1	;down
	bsr	HandleMixingSlider	;D5: position
	bra.w	.rts
.nosli1
	tst.b	onvboostslider(A5)
	beq.s	.nosli2
	
	moveq	#GD_VBoost,d0
	bsr	_GADR			;Gadgetadresse in A0
	move.l	a0,a1
	bsr	.testup

	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist+P_Vboost(a5),a0
	move.w	d5,(A0,d0.w)

	push	d5
	bsr	SetVBoost
	pop	d5			;boost * 10

	; update current boost only when the correct mode is selected
	tst.b	onplay
	beq.s	.boost_nosele
	move.w	numleft,d0
	add.w	numright,d0
	subq.w	#2,d0
	cmp.w	SelectedW(a5),d0
	bne.s	.boost_nosele

	move.l	d5,d0
	bsr	Amp_SetBoost		;bsr	do14tab
.boost_nosele
	bra.s	.rts
.nosli2
	tst.b	onPlaymodeslider(A5)
	beq.s	.nosli3

	moveq	#GD_Playmode,d0
	bsr	_GADR			;Gadgetadresse in A0
	move.l	a0,a1
	bsr	.testup

	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist+P_Surround(a5),a0
	move.b	d5,(A0,d0.w)

	move.b	d5,d0
	ext.w	d0
	ext.l	d0
	bsr	Amp_SetPlayMode
	bra.s	.rts
.nosli3
	tst.b	onSurlevslider(A5)
	beq.s	.nosurlev

	moveq	#GD_Surlevel,d0
	bsr	_GADR			;Gadgetadresse in A0
	move.l	a0,a1
	bsr	.testup

	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist+P_SurroundDelay(a5),a0
	move.b	d5,(A0,d0.w)

	move.w	d5,d0
	bsr	setPrefetch
	bra.s	.rts
.nosurlev

.endit
	moveq	#0,d0
	rts
.rts
	moveq	#1,d0
	rts
.testup
	movem.l	d1-d4/d6-a6,-(sp)

	move.l	a1,d7
	move.l	_dos(a5),a6
	moveq	#1,d1
	jsr	_LVOdelay(a6)
	move.l	d7,a1

	move.l	a1,a0			;gad
	move.l	gadsave(A5),a6		;lib
	sub.l	a2,a2			;req

	lea	numboo(a5),a1
	lea	mytags(a5),a3		;tags
	move.l	#GTSL_Level,(A3)+
	move.l	a1,(a3)+
	clr.l	(A3)+
	lea	mytags(a5),a3		;tags

	move.l	winhandle(A5),a1	;win

	jsr	_LVOGT_Getgadgetattrsa(A6)

	move.l	numboo(a5),d5

	movem.l	(sp)+,d1-d4/d6-a6
	rts


*****************************************************************************
*                 Handle Mixing Rate Slider                                 *
* Input:                                                                    *
*        D1 - MIXSL_*                                                       *
*              0=up,                                                        *
*              1=down,                                                      *
*              2=explicit mixing frequency                                  *
*              3=update number of available frequencies                     *
*              4=update level from current prefs                            *
*        D5 - current level (index) in modes 0,1, frequency in mode2        *
*              number of entries in mode 3                                  *
*        A5 - globals                                                       *
*****************************************************************************
HandleMixingSlider:
	push	d0-a6

	cmp.w	#MIXSL_SetLevel,d1
	beq.w	.updateLevel

	cmp.w	#MIXSL_ListSize,d1
	beq.s	.newlimits

	cmp.w	#MIXSL_SliderDown,d1
	beq.s	.down

	; actions when slider knob is released
	cmp.w	#MIXSL_ExplicitFreq,d1
	bne.s	.noexplicit
	move.l	d5,d1
	bra.s	.explicit
.noexplicit:
	bsr	.slev_to_mixrate	;slider level (d5) -> mixrate (d1)
.explicit:
	bsr	.setlist	;update list (d1=new value)

	bsr	MakeLV
	bsr	SetLevel
	bra.s	.ret

	; actions when slider knob is pressed
.down:
	bsr	.slev_to_mixrate	;slider level (d5) -> mixrate (d1)

	bsr	.setlist
.ret:
	bsr	SetRate
	bsr	CalcValues
	bsr	setMixRoutine		;when highboost is on, set correct filter as well

	move.l	usedmode,d0
	beq.w	.idmixmode_inactive
	move.l	d0,A1
	move.w	P_Vboost(a1),d0
	bsr	Amp_SetBoost		;bsr	do14tab
.idmixmode_inactive:

.ret_noaction:
	pop	d0-a6
	rts

	; new number of available sampling rates (d5)
.newlimits:
	lea	mytags(a5),a3		;tags

	ifne	Mixfreqlist		;use list instead of free choice (you still can write numbers to the textfield...)
		move.l	#GTSL_Min,(A3)+
		move.l	#0,(A3)+
		move.l	#GTSL_Max,(A3)+
		subq.l	#1,d5
		bge.s	.limitok
		moveq	#0,d5
.limitok:
		move.l	d5,(A3)+
	else
		move.l	#GTSL_Min,(A3)+
		move.l	#400,(A3)+
		move.l	#GTSL_Max,(A3)+
		move.l	#5800,(A3)+
	endc
	clr.l	(A3)+

	move.l	winhandle(A5),d0
	beq.s	.ret_noaction
	moveq	#GD_MixingSlider,d0
	bsr	_GADR			;Gadgetadresse in A0
	move.l	a0,d0
	beq.s	.ret_noaction

	lea	mytags(a5),a3
	move.l	winhandle(A5),a1
	suba.l	a2,a2
	move.l	gadsave(A5),a6
	jsr	_LVOGT_Setgadgetattrsa(A6)
	bra.s	.ret_noaction

.updatelevel
	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist+P_Mixrate(a5),a0
	move.l	(A0,d0.w),d1		;Mixrate
	bsr.w	.mixrate_to_slev

	move.l	winhandle(A5),d0
	beq.s	.ret_noaction
	moveq	#GD_MixingSlider,d0
	bsr	_GADR			;Gadgetadresse in A0
	move.l	a0,d0
	beq.s	.ret_noaction

	lea	mytags(a5),a3
	move.l	#GTSL_Level,(A3)+
	move.l	d1,(a3)+
	clr.l	(A3)+
	lea	mytags(a5),a3
	move.l	winhandle(A5),a1
	suba.l	a2,a2
	move.l	gadsave(A5),a6
	jsr	_LVOGT_Setgadgetattrsa(A6)
	bra.w	.ret_noaction

	;D1: new mixrate to put in preferences array (either single or global)
.setlist:
	push	d0/a0
	tst.b	onglobal(A5)
	bne.s	.sli1global
	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist+P_Mixrate(a5),a0
	move.l	d1,(A0,d0.w)
	bra.s	.sli1up
.sli1global:
	lea	Prefslist+P_Mixrate(a5),a0
	moveq	#MaxVoices-2,d0
.sli1all:
	move.l	d1,(A0)
	add.w	#P_Sizeof,a0
	dbf	d0,.sli1all
.sli1up:
	pop	d0/a0
	rts

; convert slider level to mixing rate
.slev_to_mixrate:
	push	d0/d5/a0
	ifne	Mixfreqlist
		move.l	#28603,d1		;def: something useful
		move.l	Amp_OHWMixrates,d0	;available mixing rates
		beq.s	.slev_to_mixrate_nomixrates
		move.l	d0,a0
.slev_to_mixrate_loop:
		tst.l	(a0)
		beq.s	.slev_to_mixrate_nomixrates	;keep either default or last valid mixrate
		move.l	(a0)+,d1		;load valid mixrate 
		dbf	d5,.slev_to_mixrate_loop
.slev_to_mixrate_nomixrates:
	else
		move.w	#10,d1		;old code: slider level * 10
		mulu	d5,d1
	endc
	pop	d0/d5/a0
	rts

; convert mixing rate to approximate slider level
.mixrate_to_slev:
	push	d0/d2-d5/a0
	ifne	Mixfreqlist
		move.l	d1,d5			;
		moveq	#0,d1			;def: lowest

		move.l	Amp_OHWMixrates,d0	;available mixing rates
		beq.s	.mixrate_to_slev_nomixrates
		move.l	d0,a0
		
		moveq	#-1,d3
		lsr.l	#1,d3			;7fffffff
		moveq	#-1,d4

		;check all and select lowest difference between chosen mixrate and available one
.mixrate_to_slev_loop:
		addq.l	#1,d4			;current index
		tst.l	(a0)
		beq.s	.mixrate_to_slev_done	
		move.l	(a0)+,d2		;load valid mixrate 
		sub.l	d5,d2			;diff
		bge.s	.mixrate_to_slev_hi
		neg.l	d2			;abs(diff)
.mixrate_to_slev_hi:
		cmp.l	d2,d3
		blt.s	.mixrate_to_slev_nobetter
		move.l	d4,d1			;new best index
		move.l	d2,d3			;new best diff
.mixrate_to_slev_nobetter:
		bra.s	.mixrate_to_slev_loop
.mixrate_to_slev_done:
.mixrate_to_slev_nomixrates:
	else
		divu	#10,d1			;old code: slider level = mixrate/10
		and.l	#$ffff,d1
	endc
	pop	d0/d2-d5/a0
	rts





*****************************************************************************
*                 aktuelle Anzahl von Stimmen selektieren                   *
* Input: d0 = Num Voices                                                    *
*****************************************************************************
SelectEntry:
	movem.l	d0-a6,-(Sp)
	lea	datas,a5

	subq.w	#2,d0
	bpl.s	.hi
	moveq	#0,d0
.hi	move.w	d0,SelectedW(A5)	;aktuell angeclickter Modus
	move	d0,Currentnumvoc(a5)

	bsr	UpdateList

	tst.l	winhandle(A5)
	beq.s	.err

	moveq	#GD_LISTVIEW,d0
	bsr	_GADR			;Gadgetadresse in A0
	lea	mytags(a5),a3
	move.l	#GTLV_Makevisible,(a3)+
	move.l	selectedL(a5),(a3)+
	clr.l	(a3)
	lea	mytags(a5),a3

	move.l	winhandle(A5),a1
	suba.l	a2,a2
	move.l	gadsave(A5),a6
	jsr	_LVOGT_Setgadgetattrsa(A6)
.err
	movem.l	(Sp)+,d0-a6
	rts
*****************************************************************************
*          Einsprung, was alles up to Date gebracht werden mu?              *
*****************************************************************************
;;
UpdateList:
	tst.l	winhandle(a5)
	beq.s	.end
	movem.l	d0-a6,-(Sp)

	bsr	MakeLV			;Listview in Ordnung bringen
	bsr	SetVoic
	bsr	SetVBoost
	bsr	SetVBoostSlider
	bsr	SetRate

	move.l	Amp_OHWMixrateCount,d5
	moveq	#MIXSL_ListSize,d1
	bsr	HandleMixingSlider	;update mixing slider with supplied frequency
	bsr	SetLevel		;redirects to HandleMixingSlider as well

	bsr	SetPlaymodeSlider
	bsr	SetSurlevelSlider
	bsr	SetCycle		;Mixing Mode Cycle
	movem.l	(sp)+,d0-a6
.end
	rts
	; update Mixing Mode
SetCycle:
	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist+P_Flags(a5),a0
	move.w	(A0,d0.w),d0
	moveq	#2,d1
	btst	#PBF_SincMixB,d0
	bne.s	.mod
	moveq	#1,d1
	btst    #PBF_SincMix,d0
	bne.s	.mod
	moveq	#0,d1	; no flags, use default
.mod
	moveq	#GD_MixMode,d0
	bsr	_GADR
	lea	mytags(a5),a3
	move.l	#GTCY_Active,(A3)+
	move.l	d1,(a3)+
	clr.l	(A3)+
	bra	SetGAD
SetPlaymodeSlider:
	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist(a5),a0
	move.b	P_Surround(A0,d0.w),d1	;Playmode: -64 .. 64
	ext.w	d1
	ext.l	d1

	moveq	#GD_Playmode,d0
	bsr	_GADR			;Gadgetadresse in A0
	lea	mytags(a5),a3
	move.l	#GTSL_Level,(A3)+
	move.l	d1,(a3)+
	clr.l	(A3)+
	bra	SetGAD
SetSurlevelSlider
	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist(a5),a0
	move.b	P_SurroundDelay(A0,d0.w),d1	;Playmode: -64 .. 64
	ext.w	d1
	ext.l	d1

	moveq	#GD_Surlevel,d0
	bsr	_GADR			;Gadgetadresse in A0
	lea	mytags(a5),a3
	move.l	#GTSL_Level,(A3)+
	move.l	d1,(a3)+
	clr.l	(A3)+
	bra	SetGAD

	;update mixing slider level
SetLevel:
	moveq	#MIXSL_SetLevel,d1
	bsr	HandleMixingSlider	;no argument, get value from preferences
	rts

	;update mixing rate display
SetRate:
	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist+P_Mixrate(a5),a0
	move.l	(A0,d0.w),d1

	moveq	#GD_Samplefreq,d0
	bsr	_GADR			;Gadgetadresse in A0
	lea	mytags(a5),a3
	move.l	#GTIN_Number,(A3)+
	move.l	d1,(a3)+
	clr.l	(A3)+
	bra	SetGAD

SetVBoost:
	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist+P_Vboost(a5),a0
	moveq	#0,d1
	move.w	(A0,d0.w),d1

	moveq	#GD_NumVBoost,d0
	bsr	_GADR			;Gadgetadresse in A0
	lea	mytags(a5),a3
	move.l	#GTTX_Text,(A3)+
	lea	numboo(a5),a1
	move.l	a1,(a3)+
	clr.l	(A3)+

	divu	#100,d1
	add.b	#` `,d1
	cmp.b	#` `,d1
	beq.s	.set
	add.b	#"0"-" ",d1
.set
	move.b	d1,(a1)+

	clr.w	d1
	swap	d1
	divu	#10,d1
	add.b	#`0`,d1
	move.b	d1,(a1)+

	swap	d1
	move.b	#`.`,(a1)+
	add.b	#`0`,d1
	move.b	d1,(a1)+

	clr.b	(a1)
	bra	SetGAD
SetVBoostSlider
	move.w	SelectedW(a5),d0
	mulu	#P_Sizeof,d0
	lea	Prefslist+P_Vboost(a5),a0
	moveq	#0,d1
	move.w	(A0,d0.w),d1		;VolBoost

	moveq	#GD_VBoost,d0
	bsr	_GADR			;Gadgetadresse in A0
	lea	mytags(a5),a3
	move.l	#GTSL_Level,(A3)+
	move.l	d1,(a3)+
	clr.l	(A3)+
	bra	SetGAD
SetVoic:
	moveq	#GD_VOICES,d0
	bsr	_GADR			;Gadgetadresse in A0
	lea	mytags(a5),a3
	lea	NumVoc(A5),a1
	move.l	#GTTX_Text,(A3)+
	move.l	a1,(a3)+
	clr.l	(A3)+

	;---------- hier noch test auf Global -------------
	tst.b	onglobal(a5)
	beq.s	.ori

	move.b	#`G`,(a1)+
	clr.b	(a1)
	bra.s	setgad
.ori
	move.l	SelectedL(a5),d0
	addq.l	#2,d0

	divu	#10,d0
	add.b	#"0",d0
	move.b	d0,(a1)+

	swap	d0
	add.b	#`0`,d0	
	move.b	d0,(a1)+

	clr.b	(a1)
SetGAD
	lea	mytags(a5),a3
	move.l	winhandle(A5),a1
	suba.l	a2,a2
	move.l	gadsave(A5),a6
	jsr	_LVOGT_Setgadgetattrsa(A6)
	rts
MakeLV:
	Push	All
	move.l	4.w,a6				;nur irgendeine Librarybase ben?tigt

	tst.l	winhandle(A5)
	beq.w	.lv_rts

	Push	d5
	bsr	remlist		;Liste abh?ngen

	lea	list(a5),a0
	NEWLIST	a0
	lea	Nodes(a5),a4		;1. Node

	move	Currentnumvoc(a5),d5
	moveq	#0,d4
	
	moveq	#MaxVoices-2,d7
	moveq	#2,d6
	lea	Prefslist(A5),a3
.nodeloop
	lea	list(A5),a0		;gleich nach Alloc anh?ngen
	move.l	a4,a1			;(erspart Arbeit)
	ADDTAIL
	lea	ML_Nodename(a4),a1
	move.l	a1,ln_name(A4)

	cmp	d4,d5
	bne.s	.nosele
	tst.b	onplay
	beq.s	.nosele

	cmp.w	#39,lib_version(a6)
	ifeq	V37Gadtools
	blo.s	.nosele
	else
	bra.s	.nosele
	endc

	move.b	#`|`,(a1)+		;indicator for selected row
.nosele
	addq	#1,d4

	move.w	P_Flags(a3),d0		;Flags
	and.b	#1,d0
	beq.s	.nodis
	move.b	#`(`,(a1)+
.nodis
	move.b	#` `,(a1)+

	move.l	d6,d0
	moveq	#2,d1			;Anzahl der Stellen
	bsr	todez
	ifeq	AHIOUT
		move.b	#`c`,(a1)+
		move.b	#`h`,(a1)+
		move.b	#`:`,(a1)+	
		move.b	#` `,(a1)+
	else
		; mix mode
		move.w	P_Flags(a3),d0		;Flags
		move.b	#`B`,d1
		btst	#PBF_SincMixB,d0
		bne.s	.mixmod_this
		move.b	#`P`,d1
		btst	#PBF_SincMix,d0
		bne.s	.mixmod_this
		move.b	#`A`,d1
.mixmod_this
		move.b	#`:`,(a1)+	
		move.b	d1,(a1)+		;store mixmode
	endc

	move.l	P_Mixrate(A3),d0	;Mixingrate
	moveq	#5,d1			;Anzahl der Stellen
	bsr	todez
	move.b	#` `,(a1)+

	push	d5
	moveq	#0,d5
	move.w	P_Vboost(A3),d5		;Volumeboost
	cmp.w	#99,d5
	ble.s	.no10
	move	#99,d5
.no10:
	move.l	d5,d0
	divu	#10,d0			;durch 100 (wg. Nachkomma)
	and.l	#$fff,d0
	moveq	#1,d1
	bsr	todez
	move.b	#`.`,(a1)+

	move.l	d5,d0
	divu	#10,d0	
	swap	d0			;Nachkommastelle
	add.b	#`0`,d0
	move.b	d0,(a1)+
	move.b	#` `,(a1)+
	pop	d5
					;A3 steht auf Effektwort
	moveq	#" ",d1
	move.b	P_Surround(a3),d0
	bpl.s	.hi_s
	neg.b	d0
	move.b	#`-`,d1
.hi_S	move.b	d1,(a1)+
	and.l	#$ff,d0
	moveq	#2,d1
	bsr	todez
	move.b	#` `,(a1)+

	ifeq	AHIOUT
		move.b	P_SurroundDelay(a3),d0
		and.l	#$ff,d0
		moveq	#2,d1
		bsr	todez
		move.b	#` `,(a1)+
	
		; mix mode
		move.w	P_Flags(a3),d0		;Flags

		move.b	#`B`,d1
		btst	#PBF_SincMixB,d0
		bne.s	.mixmod_this
		move.b	#`P`,d1
		btst	#PBF_SincMix,d0
		bne.s	.mixmod_this
		move.b	#`A`,d1
.mixmod_this
		move.b	d1,(a1)+		;store mixmode
		move.b	#` `,(a1)+
	endc

	;AHI: add name
	ifne	AHIOUT
	;D0 - Mode ID
	;A1 - Output array
	move.l	P_AHIMode(a3),d0
	moveq	#ML_NameLength-38,d1	;
	bsr	AHIAMP_Mode2Name
	endc
	
	move.w	P_Flags(a3),d0		;Flags
	and.b	#1,d0
	beq.s	.nodis2
	move.b	#`)`,(a1)+
.nodis2
	clr.b	(a1)

	addq.w	#1,d6
	lea	P_Sizeof(A3),a3
	lea	ML_Sizeof(a4),a4
	dbf	d7,.nodeloop

	Pop	d5
	bsr	readdlist
.lv_rts
	Pop	All
	rts
***************** Liste vom Listview-Gadget entfernen ****************
Remlist:
	movem.l	d0-a6,-(sp)
	tst.l	winhandle(a5)
	beq.s	.err

	moveq	#GD_LISTVIEW,d0
	bsr	_GADR			;Gadgetadresse in A0
	lea	mytags(a5),a3
	move.l	#GTLV_Labels,(a3)+
	move.l	#-1,(a3)+
	clr.l	(A3)
	lea	mytags(a5),a3

	move.l	winhandle(A5),a1
	suba.l	a2,a2
	move.l	gadsave(A5),a6
	jsr	_LVOGT_Setgadgetattrsa(A6)
.err
	movem.l	(sp)+,d0-a6
	rts
**************** Liste wieder an Gadget anh?ngen *********************
readdlist:
	movem.l	d0-a6,-(sp)
	tst.l	winhandle(A5)
	beq.s	.err

	moveq	#GD_LISTVIEW,d0
	bsr	_GADR			;Gadgetadresse in A0
	lea	mytags(a5),a3
	move.l	#GTLV_Labels,(a3)+
	lea	list(a5),a2
	move.l	a2,(a3)+
	move.l	#GTLV_Selected,(a3)+
	move.l	selectedL(a5),(a3)+
	clr.l	(A3)
	lea	mytags(a5),a3

	move.l	winhandle(A5),a1
	suba.l	a2,a2
	move.l	gadsave(A5),a6
	jsr	_LVOGT_Setgadgetattrsa(A6)

;	move.l	gadsave(A5),a6
;	move.l	winhandle(a5),a0
;	suba.l	a1,a1
;	jsr	_LVOGT_refreshwindow(A6)
.err
	movem.l	(sp)+,d0-a6
	rts
**************** Hex to decimal ASCII conversion ************************
ToDEZ
	movem.l	d0-a0/a2-a6,-(Sp)
	lea	dectable(pc),a0
	move.w	d1,d2
	lsl.w	#2,d2
	add.w	d2,a0
	subq.w	#1,d1
.loopl
	moveq	#-1,d3
	move.l	-(a0),d2
.loop2l
	addq.b	#1,d3
	sub.l	d2,d0
	bcc.s	.loop2l
	add.l	d2,d0
	add.b	#`0`,d3
	move.b	d3,(a1)+
	dbf	d1,.loopl
	movem.l	(sp)+,d0-a0/a2-a6
	rts

;*****************************************************************************
;*                        Amplifier-Init/Mix Routinen                        *
;*****************************************************************************
Checkfeatures:
;          - Initplayer -> Übergabe A0: Tagliste
;                                   A5: EP Globals
;                       -> zurück   D0: 0=OK, <>0=Error
;
	move.l	#EPAMT_Flags,d0
	jsr	ENPP_FindTag(A5)
	beq	.err

	moveq	#1,d2
	move.l	d0,d1
	move.l	d0,d6
	and.l	#EPAMB_WaitForStruct,d6
	and.l	#EPAMB_8Bit,d0
	bne	.ok
	moveq	#2,d2
	btst	#EPAMF_8BitUnsigned,d1
	bne	.ok

	moveq	#3,d2
	btst	#EPAMF_16Bit,d1			;erstmal nur 16 Bit signed
	beq	.err
.ok
	move.l	#EPAMT_Audiostructs,d0
	jsr	ENPP_FindTag(A5)
	beq	.err
	move.l	d0,d7

	move.l	#EPAMT_NumStructs,d0
	jsr	ENPP_FindTag(A5)
	beq	.err
	move.l	d7,a0

	subq.w	#1,d0			;Anzahl der Audio-Strukturen
	blt	.err

	moveq	#0,d1
	moveq	#0,d3
	lea	(a0),a2
.countvoc
	tst.w	AS_LeftRight(a2)
	bpl.s	.addL
	bmi.s	.addR
	bra.s	.noADD
.addL
	cmp.w	#2,AS_Leftright(a2)
	beq	.err
	addq.w	#1,d1
	bra.s	.noADD
.addR
	addq.w	#1,d3
.noadd
	lea	as_sizeof(a2),a2
	dbf	d0,.countvoc

	move.w	d3,d0
	add.w	d1,d0
	beq	.err			;wie, was, gar nichts ?

	cmp	#3,d2
	bne.s	.normal8bit

	cmp	#1,d3
	bne	.err
	cmp	#1,d1
	bne	.err

.normal8bit

	subq.w	#2,d0
	bgt.s	.hi1
	moveq	#0,d0
.hi1
	mulu	#P_Sizeof,d0
	
	lea	Prefslist+datas,a1
	lea	(a1,d0.w),a1
	
	move.w	P_Flags(A1),d0		;Flags.w (Bit0=disable)
	and.b	#PBB_Disable,d0
	bne	.err			;Modus disabled !!

	moveq	#0,d0
	rts
.err
	moveq	#-1,d0
	rts

***************************************************************************
*  set mixing routine based on preferences                                *
***************************************************************************
setMixRoutine:				;requires "usedmode" to be active
	push	all
	tst.b	onplay
	beq.w	.err

	move.l	usedmode,d0
	beq.w	.err
	move.l	d0,A1

	;------------------------------
        lea.l   mix16(pc),a2		; 16 Bit Mixing ?
        lea.l   mix16(pc),a3
        cmp.b   #FMT_S16,fformat
        bne.s   .no16

	lea	HighBoostFilter28(pc),a5
	move.l	P_Mixrate(a1),d0
	cmp.l	#HighBoost28_Maxrate,d0
	blt.s	.filter28
	lea	HighBoostFilter(pc),a5
.filter28:

	move.w	P_Flags(A1),d0		;Flags.w
	btst	#PBF_SincMixB,d0	;mix with high boost ? 
	bne.s	.use16			;
	lea	0.w,a5			;no high boost
	btst	#PBF_SincMix,d0
	beq.s	.found			;no sinc -> stay with nearest neighbor
.use16:
	lea	fm16_mix(pc),a2		;polyphase 4 tap sinc mixer, 16 Bit
	lea	fm16_mix(pc),a3		;
	bra.s	.found			; yep, don't check other variant
.no16:
	;-----------------------------

	lea	fm_mix(pc),a2		;polyphase 4 tap sinc mixer
	lea	fm_mix2(pc),a3		;

	lea	HighBoostFilter28(pc),a5
	move.l	P_Mixrate(a1),d0
	cmp.l	#HighBoost28_Maxrate,d0
	blt.s	.filter28_2
	lea	HighBoostFilter(pc),a5
.filter28_2:
	move.w	P_Flags(A1),d0		;Flags.w
	btst	#PBF_SincMixB,d0
	bne.s	.found

	lea	0.w,a5

	btst	#PBF_SincMix,d0
	bne.s	.found

	lea	mix(pc),a2		; nearest neighbor mixer
	lea	mix2(pc),a3		;
.found:
	move.l	a2,mixad1
	move.l	a3,mixad2
	move.l	a5,boostad		;high boost address
.err:
	pop	all
	rts
***************************************************************************
*              kopiert neue Audio-Struct auf die eigene AS Struktur r?ber *
***************************************************************************
Copy_AS:			;mit Umschaltung von dff09a
		Push	All
		lea	$dff09a,a6

		tst.b	onplay
		beq.w	.err

		move.l	MYEpBase(pc),a0
		move.l	epg_audiostruct(A0),d0
		beq.w	.err
		move.l	d0,a0			;Original

;		move.l	myAudioStruct(pc),a1	;meine Extra-Struktur
		lea	Extra_AS,a1

		move	numstructs(pc),d0
		subq	#1,d0
		bmi.w	.err

;		move	#1<<7,(a6)		;Audio Int verbieten
.all
		move.l	AS_ChangeFlags(a0),d1
		btst	#ASChF_Adr,d1
		beq.s	.noad
		move.l	AS_CurrentAdr(a0),AS_CurrentAdr(a1)
		clr.l	AS_CurrentPos(a1)
		clr.l	AS_CurrentFPos(a1)
		clr.b	AS_NoLoop(a1)
.noad
		btst	#ASChF_Len,d1
		beq.s	.nolen
		move.l	AS_SampleSize(a0),AS_SampleSize(a1)
.nolen
		btst	#ASChF_RepAdr,d1
		beq.s	.norepadr
		move.l	AS_RepeatAdr(a0),AS_Repeatadr(a1)
.norepadr
		btst	#ASChF_RepLen,d1
		beq.s	.noreplen
		move.l	AS_RepeatSize(a0),AS_RepeatSize(a1)
.norepLen
		move.w	AS_Period(a0),AS_Period(a1)
		move.w	AS_LeftVolume(a0),AS_LeftVolume(a1)
		move.w	AS_RightVolume(a0),AS_RightVolume(a1)
		move.w	AS_DmaBit(a0),AS_DmaBit(a1)
		move.w	AS_LeftRight(a0),AS_LeftRight(a1)
		move.b	AS_LoopFlag(a0),AS_LoopFlag(a1)
		clr.l	AS_ChangeFlags(a0)

		lea	AS_Sizeof(a0),a0
		lea	AS_Sizeof(a1),a1
		dbf	d0,.all

;		move	#1<<7!$8000,(a6)	;Kram kopiert, also Audio Int wieder an
.err
		Pop	All
		rts


Copy_AS2:
		Push	All

		tst.b	onplay
		beq.s	.err

		move.l	myAudioStruct(pc),a1	;meine Extra-Struktur

		move.l	MYEpBase(pc),a0
		move.l	epg_audiostruct(A0),d0
		beq.s	.err
		move.l	d0,a0			;Original

		move	numstructs(pc),d0
		subq	#1,d0
		bmi.s	.err

.all
		moveq	#AS_Sizeof-1,d1
.cp		move.b	(a0)+,(a1)+
		dbf	d1,.cp

		dbf	d0,.all
.err
		Pop	All
		rts
*************************************************************************
* clear precalculated periods in audio structure                        *
* - inputs:  -                                                          *
* - outputs: - (audio struct changed)                                   *
*************************************************************************
ResetASPeriods:
	Push	d0/a4

	move.l	myaudiostruct(pc),d0
	beq.s	.RASP_end
	move.l	d0,a4

	move	numstructs(pc),d0
	subq	#1,d0
	bmi.w	.RASP_end

.RASP_loop:
	clr.w	ASP_LastPeriod(a4)
	clr.w	ASP_Anticlickpos(a4)

	lea	AS_Sizeof(a4),a4
	dbf	d0,.RASP_loop

.RASP_end:
	Pop	d0/a4
	rts


*************************************************************************
*              Mem fuer Tabellen besorgen und Tabellen init              *
*************************************************************************
InitTables:
	movem.l	d1-a6,-(Sp)
	lea	datas,a5

	ifeq	nosurroundtab
		;------- 128k Surround/Mono Table ---------
		clr.l	SurroundTabAddr
		clr.l	SurroundTabAddrMid

		tst.w	MC68040on(A5)		;no surround table on 68040+ -> requires rewrite of copybufmono
;		bne.s	.nosurtab


		moveq	#2,d0
		swap	d0			;move.l	#64*1024*2,d0
		move.l	d0,SurroundTabSize
		moveq	#MEMF_PUBLIC,d1
		move.l	4.w,a6
		jsr	_LVOallocmem(A6)
		move.l	d0,SurroundTabAddr
		beq.w	myFreemem2
		moveq	#1,d1	;1
		swap	d1	;$10000
		add.l	d1,d0	;middle of table
		move.l	d0,SurroundTabAddrMid
.nosurtab
	endc
	

	;------ 16 to 14 bit conversion table -----
	moveq	#2,d0
	swap	d0			;move.l	#64*1024*2,d0
        moveq   #MEMF_PUBLIC!MEMF_FAST,d1
        move.l  4.w,a6			;redundant
        jsr     _LVOallocmem(A6)
        move.l  d0,buff14
        beq.w   myFreemem2
        add.l   #$10000,d0
        move.l  d0,buff14_2

	;------------ mixing table ----------------
	move.l	#66*256*2*2,d0		;65 volume steps, 256 values, 2 channels (left/right), 2 bytes per entry

	cmp.w   #FMT_S16,fformat
        blo     .no16it
	move.l  #$10000*2*2+256,d0        ;65000 words * 2 channels
.no16it
	move.l  d0,Vtabsize

	moveq	#MEMF_PUBLIC,d1
	move.l	4.w,a6
	jsr	_LVOallocmem(A6)
	move.l	d0,Vtab
	beq.b	myFreemem2

	add.l	#255,d0
	and.l	#~$ff,d0
	move.l	d0,Vtabaddr

	move.l	#66*256*2,d1
	cmp.w   #FMT_S16,fformat
        blo     .no16it_2
	move.l	#$10000*2,d1
.no16it_2
	add.l	d1,d0
	move.l	d0,VtabaddrR		; Vtab for right channel

	;bsr	do14tab			;bsr	makedivtabs -> relegated to HW output
	bsr	Makevoltable

	bsr	MakeSilenceBuffer	;Silence buffer for FM Fadeout (depends on signed/unsigned formats)

	movem.l	(sp)+,d1-a6
	moveq	#0,d0
	rts
***************************************************************************
*                 Speicher von Divtab / VolTab freigeben                  *
* no reference to a5 in here                                              *
***************************************************************************
myFreemem:
	movem.l	d1-a6,-(sp)
myFreemem2:
	move.l	4.w,a6

	move.l	Vtab(pc),d0
	beq.s	.empty1
	move.l	d0,a1
	move.l	Vtabsize,d0
	jsr	_LVOfreemem(A6)
	clr.l	Vtab
.empty1
        move.l  buff14(pc),d0
        beq.s   .empty3
        move.l  d0,a1
        move.l  #64*1024*2,d0
        jsr     _LVOfreemem(A6)
        clr.l   buff14
        clr.l   buff14_2
.empty3
	ifeq	nosurroundtab

		move.l	SurroundTabAddr(pc),d0
		beq.s	.empty2
		move.l	d0,a1
		move.l	SurroundTabSize(pc),d0
		jsr	_LVOfreemem(A6)
		clr.l	SurroundTabAddr
		clr.l	SurroundTabAddrMid
.empty2
	endc
	
	movem.l	(sp)+,d1-a6
	moveq	#-1,d0
	rts

*****************************************************************************
*              Mono/Surround table                                          *
* - Input:  D0 - mode: -64...64 (=Mono....Stereo....Surround                *
*              - SurroundTabAddr   must be allocated with 128k FastRAM      *
* - Output:                                                                 *
*****************************************************************************
MakeSurroundTab:
	ifeq	nosurroundtab
	
		push	all
		move.l	SurroundTabAddr(pc),d1	;will happen on 68040+
		beq	.fail_no_tab
		move.l	d1,A0		;beginning of negative range

		moveq	#1,d1		;1
		swap	d1		;$10000
		lea	(A0,D1.l),A0	; middle of table (64k DWORDS), beginning of positive range
		movea.l	A0,A1

		ext.w	d0		; effect volume (-64...0...64)
		neg.w	d0		; reverse sign (negative = surround, positive = mono)

		move	#32768-1,d7	; 
		moveq	#0,d1		;start at 0
		moveq	#-1,d3
.surtab
		move.l	d1,d2
		move.l	d3,d4
		muls	d0,d2		;*effect volume
		muls	d0,d4		;*effect volume
		lsr.l	#6,d2		;/(max. volume)
		lsr.l	#6,d4		;/(max. volume)
		move	d2,(a0)+	;positive range
		move	d4,-(a1)	;negative range
		addq	#1,d1
		subq	#1,d3
		dbf	d7,.surtab

.fail_no_tab:
		pop	all
	endc
	rts



***********************************************************************************
* create volume table(s)                                                          *
* - one for left, one for right channel                                           *
*    - incorporates input volumes 0..64                                           *
*    - incorporates global volume and balance (hence one for left, one for right) *
***********************************************************************************
Makevoltable:
	push	all
	moveq	#0,d0 ; left channel
	bsr	mvt_getvolume	;d0: max. amplitude as 15 bit (0..32767)

	move.l	Vtabaddr,d1
	beq.s	.rts
	move.l	d1,a0
	bsr	mvt_onetab

	moveq	#1,d0 ; right channel
	bsr	mvt_getvolume	;d0: max. amplitude as 15 Bit (0.32767)

	move.l	VtabaddrR,a0
	bsr	mvt_onetab
.rts
	pop	all
	rts
	; calculate effective volume for left/right (0..255)
	; - number of active channels
	; - global volume
	; - effect volume
	;
mvt_getvolume:
		movem.l	d1-d2/a5,-(sp)
	ifeq	mixdebug

	; global volume / balance
		move.l	MyEPBase(pc),d1
		beq.w	.quit
		move.l	d1,a5
		move.w	DTG_SndRBal(a5),d1
		tst.w	d0
		bne.s	.right
                move.w  DTG_SndLBal(a5),d1
.right:
                mulu    DTG_SndVol(a5),d1
                lsr.w   #6,d1                           ; global volume 0..64 for left or right channel

		move.w	#32767,d0			; default max. amplitude
		mulu.w	d1,d0				; max. amplitude * volume
		lsr.l	#6,d0				; max. amplitude after volume/balance

	; number of channels
		move.w	numleft(pc),d1
		cmp.w	numright(pc),d1
		bgt.s	.hi
		move.w	numright(pc),d1
.hi
		tst.w	d1
		beq.s	.skipdiv
		divu	d1,d0				; max. amplitude / max. number of channels
.skipdiv
		ext.l	d0


;		FM_sinc_norm 
		move.l	usedmode,d1
		beq.w	.fmix_err
		move.l	d1,A1
		move.w	P_Flags(A1),d1		;Flags.w
		btst	#PBF_SincMix,d1
		beq.s	.fmix_err
		mulu	#FM_sinc_norm,d0
		add.l	d0,d0
		swap	d0
		ext.l	d0
.fmix_err

		;highboost filter
		move.l	usedmode,d1
		beq.w	.nohighboost
		move.l	d1,A1
		move.w	P_Flags(A1),d1		;Flags.w
		btst	#PBF_SincMixB,d1
		beq.s	.nohighboost

		move.l	P_Mixrate(A1),d1	;Flags.w
		cmp.l	#HighBoost28_Maxrate,d1
		bge.s	.high44
		mulu.w	#50000,d0
		bra.s	.highboost_apply
.high44:
		mulu.w	#42000,d0		;TODO: verify that this constant is sufficiently low
.highboost_apply:
		swap	d0
		ext.l	d0
		;lsr.l	#3,d0			;divide by 8 (4 would probably be sufficient)
.nohighboost

	; effect volume (either surround or mono, 0..64)
		move	effectvol(pc),d2		; effect volume
		add.w	#64,d2				; sum of "natural" volume and effect volume
		lsl.l	#6,d0				; amplitude * 64
		divu	d2,d0				; amplitude * 64 / (natural volume + effect volume)
		ext.l	d0				; 
.quit
	
		; sanity check
		else
			move.l	#$7fff,d0
		endc
			and.l	#$7fff,d0			; make sure, the value is never >15 bit

		;move.l	d0,0		; !!DEBUG

		movem.l	(sp)+,d1-d2/a5
		rts

	; one table, either for left or right channel
	; a0 - table (space for 65*256*2 bytes)
	; d0 - global amplitude for left or right channel (0...32767)
mvt_onetab:
	cmp.b	#FMT_S8,fformat	 ; signed
	beq.b	mvt_signed
;	bra	signed

	cmp.b	#FMT_S16,fformat ; 16 bit signed
	beq.b	mvt_signed16

	; TODO: 16 bit signed
	
mvt_unsigned:
	moveq	#0,d3		;volume
.loopvol:
	move	d0,d6		;global max. amplitude
	mulu	d3,d6		;*(current volume in table)
	lsr.l	#6,d6		;/(max volume)

	moveq	#0,d4		;data (0...255)
.loopamplitude:
	move.w	d4,d5		;0...255
	sub.w	#128,d5		;0...255 -> -128...127
	muls	d6,d5		;*current max. amplitude
	asr.l	#7,d5		;/max. amplitude of input samples (-128...127)
	move.w	d5,(a0)+	;signed 16 bit for mixing

	addq	#1,d4
	cmp	#256,d4
	bne.b	.loopamplitude

	addq	#1,d3
	cmp	#65,d3
	bne.b	.loopvol
	rts

	; 8 Bit to 16 bit signed volume table
	; addressing using unsigned extension of the sample data
	; e.g.
	;  moveq #0,d0           ; zero upper bits
	;  move.b (a0),d0        ; get 8 Bit data 
	;  move.w (a1,d0.w*2),d0 ; treat 8 bit data as unsigned, i.e. positive range (0..127), followed by negative range (128..255)
mvt_signed:
	moveq	#0,d3		;volume
.loopvol:
	move	d0,d6		;global max. amplitude
	mulu	d3,d6		;*(current volume in table)
	lsr.l	#6,d6		;/(max volume)

	moveq	#0,d4		;data (0...255)
.loopamplitude:
	move.b	d4,d5
	ext.w	d5		;signed representation of current amplitude (-128...127)
	muls	d6,d5		;*current max. amplitude
	asr.l	#7,d5		;/max. amplitude of input samples (-128...127)
	move.w	d5,(a0)+	;signed 16 bit for mixing

	addq	#1,d4
	cmp	#256,d4
	bne.b	.loopamplitude

	addq	#1,d3
	cmp	#65,d3
	bne.b	.loopvol
	rts

	; the table is addressed with inputs treated as 32 bit unsigned, hence
	; 0...32767,-32768...-1
mvt_signed16:
;	move	d0,0		; !!DEBUG

	move.l	#65536-1,d7
	moveq	#0,d1
.loop_pos
	move	d1,d2
	muls	d0,d2		;32767 * level
	swap	d2		;get upper bits
	move	d2,(a0)+	;store

	addq.l	#1,d1
	subq.l	#1,d7
	bge.s	.loop_pos

	rts

****************************************************************************
* Initialize Filtered Mixing Silence Buffer,                               *
* depending on signed/unsigned samples                                     *
* Input: -                                                                 *
****************************************************************************
MakeSilenceBuffer:	;
	push	d0-d2/a0
	lea	SilenceBuffer,a0
	move.b	fformat(pc),d0

	move.b	#128,d1		; zero for unsigned signed 8 Bit
	cmp.b	#FMT_U8,d0
	beq.s	set8$

	moveq	#0,d1		; zero for signed 8/16 Bit

	cmp.b	#FMT_S8,d0
	beq.s	set8$

	;FMT_S16, d1 still zero
	moveq	#MIX_FM_FADEOUT_TIME-1,d0	;if zero, skip the copy
	blt.s	done$
loop16$
	move.w	d1,(a0)+
	dbf	d0,loop16$
	bra.s	done$
set8$
	moveq	#MIX_FM_FADEOUT_TIME-1,d0	;if zero, skip the copy
	blt.s	done$
loop8$
	move.b	d1,(a0)+
	dbf	d0,loop8$
done$

	pop	d0-d2/a0
	rts
****************************************************************************
*                  Delay setzen fuer Surround Modus                         *
* Input: d0 = Wert/10                                                      *
****************************************************************************
;
setPrefetch:
	push	d0/d1

	mulu	#10,d0		;

	and.w	#$fffc,d0	; make sure, this is dividable by 4
	bne.s	.hi1
	moveq	#4,d0
.hi1
	cmp.w	#SamLEN-8,d0
	ble.s	.lo
	move.w	#SamLEN-8,d0
.lo
	move.w	d0,prefetchR

	move.b	Begin+OnMonoMode(pc),d1
	beq.s	.works
	sub	#64,d0
	bgt.s	.works
	moveq	#4,d0
.works
	move.w	d0,prefetchL

	pop	d0/d1
	rts

***************************************************************************
* Update mixing rate in active output hw (if currently playing)           *
***************************************************************************
CalcValues:
	move.b	onplay(pc),d0
	beq.w	.skipcalc

	move.l	usedmode,a0
	move.l	(a0),d0
	move.l	d0,playmixrate
	bsr	Amp_SetMixRate
.skipcalc:
	rts


;-----------------------------------------------------------------------------------------------------
; copy audio struct back if extra audio struct is in use
;
CopyASBack:
		move.l	myAudioStruct(pc),d0
		cmp.l	#Extra_AS,d0
		bne.s	.out
		move.l	d0,a0

		move.l	MYEpBase(pc),a1
		move.l	epg_audiostruct(A1),d0
		beq.w	.out
		move.l	d0,a1

		move	numStructs(pc),d0
		subq	#1,d0
.cp
		movem.l	(a0)+,d1-d7/a2-a6	;7+5 Register = 12*4=48 Bytes
		movem.l	d1-d7/a2-a6,(a1)
		lea	12*4(a1),a1
		movem.l	(a0)+,d1-d7/a2-a6	;7+5 Register = 12*4=48 Bytes
		movem.l	d1-d7/a2-a6,(a1)
		lea	12*4(a1),a1
		move.l	(a0)+,(a1)+
		dbf	d0,.cp
.out
		rts

;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
;--------------------------- 8 Bit Filtered Mixing Routine (Part1 - Move Data) -----------------------
;-----------------------------------------------------------------------------------------------------
; Input:
;        A4   - AudioStruct for current channel and panning
;        A2   - Output buffer for "SamLEN*2" shorts
;        D0   - number of samples (SamLEN*2)
;        A6   - Volume table
; Output:
;        (A2) - Output samples
;        D0   - 1=mixing done, 0=mixing skipped (no sample played)
;
; Used Registers: all, none saved
;        D0   - current offset towards A0 (upper word is current fractional position)
;        D4   - interpolation step size, lower word integer part, upper word fractional part
;        A0   - current source sample, offset by (already) played bytes
;        A5   - remaining samples (except in fast loop, where it is -16)
;        A6   - Volume Table
; Notes:
;  - 4 Tap FIR interpolation filter (Windowed Sinc), tradeoff between multiplies per output sample 
;    and sound quality
;  - TODO: ping-pong backwards preload and FM mix, currently uses nearest neighbor
;-----------------------------------------------------------------------------------------------------

; check whether instrument is looping
;  - branch to bailout if end of one-shot instrument is reached
;  - branch to PingPong loop code if ping-pong instruments are in use
;  - issue "interrupt" if necessary
;  - output registers (if not branched away):
;     D0 - current integer position (upper half of D0 untouched)
;     A0 - current address (at this point, exactly the repeat address)
;     D6 - new length (range check necessary, if >$7fff)
fm_mix:
	ifne	benchmark
	move	#$f00,$dff180
	endc

	; filtered mixing routine can`t handle downsampling, refer to regular "mix" in those cases
	move.w	AS_Period(a4),d6	;current period
	beq.w	.fm_mix_ret_nop
	cmp.w	mixingperiod4,d6	;must be larger than mixing period, else use regular routine
	ble.w	mix			;ok, period is wrong, use regular mixer	

	; init macro
	MIX_INIT	.fm_mix_ret_nop,1,0,0 ; set up registers (common routine to all mixers)

	btst	#ASLoopF_FMInit,AS_CurrentFPOS+2(a4) 
	bne.s	.fm_mix_fminitialized	; Filtered Mixing initialized ?
	bsr.w	fm_mix_preload		; preload samples into filter buffer
	move.w	AS_LeftVolume(A4),AS_FMVol(A4)
.fm_mix_fminitialized:

	btst	#ASLoopF_BackLooping,AS_CurrentFPOS+2(a4) ; D6 -> stop position for backward loop
	bne.w	.fm_mix_Ping_backwards

	MIX_REMBYTES	d6,d1,0		;total length - played length = remaining bytes, 2nd arg: scratch register, 3rd arg: 1=16 Bit mode, 0=8 Bit mode

;-----------------------------------------------------------------------------------------------
; main loop: check vital parameters and time-scale the inputs
;-----------------------------------------------------------------------------------------------
.fm_mix_mainloop:			;start point for "fast" loop and "repeated" samples
	MIX_LIMITER	d6,norm		;limit remaining samples in D6 to $7fff (16 bit signed max.)

	ifne	benchmark
	move	#$00f,$dff180
	endc

	moveq	#0,d3
	move.b	AS_FMPos(a4),d3			;ringbuffer index
	lsl.w	#FMB_Accuracy+FMB_Taps+1,d3	;accuracy bits + taps bits + word size
	lea	FM_tab_100,a3
	add.l	d3,a3				;table for current ringbuffer position

	lea	-16(a5),a5 ; simplifies check for "less than 16 samples" to generate

	move.l	d0,d7			; current position in instrument
	add.l	a1,d7			; + 16 output samples wrt. sampling rate
	cmp.w	d7,d6			; >= end of instrument
	blo.w	.fm_mix_underflow16	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)
	move.w	a5,d1			; 16 samples left to write in output buffer ?
	blt.w	.fm_mix_underflow16 	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

;	tst.w	ASP_Anticlickpos(a4)
;	beq.w	.fm_mix_fastloop
;-----------------------------------------------------------------------------------------------
; main loop - anticlick
; d1,d2,d3,d5,d7 are free
;-----------------------------------------------------------------------------------------------
.fm_mix_fastloop_ac:
		fm_mix_one_ac	1
		fm_mix_one_ac	2
		fm_mix_one_ac	3
		fm_mix_one_ac	4
		fm_mix_one_ac	5
		fm_mix_one_ac	6
		fm_mix_one_ac	7
		fm_mix_one_ac	8
		fm_mix_one_ac	9
		fm_mix_one_ac	10
		fm_mix_one_ac	11
		fm_mix_one_ac	12
		fm_mix_one_ac	13
		fm_mix_one_ac	14
		fm_mix_one_ac	15
		fm_mix_one_ac	16

		lea	-16(a5),a5	; no condition codes changed (=16 samples processed)
		move.l	a5,d1							          
		swap	d1		; if( remaining_output_bytes < 16 ) 0xffff else 0x0000
		or.w	d6,d1		; if( remaining_output_bytes < 16 ) -1 else remaining_input_bytes
		move.l	d0,d7		;processed bytes
		add.l	a1,d7		;last byte to be processed
		cmp.w	d7,d1		;if( D1 < D7 ) -> stop
		bgt.w	.fm_mix_fastloop_ac
	
		bra	.fm_mix_underflow16
;-----------------------------------------------------------------------------------------------
; main loop
; d1,d2,d3,d5,d7 are free
;-----------------------------------------------------------------------------------------------
.fm_mix_fastloop:
		fm_mix_one	1
		fm_mix_one	2
		fm_mix_one	3
		fm_mix_one	4
		fm_mix_one	5
		fm_mix_one	6
		fm_mix_one	7
		fm_mix_one	8
		fm_mix_one	9
		fm_mix_one	10
		fm_mix_one	11
		fm_mix_one	12
		fm_mix_one	13
		fm_mix_one	14
		fm_mix_one	15
		fm_mix_one	16

		lea	-16(a5),a5	; no condition codes changed (=16 samples processed)
		move.l	a5,d1							          
		swap	d1		; if( remaining_output_bytes < 16 ) 0xffff else 0x0000
		or.w	d6,d1		; if( remaining_output_bytes < 16 ) -1 else remaining_input_bytes
		move.l	d0,d7		;processed bytes
		add.l	a1,d7		;last byte to be processed
		cmp.w	d7,d1		;if( D1 < D7 ) -> stop
		bgt.w	.fm_mix_fastloop
;-----------------------------------------------------------------------------------------------

;-----------------------------------------------------------------------------------------------
; less than 16 bytes to go or sample ending
;-----------------------------------------------------------------------------------------------
.fm_mix_underflow16:
		lea	16(a5),a5 ; restore A5 with remaining bytes to be filled (we started with -16 in fast loop)

		ifne	benchmark
		move	#$444,$dff180
		endc

.fm_mix_underflow_loop:
		move.w	a5,d1
		ble.w	.fm_mix_buffer_full ; done, buffer filled

		fm_mix_one_ac	17	; includes anticlick
		
		cmp	d0,d6		;
		lea	-1(a5),a5	; samples -1 (before jumping out)
		bhi.w	.fm_mix_underflow_loop ; 
		; loop until end of instrument
;-----------------------------------------------------------------------------------------------

;-----------------------------------------------------------------------------------------------
; we land here with A5 having the number of remaining bytes
; decide whether to loop the current sample or to clear the rest of the buffer
;-----------------------------------------------------------------------------------------------

	; non-looping instruments require some extra treatment:
	;  - we have a look-ahead in here, so that 2 "future" samples are already stored
	;  - provide a "dummy" extension with "0"-samples
	;  - no problem with "repeating" instruments since the integer position is reset and
	;    new samples are loaded as needed
 	MIX_FM_FADEOUT .fm_mix_clear_rest,.fm_mix_mainloop ; 

	;-------------------------- "usual" loop code ------------------------------------------

	MIX_LOOPCHECK	.fm_mix_clear_rest,.fm_mix_PingPongLoop,0 ;1st arg: bra target for one-shot samples, 2nd arg: bra target for ping-pong loop code;3rd arg: 16 bit mode

	bra.w	.fm_mix_mainloop

;----------------- ping-pong loop first entry --------------------------
.fm_mix_PingPongLoop
	MIX_PINGPONG_INIT	.fm_mix_Ping_backwards	;arg1: where to continue when samples are still left in output
	; caution: continues directly below at "buffer_full"
	; when we land here, pingpong is needed but the buffer is full

;-------- DONE: one buffer is filled, now just store parameters and leave
.fm_mix_buffer_full:

	ifne	benchmark
	move	#$000,$dff180
	endc

	; check whether we reached the end of the instrument before leaving
	MIX_POSTCHECK	.fm_mix_bfull_ret,0

.fm_mix_bfull_ret:
	moveq	#1,d0
	rts

;--------------------- clear remaining bytes - TODO: filter here also ----------------
.fm_mix_clear_rest:
		move.l	A5,d1		;remaining bytes
		moveq	#$0,d0
		subq	#1,d1		;for dbf loop
		bmi.s	.fm_mix_ret_clear
.fm_mix_clear_loop:
		move	d0,(a2)+
		dbf	d1,.fm_mix_clear_loop
.fm_mix_ret_clear
		moveq	#1,d0
		rts

; no explicit clear when empty-buffer aware routines are in place
.fm_mix_ret_nop:
		moveq	#0,d0
		rts

;-------------- Ping-Pong Loop entry point for backwards loop -------------------
;- custom routine, repeating the normal interpolation but checking against negative offsets
.fm_mix_Ping_backwards	;yep, do loop backwards
					;in D6, the remaining length is kept (current position = rest until "0" is reached)
		move.l	a1,d1		;invert end position check
		neg.w	d1
		move.l	d1,a1

		MIX_LIMITER	d6,ping	;limit D6 to 0x7fff

		lea	-16(a5),a5 ; simplifies check for "less than 16 samples"
		moveq	#0,d5
		moveq	#0,d2
		moveq	#0,d3
.fm_mix_PingB_fastloop:
		move.w	a5,d1		 ; sets condition codes but not "X"
		lea	(A1,D0.l),A3	 ; see above, (needs addq #1,d1)
		blt.w	.fm_mix_PingB_underflow16 ; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

		;if position after 16 bytes (d0+next samples) <= start of loop, then leave here
		lea	(a3,d6.w),A3
		move.w	a3,d3
		blt.w	.fm_mix_PingB_underflow16	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

		rept	8
			move.b	(a0,d0.w),d2
			sub.l	d4,d0
			subx.w	d5,d0
			move.w	(A6,d2.w*2),d3
			swap	d3
			move.b	(a0,d0.w),d2
			sub.l	d4,d0
			subx.w	d5,d0
			move.w	(A6,d2.w*2),d3
			move.l	d3,(a2)+
		endr
		lea	-16(a5),a5	; no condition codes changed
		bra.w	.fm_mix_PingB_fastloop

		; slow loop, A5 almost empty and/or near end of loop area
.fm_mix_PingB_underflow16:
		lea	16(a5),a5 ; restore A5 with remaining bytes to be filled (we started with -16 in fast loop)

		ifne	benchmark
			move	#$444,$dff180
		endc

		; NOTE: X bit still set until this point
		; less than 16 bytes to be written (bad choice of SamLEN) 
		; or (more likely) near the end of the sample
.fm_mix_PingB_underflow_loop:
		move.w	a5,d1
		ble.w	.fm_mix_Ping_buffer_full

		move.b	(a0,d0.w),d2	; one byte at a time, happens rarely (!!)
		sub.l	d4,d0
		subx.l	d5,d0
		move.w	(A6,d2.w*2),(A2)+

		lea	-1(a5),a5	; samples -1 (before jumping out)

		move	d0,d3
		add	d6,d3
		blt.w	.fm_mix_Ping_directionChange ; if position is at beginning of loop area, we are done here
		
		bra.b	.fm_mix_PingB_underflow_loop
		; loop until end of sample

.fm_mix_Ping_directionChange:
		; we were looping backwards, now go forwards
		MIX_PING_DIRECTIONCHANGE        ;change mixing direction offsets, issue interrupt (if necessary)
		bsr	fm_mix_preload
		bra.w	.fm_mix_mainloop

.fm_mix_Ping_buffer_full
		MIX_PING_BUFFERFULL	.fm_mix_ping_return
.fm_mix_Ping_return
		moveq	#1,d0
		rts


;
; preload FM_Preload samples to AS_FMSamples buffer
; called at each start of an instrument (after a "silence" 
; or playing a new note with previously looping instruments)
; 
; -> not for repeating instruments, there the buffer is already available
;
; input:
;  - AS_Struct in A4
; output:
;  - D0 offset + 2
;
; notes:
;  - AS_FMPos may be at any state, sanity check is done here
fm_mix_preload:
	ifne	clickreduce
		move.w	#AC_INITIAL,ASP_Anticlickpos(a4)	;initialize Anticlick routine

	; relatively weak click reduce function:
	; - lowpass filter across buffered samples
	; -> extends just 3 buffered samples, hence very short period
		push	d1-d5

		move.b	AS_FMPos(A4),d3
		move.b	d3,d1
		subq.b	#3,d1					;get last "current" sample (two positions back in buffer)
		and.w	#FM_Taps-1,d1
		move.w	AS_FMSamples1(a4,d1.w*2),d4		;last "current" for IIR implementation = y[n-1]

		;IIR lowpass: y[n] = Anticlickpos * y[n-1] - (Anticlickpos-64) * x[n]
	ifne	anticlick
		move.w	ASP_Anticlickpos(a4),d5			;anticlick
	endc
		moveq	#FM_Preload-1,d1
.fm_mix_preload_loop:
	ifne	anticlick
		muls	d5,d4					;y[n-1] * Anticlickpos
		sub.w	#AC_REF,d5				;Anticlickpos-64
	else
		asr.w	#1,d4					;0.5 y[n-1]
	endc
		moveq	#0,d2					;todo: move out of loop
		move.b	(a0,d0.w),d2				;8 Bit input
		move.w	(A6,d2.w*2),d2				;16 Bit signed internal
	ifne	anticlick
		muls	d5,d2					;x[n]*(Anticlickpos-64)
		add.w	#AC_REF,d5				;Anticlickpos
		sub.l	d2,d4					;y[n]=acp*y[n-1] - (acp-64)*x[n]
		moveq	#0,d2
		asr.l	#AC_SHIFT,d4				;normalize
		subq.w	#1,d5					;d5: (0...AC_REF)-1 -> x bit set if underrun
		addx.w	d2,d5					;if( d5 == -1 ) d5 = 0
	else
		asr.w	#1,d2					;0.5 x[n]
		add.w	d2,d4					;y[n]=0.5( x[n]+y[n-1] )
	endc
		and.w	#FM_Taps-1,d3				;make sure, the position is sane (assumes 2^n taps)
		move	d4,AS_FMSamples1(a4,d3.w*2)		;Volume correction, store
		addq.w	#1,d0
		addq.w	#1,d3
		dbf	d1,.fm_mix_preload_loop

		;D0 points to the next sample to be loaded -> compatible with other mixing routines
		bset	#ASLoopF_FMInit,AS_CurrentFPOS+2(a4) ; Filtered Mixing initialized ?
	ifne	anticlick
		move.w	d5,ASP_Anticlickpos(a4)			;store anticlick status
	endc
		move.b	d3,AS_FMPos(A4)
		pop	d1-d5

	else ; clickreduce condition "0" = off follows below

		push	d1-d3

		move.b	AS_FMPos(A4),d3

		moveq	#FM_Preload-fadein-1,d1
		moveq	#0,d2
		move.l	d2,AS_FMSamples1(A4)			;clear sample buffer 
		move.l	d2,AS_FMSamples2(A4)			;clear sample buffer
.fm_mix_preload_loop:
		and.w	#FM_Taps-1,d3				;make sure, the position is sane (assumes 2^n taps)
		move.b	(a0,d0.w),d2				;8 Bit input
		move.w	(A6,d2.w*2),AS_FMSamples1(a4,d3.w*2)	;Volume correction, store
		addq.w	#1,d0
		addq.w	#1,d3
		dbf	d1,.fm_mix_preload_loop

		;D0 points to the next sample to be loaded -> compatible with other mixing routines
		;subq.w	#1,d0 ; point to last sample stored in buffer so +1 will load "next" in main loop (OBSOLETE)
		bset	#ASLoopF_FMInit,AS_CurrentFPOS+2(a4) ; Filtered Mixing initialized ?

		; fade instrument in by 1 sample (to reduce clicking)
		; longer fade-in times will require to re-calculate d0
		ifne	fadein
		sub.l	d4,d0		; go back 1 sample (so that the next sample will be loaded after the first byte)
		add.w	d4,d0		; just to make sure, in fact d4.w _should_ be zero as we check for downsampling cases
		endc

		move.b	d3,AS_FMPos(A4)
		pop	d1-d3
	endc
		rts

;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
;--------------------------- 8 Bit Filtered Mixing Routine (Part2 - Add Data) ------------------------
;-----------------------------------------------------------------------------------------------------
; Input:
;        A4   - AudioStruct for current channel and panning
;        A2   - Output buffer for "SamLEN*2" shorts
;        D0   - number of samples (SamLEN*2)
;        A6   - Volume table
; Output:
;        (A2) - Output samples
;        D0   - 1=mixing done, 0=mixing skipped (no sample played)
;
; Used Registers: all, none saved
;        D0   - current offset towards A0 (upper word is current fractional position)
;        D4   - interpolation step size, lower word integer part, upper word fractional part
;        A0   - current source sample, offset by (already) played bytes
;        A5   - remaining samples (except in fast loop, where it is -16)
;        A6   - Volume Table
; Notes:
;  - 4 Tap FIR interpolation filter (Windowed Sinc), tradeoff between multiplies per output sample 
;    and sound quality
;  - TODO: ping-pong backwards preload and FM mix, currently uses nearest neighbor
;-----------------------------------------------------------------------------------------------------

fm_mix2:
	ifne	benchmark
	move	#$f00,$dff180
	endc

	move.w	AS_Period(a4),d6	;current period
	beq.w	.fm_mix2_ret_nop
	cmp.w	mixingperiod4,d6	;must be larger than mixing period, else use regular routine
	ble.w	mix2			;ok, period is wrong, use regular mixer	

	MIX_INIT	.fm_mix2_ret_nop,1,0,0 ; set up registers (common routine to all mixers)


	btst	#ASLoopF_FMInit,AS_CurrentFPOS+2(a4) ; Filtered Mixing initialized ?
	bne.s	.fm_mix2_fminitialized
	bsr.w	fm_mix_preload		; preload samples into filter buffer

.fm_mix2_fminitialized:

	btst	#ASLoopF_BackLooping,AS_CurrentFPOS+2(a4) ; D6 -> stop position for backward loop
	bne.w	.fm_mix2_Ping_backwards

	MIX_REMBYTES	d6,d1,0		;total length - played length = remaining bytes, 2nd arg: scratch register, 3rd arg: 1=16 Bit mode, 0=8 Bit mode

	; main loop for 16 samples in one run 
.fm_mix2_mainloop:			;start point for "fast" loop and "repeated" samples
	MIX_LIMITER	d6,norm		;limit remaining samples in D6 to $7fff (16 bit signed max.)

	ifne	benchmark
		move	#$ff0,$dff180
	endc

	moveq	#0,d3
	move.b	AS_FMPos(a4),d3			;ringbuffer index
	lsl.w	#FMB_Accuracy+FMB_Taps+1,d3	;accuracy bits + taps bits + word size
	lea	FM_tab_100,a3
	add.l	d3,a3

	lea	-16(a5),a5 ; simplifies check for "less than 16 samples" to generate
	;max. sample length + current position in sample >= end of sample ?
	move.l	d0,d7
	add.l	a1,d7
	cmp.w	d7,d6			;
	blo.w	.fm_mix2_underflow16	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)
	move.w	a5,d1			;
	blt.w	.fm_mix2_underflow16 	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

;	tst.w	ASP_Anticlickpos(a4)
;	beq.w	.fm_mix2_fastloop
;-----------------------------------------------------------------------------------------------
; main loop - anticlick
; d1,d2,d3,d5,d7 are free
;-----------------------------------------------------------------------------------------------
.fm_mix2_fastloop_ac:
		fm_mix_one_ac_2	1
		fm_mix_one_ac_2	2
		fm_mix_one_ac_2	3
		fm_mix_one_ac_2	4
		fm_mix_one_ac_2	5
		fm_mix_one_ac_2	6
		fm_mix_one_ac_2	7
		fm_mix_one_ac_2	8
		fm_mix_one_ac_2	9
		fm_mix_one_ac_2	10
		fm_mix_one_ac_2	11
		fm_mix_one_ac_2	12
		fm_mix_one_ac_2	13
		fm_mix_one_ac_2	14
		fm_mix_one_ac_2	15
		fm_mix_one_ac_2	16

		lea	-16(a5),a5	; no condition codes changed (=16 samples processed)
		move.l	a5,d1							          
		swap	d1		; if( remaining_output_bytes < 16 ) 0xffff else 0x0000
		or.w	d6,d1		; if( remaining_output_bytes < 16 ) -1 else remaining_input_bytes
		move.l	d0,d7		;processed bytes
		add.l	a1,d7		;last byte to be processed
		cmp.w	d7,d1		;if( D1 < D7 ) -> stop
		bgt.w	.fm_mix2_fastloop_ac
	
		bra	.fm_mix2_underflow16
;-----------------------------------------------------------------------------------------------
; main loop 
; d1,d2,d3,d5,d7 are free
;-----------------------------------------------------------------------------------------------
.fm_mix2_fastloop:
		fm_mix_one_2	1
		fm_mix_one_2	2
		fm_mix_one_2	3
		fm_mix_one_2	4
		fm_mix_one_2	5
		fm_mix_one_2	6
		fm_mix_one_2	7
		fm_mix_one_2	8
		fm_mix_one_2	9
		fm_mix_one_2	10
		fm_mix_one_2	11
		fm_mix_one_2	12
		fm_mix_one_2	13
		fm_mix_one_2	14
		fm_mix_one_2	15
		fm_mix_one_2	16

		lea	-16(a5),a5	; no condition codes changed 	                  
		move.l	a5,d1							          
		swap	d1	; if( remaining_output_bytes < 16 ) 0xffff else 0x0000
		or.w	d6,d1	; if( remaining_output_bytes < 16 ) -1 else remaining_input_bytes
		move.l	d0,d7	;processed bytes
		add.l	a1,d7	;last byte to be processed
		cmp.w	d7,d1	;if( D1 < D7 ) -> stop
		bgt.w	.fm_mix2_fastloop

		; less than 16 bytes to go or sample ending
.fm_mix2_underflow16:
		lea	16(a5),a5 ; restore A5 with remaining bytes to be filled (we started with -16 in fast loop)

		ifne	benchmark
		move	#$444,$dff180
		endc

.fm_mix2_underflow_loop:
		move.w	a5,d1
		ble.w	.fm_mix2_buffer_full ; done, buffer filled

		fm_mix_one_ac_2	17
		
		cmp	d0,d6		;
		lea	-1(a5),a5	; samples -1 (before jumping out)
		bhi.w	.fm_mix2_underflow_loop ; (FIXME: bhs or bhi ?) (most probably the latter)
		; loop until end of instrument
;-----------------------------------------------------------------------------------------------

;-----------------------------------------------------------------------------------------------
; we land here with A5 having the number of remaining bytes
; decide whether to loop the current sample or to clear the rest of the buffer
;-----------------------------------------------------------------------------------------------
	; non-looping instruments require some extra treatment:
	;  - we have a look-ahead in here, so that 2 "future" samples are already stored
	;  - provide a "dummy" extension with "0"-samples
	;  - no problem with "repeating" instruments since the integer position is reset and
	;    new samples are loaded as needed
	MIX_FM_FADEOUT .fm_mix2_clear_rest,.fm_mix2_mainloop ;

	;-------------------------- "usual" loop code ------------------------------------------

.fm_mix2_reload_or_stop_sample:

	MIX_LOOPCHECK	.fm_mix2_clear_rest,.fm_mix2_PingPongLoop,0	;1st arg: bra target for one-shot samples, 2nd arg: bra target for ping-pong loop code

	bra.w	.fm_mix2_mainloop			;try again (TODO: check for actual end like in old routine, but this works as well)

;----------------- ping-pong loop first entry --------------------------
.fm_mix2_PingPongLoop:
	MIX_PINGPONG_INIT	.fm_mix2_Ping_backwards	;arg1: where to continue when samples are still left in output
	; caution: continues directly below at "buffer_full"
	; when we land here, pingpong is needed but the buffer is full
;-------- DONE: one buffer is filled, now just store parameters and leave
.fm_mix2_buffer_full:

	ifne	benchmark
	move	#$000,$dff180
	endc

	; check whether we reached the end of the instrument before leaving
	MIX_POSTCHECK	.fm_mix2_bfull_ret,0

.fm_mix2_clear_rest:		; no clear in second mix routine -> just bail out
.fm_mix2_bfull_ret:
	moveq	#1,d0
	rts
; no explicit clear when empty-buffer aware routines are in place
.fm_mix2_ret_nop:
	moveq	#0,d0
	rts

;-------------- Ping-Pong Loop entry point for backwards loop -------------------
;- custom routine, repeating the normal interpolation but checking against negative offsets
.fm_mix2_Ping_backwards	;yep, do loop backwards
					;in D6, the remaining length is kept (current position = rest until "0" is reached)
		move.l	a1,d1		;invert end position check
		neg.w	d1
		move.l	d1,a1
		MIX_LIMITER	d6,ping	;limit D6 to 0x7fff

		lea	-16(a5),a5 ; simplifies check for "less than 16 samples"
		moveq	#0,d5
		moveq	#0,d2	
.fm_mix2_PingB_fastloop:
		move.w	a5,d1		 ; sets condition codes but not "X"
		lea	(A1,D0.l),A3	 ; see above, (needs addq #1,d1)
		blt.w	.fm_mix2_PingB_underflow16 ; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

		;if position after 16 bytes (d0+next samples) <= start of loop, then leave here

		lea	(a3,d6.w),A3
		move.w	a3,d3
		blt.w	.fm_mix2_PingB_underflow16	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)


		rept	8
			move.b	(a0,d0.w),d2
			sub.l	d4,d0
			subx.w	d5,d0
			move.w	(A6,d2.w*2),d3
			add.w	d3,(a2)+
			move.b	(a0,d0.w),d2
			sub.l	d4,d0
			subx.w	d5,d0
			move.w	(A6,d2.w*2),d3
			add.w	d3,(a2)+
		endr
		lea	-16(a5),a5	; no condition codes changed
		bra.w	.fm_mix2_PingB_fastloop

		; slow loop, A5 almost empty and/or near end of loop area
.fm_mix2_PingB_underflow16:
		lea	16(a5),a5 ; restore A5 with remaining bytes to be filled (we started with -16 in fast loop)

		ifne	benchmark
			move	#$444,$dff180
		endc

		; NOTE: X bit still set until this point
		; less than 16 bytes to be written (bad choice of SamLEN) 
		; or (more likely) near the end of the sample
.fm_mix2_PingB_underflow_loop:
		move.w	a5,d1
		ble.w	.fm_mix2_Ping_buffer_full

		move.b	(a0,d0.w),d2	; one byte at a time, happens rarely (!!)
		sub.l	d4,d0
		subx.w	d5,d0
		move.w	(A6,d2.w*2),d3
		add.w	d3,(A2)+

		lea	-1(a5),a5	; samples -1 (before jumping out)

		move	d0,d3
		add	d6,d3
		blt.w	.fm_mix2_Ping_directionChange ; if position is at beginning of loop area, we are done here

		bra.b	.fm_mix2_PingB_underflow_loop
		; loop until end of sample

.fm_mix2_Ping_directionChange:
		; we were looping backwards, now go forwards
		MIX_PING_DIRECTIONCHANGE        ;change mixing direction offsets, issue interrupt (if necessary)
		bra.w	.fm_mix2_mainloop

.fm_mix2_Ping_buffer_full
		MIX_PING_BUFFERFULL	.fm_mix2_ping_return
.fm_mix2_Ping_return
		moveq	#1,d0
		rts


;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
;--------------------------------- 8 Bit Mixing Routine (Part1 - Move Data) --------------------------
;-----------------------------------------------------------------------------------------------------
; Input:
;        A4   - AudioStruct for current channel and panning
;        A2   - Output buffer for "SamLEN*2" shorts
;        D0   - number of samples (SamLEN*2)
; Output:
;        (A2) - Output samples
;        D0   - 1=mixing done, 0=mixing skipped (no sample played)
;
; Used Registers: all, none saved
;        D0   - current offset towards A0 (upper word is current fractional position)
;        D4   - interpolation step size, lower word integer part, upper word fractional part
;        A0   - current source sample, offset by (already) played bytes
;        A5   - remaining samples (except in fast loop, where it is -16)
;        A6   - Volume Table
; Notes:
;  - keep in mind that the arithmetics are done to preserve the "X" bit which is used to carry the
;    fractional offset overflows to D0, hence beware of integer "add/sub" within the loops
;
;-----------------------------------------------------------------------------------------------------
mix:
; d7 is free
; d5 could be freed, replace add.l d5,d0 by sub.w d4,d0 add.l d4,d0 (both)

	ifne	benchmark
	move	#$f00,$dff180
	endc

	MIX_INIT	.mix_ret_nop,1,0,0 ; set up registers (return address, volume table adjust=yes, single byte offsets in mixing = 8 Bit inputs)
	moveq	#0,d3			;current processed sample, needs to be zero
	moveq	#0,d2			;current input sample, zero because of implicit byte->word mapping

	btst	#ASLoopF_BackLooping,AS_CurrentFPOS+2(a4) ; D6 -> stop position for backward loop
	bne.w	.mix_Ping_backwards

	MIX_REMBYTES	d6,d1,0		;total length - played length = remaining bytes, 2nd arg: scratch register, 3rd arg: 1=16 Bit mode, 0=8 Bit mode

;-----------------------------------------------------------------------------------------------
; main loop: check vital parameters and time-scale the inputs
;-----------------------------------------------------------------------------------------------
.mix_mainloop:				;start point for "fast" loop and "repeated" samples
	MIX_LIMITER	d6,norm		;limit remaining samples in D6 to $7fff (16 bit signed max.)

	ifne	benchmark
		move	#$00f,$dff180
	endc
	
	lea	-16(a5),a5 ; simplifies check for "less than 16 samples" (will be corrected after fast loop)

	sub.w	d4,d0	   ; integer position decrement 
	add.l	d4,d0	   ; fractional position increment -> sets X bit (and restores integer position)

	lea	(A1,D0.l),A3	 ; see above in MIX_INIT macro, (needs addq #2,d1)

	;max. sample length + current position in sample >= end of sample ? (when 16 samples would follow)
	cmp.w	A3,d6			; sets condition codes but not "X"
	blo.w	.mix_underflow16	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

	move.w	a5,d1		 ; sets condition codes but not "X" (tst.w does not work on 68000)
	blt.w	.mix_underflow16 ; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

	; d1,d5,d7 are free

.mix_fastloop:
		rept	8
		 move.b	(a0,d0.w),d2					;
		 addx.l	d4,d0						;
		 move.w	(A6,d2.w*2),d3
		 move.b	(a0,d0.w),d2					;
		 addx.l	d4,d0						;
		 swap	d3						;
		 move.w	(a6,d2.w*2),d3					;
		 move.l	d3,(a2)+					;
		endr

		lea	-16(a5),a5	; no condition codes changed 	                  ;8

		move.l	a5,d1							          ;4
		swap	d1	; if( remaining_output_bytes < 16 ) 0xffff else 0x0000    ;4
		or.w	d6,d1	; if( remaining_output_bytes < 16 ) -1 else               ;4 
				;                                   remaining_input_bytes
		lea	(A1,D0.l),A3	; input: processed bytes                          ;12
		cmp.w	A3,d1	; if( D1 < A3 ) -> stop		                          ;6
		bgt.w	.mix_fastloop                                                     ;10
											  ;=48 cycles every 16 bytes
											  ; total 8*84+48 for 16 output bytes = 45 cycles/byte
		; less than 16 bytes to go or sample ending
.mix_underflow16:
	lea	16(a5),a5 ; restore A5 with remaining bytes to be filled (we started with -16 in fast loop)

	ifne	benchmark
	move	#$444,$dff180
	endc

	; NOTE: X bit still set until this point
	; less than 16 bytes to be written (bad choice of SamLEN) 
	; or (more likely) near the end of the sample

.mix_underflow_loop:
		; TODO: optimize, rarely used for PT alike but a burden with chiptunes
		move.w	a5,d1
		ble.w	.mix_buffer_full ; done, buffer filled

		move.b	(a0,d0.w),d2
		addx.l	d4,d0
		move.w	(A6,d2.w*2),(A2)+
		
		cmp	d0,d6		; keeps X bit
		lea	-1(a5),a5	; samples -1 (before jumping out)
		bhi.b	.mix_underflow_loop ; (FIXME: bhs or bhi ?) (most probably the latter)
		; loop until end of sample

	add.w	d4,d0	   ; integer position increment 
	sub.l	d4,d0	   ; fractional position decrement -> sets X bit (and restores integer position)

;--------------------------
	; we land here with A5 having the number of remaining bytes
	; decide whether to loop the current sample or to clear the rest of the buffer
	MIX_LOOPCHECK	.mix_clear_rest,.mix_PingPongLoop,0 ;1st arg: bra target for one-shot samples, 2nd arg: bra target for ping-pong loop code;3rd arg: 16 bit mode
	bra.w	.mix_mainloop			;try again (TODO: check for actual end like in old routine, but this works as well)

;----------------- ping-pong loop first entry --------------------------
.mix_PingPongLoop
	MIX_PINGPONG_INIT	.mix_Ping_backwards	;arg1: where to continue when samples are still left in output
	; caution: continues directly below at "buffer_full"
	; when we land here, pingpong is needed but the buffer is full

;-------- DONE: one buffer is filled, now just store parameters and leave
.mix_buffer_full:
	add.w	d4,d0	   ; integer position increment 
	sub.l	d4,d0	   ; fractional position decrement -> sets X bit (and restores integer position)

	ifne	benchmark
	move	#$000,$dff180
	endc

	MIX_POSTCHECK   .mix_bfull_ret,0

.mix_bfull_ret:
	moveq	#1,d0
	rts

	; clear remaining bytes
.mix_clear_rest:
		move.l	A5,d1		;remaining bytes
		moveq	#$0,d0
		subq	#1,d1		;for dbf loop
		bmi.s	.mix_ret_clear
.mix_clear_loop:
		move	d0,(a2)+
		dbf	d1,.mix_clear_loop
.mix_ret_clear
		moveq	#1,d0
		rts

; no explicit clear when empty-buffer aware routines are in place
.mix_ret_nop:
		moveq	#0,d0
		rts

;-------------- Ping-Pong Loop entry point for backwards loop -------------------
;- custom routine, repeating the normal interpolation but checking against negative offsets
.mix_Ping_backwards	;yep, do loop backwards

					;in D6, the remaining length is kept (current position = rest until "0" is reached)
		move.l	a1,d1		;invert end position check
		neg.w	d1
		move.l	d1,a1
		MIX_LIMITER	d6,ping	;limit D6 to 0x7fff

		lea	-16(a5),a5 ; simplifies check for "less than 16 samples"
		add.w	d4,d0	   ; integer position increment 
		sub.l	d4,d0	   ; fractional position decrement -> sets X bit (and restores integer position)
.mix_PingB_fastloop:
		move.w	a5,d1		 ; sets condition codes but not "X"
		lea	(A1,D0.l),A3	 ; see above, (needs addq #1,d1)
		blt.w	.mix_PingB_underflow16 ; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

		;if position after 16 bytes (d0+next samples) <= start of loop, then leave here
		cmp.w	A3,d6			; sets condition codes but not "X"
		bhs.w	.mix_PingB_underflow16	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

		rept	8
			move.b	(a0,d0.w),d2
			subx.l	d4,d0
			move.w	(A6,d2.w*2),d3
			swap	d3
			move.b	(a0,d0.w),d2
			subx.l	d4,d0
			move.w	(A6,d2.w*2),d3
			move.l	d3,(a2)+
		endr
		lea	-16(a5),a5	; no condition codes changed
		bra.w	.mix_PingB_fastloop

		; slow loop, A5 almost empty and/or near end of loop area
.mix_PingB_underflow16:
		lea	16(a5),a5 ; restore A5 with remaining bytes to be filled (we started with -16 in fast loop)

		ifne	benchmark
			move	#$444,$dff180
		endc

		; NOTE: X bit still set until this point
		; less than 16 bytes to be written (bad choice of SamLEN) 
		; or (more likely) near the end of the sample
.mix_PingB_underflow_loop:
		move.w	a5,d1
		ble.w	.mix_Ping_buffer_full

		move.b	(a0,d0.w),d2	; one byte at a time, happens rarely (!!)
		subx.l	d4,d0
		move.w	(A6,d2.w*2),(A2)+

		lea	-1(a5),a5	; samples -1 (before jumping out)

		cmp.w	d0,d6			; sets condition codes but not "X"
		bhs.w	.mix_Ping_directionChange ; if position is at beginning of loop area, we are done here

		bra.b	.mix_PingB_underflow_loop
		; loop until end of sample

.mix_Ping_directionChange:
		; we were looping backwards, now go forwards
		MIX_PING_DIRECTIONCHANGE	;change mixing direction offsets, issue interrupt (if necessary)
		bra.w	.mix_mainloop

.mix_Ping_buffer_full:
		MIX_PING_BUFFERFULL	.mix_ping_return
.mix_Ping_return
		moveq	#1,d0
		rts
mix_int:
		movem.l	d0-a6,-(sp)		;nur zur Sicherheit
		move.l	AS_Int(A4),a1
		move.l	4.w,a6
		jsr	_LVOCause(a6)
		movem.l	(sp)+,d0-a6		;man weiss ja nie...
		rts


;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
;--------------------------------- 8 Bit Mixing Routine (Part2 - ADD Data) ---------------------------
;-----------------------------------------------------------------------------------------------------
; NEW ROUTINE
; Input:
;        A4   - AudioStruct for current channel and panning
;        A2   - Output buffer for "SamLEN*2" shorts
;        D0   - number of samples (SamLEN*2)
; Output:
;        (A2) - Output samples
;        D0   - 1=mixing done, 0=mixing skipped (no sample played)
;
; Used Registers: all, none saved
;        D0   - current offset towards A0 (upper word is current fractional position)
;        D4   - interpolation step size, lower word integer part, upper word fractional part
;        A0   - current source sample, offset by (already) played bytes
;        A5   - remaining samples (except in fast loop, where it is -16)
;        A6   - volume table
; Notes:
;  - keep in mind that the arithmetics are done to preserve the "X" bit which is used to carry the
;    fractional offset overflows to D0, hence beware of integer "add/sub" within the loops
;
;-----------------------------------------------------------------------------------------------------
mix2:
; d7 is free

	ifne	benchmark
	move	#$ff0,$dff180
	endc

	MIX_INIT	.mix2_nop_return,1,0,0 ; set up registers (return address, volume table adjust=yes, single byte offsets in mixing = 8 Bit inputs)
	moveq	#0,d3			;current processed sample, needs to be zero
	moveq	#0,d2			;current input sample (will be loaded as bytes and treated as words), hence 0

	btst	#ASLoopF_BackLooping,AS_CurrentFPOS+2(a4) ; D6 -> stop position for backward loop
	bne.w	.mix2_Ping_backwards

	MIX_REMBYTES	d6,d1,0		;total length - played length = remaining bytes, 2nd arg: scratch register, 3rd arg: 1=16 Bit mode, 0=8 Bit mode

;-----------------------------------------------------------------------------------------------
; main loop: check vital parameters and time-scale the inputs
;-----------------------------------------------------------------------------------------------
.mix2_mainloop:				;start point for "fast" loop and "repeated" samples
	MIX_LIMITER	d6,norm		;limit remaining samples in D6 to $7fff (16 bit signed max.)

;	ifne	benchmark
;		move	#$00f,$dff180
;	endc
	
	lea	-16(a5),a5 ; simplifies check for "less than 16 samples"

	sub.w	d4,d0	   ; integer position decrement 
	add.l	d4,d0	   ; fractional position increment -> sets X bit (and restores integer position)

	lea	(A1,D0.l),A3	 ; see above, (needs addq #1,d1)
	;max. sample length + current position in sample >= end of sample ?
	cmp.w	A3,d6			; sets condition codes but not "X"
	blo.w	.mix2_underflow16	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

	move.w	a5,d1		 ; sets condition codes but not "X" (tst.w does not work on 68000)
	blt.w	.mix2_underflow16 ; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)


.mix2_fastloop:
		rept	8
		; TODO: interleave
		 move.w	(a2),a3		;old buffer state
		 move.b	(a0,d0.w),d2
		 addx.l	d4,d0
		 adda.w	(A6,d2.w*2),a3  ; this add doesn`t change X bit (!)
		 move.w	A3,(a2)+

		 move.w	(a2),a3		;old buffer state
		 move.b	(a0,d0.w),d2
		 addx.l	d4,d0
		 adda.w	(A6,d2.w*2),a3
		 move.w	A3,(a2)+
		endr

		lea	-16(a5),a5	; no condition codes changed (TODO: suba.l dx,A5)

		move.l	a5,d1
		swap	d1	; if( remaining_output_bytes < 16 ) 0xffff else 0x0000
		or.w	d6,d1	; if( remaining_output_bytes < 16 ) -1 else
				;                                   remaining_input_bytes
		lea	(A1,D0.l),A3	; input: processed bytes
		cmp.w	A3,d1	; if( D1 < A3 ) -> stop
		bgt.w	.mix2_fastloop

	; less than 16 bytes to go or sample ending
.mix2_underflow16:
	lea	16(a5),a5 ; restore A5 with remaining bytes to be filled (we started with -16 in fast loop)

	ifne	benchmark
	move	#$444,$dff180
	endc

	; NOTE: X bit still set until this point
	; less than 16 bytes to be written (bad choice of SamLEN) 
	; or (more likely) near the end of the sample

.mix2_underflow_loop:
	move.w	a5,d1
	ble.w	.mix2_buffer_full ; done, buffer filled

	move.w	(a2),a3
	move.b	(a0,d0.w),d2
	addx.l	d4,d0
	adda.w	(A6,d2.w*2),a3
	move.w	a3,(a2)+
		
	cmp	d0,d6		; keeps X bit
	lea	-1(a5),a5	; samples -1 (before jumping out)
	bhi.b	.mix2_underflow_loop ; (FIXME: bhs or bhi ?) (bhi !)
	; loop until end of sample

	add.w	d4,d0	   ; integer position increment 
	sub.l	d4,d0	   ; fractional position decrement -> sets X bit (and restores integer position)

;--------------------------
	; we land here with A5 having the number of remaining bytes
	; decide whether to loop the current sample or to clear the rest of the buffer
	MIX_LOOPCHECK	.mix2_rts2,.mix2_PingPongLoop,0 ;1st arg: bra target for one-shot samples, 2nd arg: bra target for ping-pong loop code;3rd arg: 16 bit mode
	bra.w	.mix2_mainloop			;try again (TODO: check for actual end like in old routine, but this works as well)

;----------------- ping-pong loop first entry --------------------------
.mix2_PingPongLoop
	MIX_PINGPONG_INIT	.mix2_Ping_backwards	;arg1: where to continue when samples are still left in output
	; caution: continues directly below at "buffer_full"
	; when we land here, pingpong is needed but the buffer is full

;-------- DONE: one buffer is filled, now just store parameters and leave
.mix2_buffer_full:
	add.w	d4,d0	   ; integer position increment 
	sub.l	d4,d0	   ; fractional position decrement -> sets X bit (and restores integer position)

	ifne	benchmark
	move	#$000,$dff180
	endc

	MIX_POSTCHECK	.mix2_rts2,0

.mix2_rts2:
	moveq	#1,d0			;OK, we have done something
	rts
.mix2_nop_return:
	moveq	#0,d0
	rts

;-------------- Ping-Pong Loop entry point for backwards loop -------------------
;- custom routine, repeating the normal interpolation but checking against negative offsets
.mix2_Ping_backwards	;yep, do loop backwards

					;in D6, the remaining length is kept (current position = rest until "0" is reached)
		move.l	a1,d1		;invert end position check
		neg.w	d1
		move.l	d1,a1
		MIX_LIMITER	d6,ping	;limit D6 to 0x7fff

		lea	-16(a5),a5 ; simplifies check for "less than 16 samples"
		add.w	d4,d0	   ; integer position increment 
		sub.l	d4,d0	   ; fractional position decrement -> sets X bit (and restores integer position)
.mix2_PingB_fastloop:
		move.w	a5,d1		 ; sets condition codes but not "X"
		lea	(A1,D0.l),A3	 ; see above, (needs addq #1,d1)
		blt.w	.mix2_PingB_underflow16 ; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

		;if position after 16 bytes (d0+next samples) <= start of loop, then leave here
		cmp.w	A3,d6			; sets condition codes but not "X"
		bhs.w	.mix2_PingB_underflow16	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

		rept	8
			move.w	(a2),a3
			move.b	(a0,d0.w),d2
			subx.l	d4,d0
			adda.w	(A6,d2.w*2),a3
			move.w	A3,(a2)+

			move.w	(a2),a3
			move.b	(a0,d0.w),d2
			subx.l	d4,d0
			adda.w	(A6,d2.w*2),a3
			move.w	A3,(a2)+
		endr
		lea	-16(a5),a5	; no condition codes changed
		bra.w	.mix2_PingB_fastloop

		; slow loop, A5 almost empty and/or near end of loop area
.mix2_PingB_underflow16:
		lea	16(a5),a5 ; restore A5 with remaining bytes to be filled (we started with -16 in fast loop)

		ifne	benchmark
			move	#$444,$dff180
		endc

		; NOTE: X bit still set until this point
		; less than 16 bytes to be written (bad choice of SamLEN) 
		; or (more likely) near the end of the sample
.mix2_PingB_underflow_loop:
		move.w	a5,d1
		ble.w	.mix2_Ping_buffer_full

		move.w	(a2),a3
		move.b	(a0,d0.w),d2	; one byte at a time, happens rarely (!!)
		subx.l	d4,d0
		adda.w	(A6,d2.w*2),a3
		move.w	a3,(a2)+

		lea	-1(a5),a5	; samples -1 (before jumping out)

		cmp.w	d0,d6			; sets condition codes but not "X"
		bhs.w	.mix2_Ping_directionChange ; if position is at beginning of loop area, we are done here

		bra.b	.mix2_PingB_underflow_loop
		; loop until end of sample

.mix2_Ping_directionChange:
		; we were looping backwards, now go forwards
		MIX_PING_DIRECTIONCHANGE        ;change mixing direction offsets, issue interrupt (if necessary)
		bra.w	.mix2_mainloop

.mix2_Ping_buffer_full
		MIX_PING_BUFFERFULL	.mix2_ping_return
.mix2_Ping_return
		moveq	#1,d0
		rts


;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
;--------------------------------- 16 Bit Mixing Routine (Part1 - Move Data) -------------------------
;-----------------------------------------------------------------------------------------------------
; NEW ROUTINE
; Input:
;        A4   - AudioStruct for current channel and panning
;        A2   - Output buffer for "SamLEN*2" shorts
;        D0   - number of samples (SamLEN*2)
; Output:
;        (A2) - Output samples
;        D0   - 1=mixing done, 0=mixing skipped (no sample played)
;
; Used Registers: all, none saved
;        D0   - current offset towards A0 (upper word is current fractional position)
;        D4   - interpolation step size, lower word integer part, upper word fractional part
;        A0   - current source sample, offset by (already) played bytes
;        A5   - remaining samples (except in fast loop, where it is -16)
;        A6   - Volume Table
; Notes:
;  - keep in mind that the arithmetics are done to preserve the "X" bit which is used to carry the
;    fractional offset overflows to D0, hence beware of integer "add/sub" within the loops
;  - ping-pong loop not working yet, needs to be ported to 16 bit semantics (!)
;-----------------------------------------------------------------------------------------------------
mix16:
; d7 is free

	ifne	benchmark
	move	#$333,$dff180
	endc

	; init macro
	MIX_INIT	.mix16_ret_nop,0,1,0	; set up registers (common routine to all mixers, no volume table, add offset twice)

	moveq	#0,d2			;
	moveq	#0,d3			;current processed sample, needs to be zero

	btst	#ASLoopF_BackLooping,AS_CurrentFPOS+2(a4) ; D6 -> stop position for backward loop
	bne.w	.mix16_Ping_backwards

	MIX_REMBYTES	d6,d1,1		;total length - played length = remaining bytes, 2nd arg: scratch register, 3rd arg: 1=16 Bit mode, 0=8 Bit mode

; FIXME: really necessary ? In essence, this code fragment loops the sample in total without checking flags, disabled for now
;	bpl.s	.mix16_mainloop		;not beyond sample length -> skip
;	move.l	AS_SampleSize(a4),d6	;fetch total length
;	move.l	AS_Currentadr(a4),a0	;start at beginning of sample

	; main loop for 16 samples in one run 
.mix16_mainloop:				;start point for "fast" loop and "repeated" samples
	MIX_LIMITER	d6,norm		;limit remaining samples in D6 to $7fff (16 bit signed max.)

	ifne	benchmark
		move	#$444,$dff180
	endc
	
	lea	-16(a5),a5 ; simplifies check for "less than 16 samples"

	ifne	fastmix
	sub.w	d4,d0	   ; integer position decrement 
	add.l	d4,d0	   ; fractional position increment -> sets X bit (and restores integer position)
	else
	moveq	#0,d5
	endc
	
	lea	(A1,D0.l),A3	 ; see above, (needs addq #1,d1)
	;max. sample length + current position in sample >= end of sample ?
	cmp.w	A3,d6			; sets condition codes but not "X"
	blo.w	.mix16_underflow16	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

	move.w	a5,d1		 ; sets condition codes but not "X" (tst.w does not work on 68000)
	blt.w	.mix16_underflow16 ; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

; d1,d5,d7 are free
; d0 starts with zero in lower word

.mix16_fastloop:
		rept	8
		 move.w	(a0,d0.w*2),d2					;
		 ifne	fastmix
		 addx.l	d4,d0						;
		 else
		 add.l	d4,d0						;
		 addx.l	d5,d0
		 endc
		 move.w	(A6,d2.l*2),d3
		 move.w	(a0,d0.w*2),d2					;
		 ifne	fastmix
		 addx.l	d4,d0						;
		 else
		 add.l	d4,d0						;
		 addx.l	d5,d0
		 endc
		 swap	d3						;
		 move.w	(a6,d2.l*2),d3					;
		 move.l	d3,(a2)+					;
		endr

		lea	-16(a5),a5	; no condition codes changed 	                  ;8

		move.l	a5,d1							          ;4
		swap	d1	; if( remaining_output_bytes < 16 ) 0xffff else 0x0000    ;4
		or.w	d6,d1	; if( remaining_output_bytes < 16 ) -1 else               ;4 
				;                                   remaining_input_bytes
		lea	(A1,D0.l),A3	; input: processed bytes                          ;12
		cmp.w	A3,d1	; if( D1 < A3 ) -> stop		                          ;6
		bgt.w	.mix16_fastloop                                                     ;10
											  ;=48 cycles every 16 bytes
											  ; total 8*84+48 for 16 output bytes = 45 cycles/byte
		; less than 16 bytes to go or sample ending
.mix16_underflow16:
	lea	16(a5),a5 ; restore A5 with remaining bytes to be filled (we started with -16 in fast loop)

	ifne	benchmark
	move	#$444,$dff180
	endc

	; NOTE: X bit still set until this point
	; less than 16 bytes to be written (bad choice of SamLEN) 
	; or (more likely) near the end of the sample

.mix16_underflow_loop:
		; TODO: optimize, rarely used for PT alike but a burden with chiptunes
		move.w	a5,d1
		ble.w	.mix16_buffer_full ; done, buffer filled

		move.w	(a0,d0.w*2),d2
		ifne	fastmix
		 addx.l	d4,d0
		else
		 add.l	d4,d0
		 addx.l	d5,d0
		endc
		move.w	(A6,d2.l*2),(A2)+
		
		cmp	d0,d6		; keeps X bit
		lea	-1(a5),a5	; samples -1 (before jumping out)
		bhi.b	.mix16_underflow_loop ; (FIXME: bhs or bhi ?) (most probably the latter)
		; loop until end of sample

	ifne	fastmix
	 add.w	d4,d0	   ; integer position increment 
	 sub.l	d4,d0	   ; fractional position decrement -> sets X bit (and restores integer position)
	endc
;	subx.l	d1,d0

;--------------------------
	; we land here with A5 having the number of remaining bytes
	; decide whether to loop the current sample or to clear the rest of the buffer
	MIX_LOOPCHECK   .mix16_clear_rest,.mix16_PingPongLoop,1   ;1st arg: bra target for one-shot samples, 2nd arg: bra target for ping-pong loop code, 3rd arg: 16 bit mode (1)
	bra.w	.mix16_mainloop			;try again (TODO: check for actual end like in old routine, but this works as well)

;----------------- ping-pong loop first entry --------------------------
.mix16_PingPongLoop
	MIX_PINGPONG_INIT	.mix16_Ping_backwards,1

;-------- DONE: one buffer is filled, now just store parameters and leave
.mix16_buffer_full:
;	moveq	#0,d1
	ifne	fastmix
	 add.w	d4,d0	   ; integer position increment 
	 sub.l	d4,d0	   ; fractional position decrement -> sets X bit (and restores integer position)
	endc
;	subx.l	d1,d0

	ifne	benchmark
	move	#$000,$dff180
	endc

	MIX_POSTCHECK   .mix16_bfull_ret,1

.mix16_bfull_ret:
	moveq	#1,d0
	rts

	; clear remaining bytes
.mix16_clear_rest:
		move.l	A5,d1		;remaining bytes
		moveq	#$0,d0
		subq	#1,d1		;for dbf loop
		bmi.s	.mix16_ret_clear
.mix16_clear_loop:
		move	d0,(a2)+
		dbf	d1,.mix16_clear_loop
.mix16_ret_clear
		moveq	#1,d0
		rts

; no explicit clear when empty-buffer aware routines are in place
.mix16_ret_nop:
		moveq	#0,d0
		rts

;-------------- Ping-Pong Loop entry point for backwards loop -------------------
;- custom routine, repeating the normal interpolation but checking against negative offsets
.mix16_Ping_backwards	;yep, do loop backwards
					;in D6, the remaining length is kept (current position = rest until "0" is reached)
		move.l	a1,d1		;invert end position check
		neg.w	d1
		move.l	d1,a1
		MIX_LIMITER	d6,ping	;limit D6 to 0x7fff

		lea	-16(a5),a5 ; simplifies check for "less than 16 samples"
		add.w	d4,d0	   ; integer position increment 
		sub.l	d4,d0	   ; fractional position decrement -> sets X bit (and restores integer position)
.mix16_PingB_fastloop:
		move.w	a5,d1		 ; sets condition codes but not "X"
		lea	(A1,D0.l),A3	 ; see above, (needs addq #1,d1)
		blt.w	.mix16_PingB_underflow16 ; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

		;if position after 16 bytes (d0+next samples) <= start of loop, then leave here
		cmp.w	A3,d6			; sets condition codes but not "X"
		bhs.w	.mix16_PingB_underflow16	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

		rept	8
			move.b	(a0,d0.w),d2
			subx.l	d4,d0
			move.w	(A6,d2.w*2),d3
			swap	d3
			move.b	(a0,d0.w),d2
			subx.l	d4,d0
			move.w	(A6,d2.w*2),d3
			move.l	d3,(a2)+
		endr
		lea	-16(a5),a5	; no condition codes changed
		bra.w	.mix16_PingB_fastloop

		; slow loop, A5 almost empty and/or near end of loop area
.mix16_PingB_underflow16:
		lea	16(a5),a5 ; restore A5 with remaining bytes to be filled (we started with -16 in fast loop)

		ifne	benchmark
			move	#$444,$dff180
		endc

		; NOTE: X bit still set until this point
		; less than 16 bytes to be written (bad choice of SamLEN) 
		; or (more likely) near the end of the sample
.mix16_PingB_underflow_loop:
		move.w	a5,d1
		ble.w	.mix16_Ping_buffer_full

		move.b	(a0,d0.w),d2	; one byte at a time, happens rarely (!!)
		subx.l	d4,d0
		move.w	(A6,d2.w*2),(A2)+

		lea	-1(a5),a5	; samples -1 (before jumping out)

		cmp.w	d0,d6			; sets condition codes but not "X"
		bhs.w	.mix16_Ping_directionChange ; if position is at beginning of loop area, we are done here

		bra.b	.mix16_PingB_underflow_loop
		; loop until end of sample

.mix16_Ping_directionChange:
		; we were looping backwards, now go forwards
		MIX_PING_DIRECTIONCHANGE
		bra.w	.mix16_mainloop

.mix16_Ping_buffer_full:
		MIX_PING_BUFFERFULL	.mix16_Ping_return
.mix16_Ping_return
		moveq	#1,d0
		rts


;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
;--------------------------- 16 Bit Filtered Mixing Routine (Part1 - Move Data) ----------------------
;-----------------------------------------------------------------------------------------------------
; Input:
;        A4   - AudioStruct for current channel and panning
;        A2   - Output buffer for "SamLEN*2" shorts
;        D0   - number of samples (SamLEN*2)
;        A6   - Volume table
; Output:
;        (A2) - Output samples
;        D0   - 1=mixing done, 0=mixing skipped (no sample played)
;
; Used Registers: all, none saved
;        D0   - current offset towards A0 (upper word is current fractional position)
;        D4   - interpolation step size, lower word integer part, upper word fractional part
;        D5   - upper word of D4, 0 in lower word
;        A0   - current source sample, offset by (already) played bytes
;        A5   - remaining samples (except in fast loop, where it is -16)
;        A6   - Volume Table
; Notes:
;  - 4 Tap FIR interpolation filter (Windowed Sinc), tradeoff between multiplies per output sample 
;    and sound quality
;  - TODO: ping-pong backwards preload and FM mix, currently uses nearest neighbor
;-----------------------------------------------------------------------------------------------------
fm16_mix:
	ifne	benchmark
	move	#$f00,$dff180
	endc

	; filtered mixing routine can`t handle downsampling, refer to regular "mix" in those cases
	move.w	AS_Period(a4),d6	;current period
	beq.w	.fm16_mix_ret_nop
	cmp.w	mixingperiod4,d6	;must be larger than mixing period, else use regular routine
;	blt.w	mix16			;ok, period is wrong, use regular mixer	
	ble.w	fm16down_mix	; use downmixing routine 
;blt.w

	; init macro
	MIX_INIT	.fm16_mix_ret_nop,0,1,0 ; set up registers (common routine to all mixers, no volume table, add offset twice)

	btst	#ASLoopF_FMInit,AS_CurrentFPOS+2(a4) 
	bne.s	.fm16_mix_fminitialized	; Filtered Mixing initialized ?
	bsr.w	.fm16_mix_preload		; preload samples into filter buffer
.fm16_mix_fminitialized:

	moveq	#0,d2			; irrelevant ?!
	moveq	#0,d3			;

	btst	#ASLoopF_BackLooping,AS_CurrentFPOS+2(a4) ; D6 -> stop position for backward loop
	bne.w	.fm16_mix_Ping_backwards

	MIX_REMBYTES	d6,d1,1		;total length - played length = remaining bytes, 2nd arg: scratch register, 3rd arg: 1=16 Bit mode, 0=8 Bit mode

;-----------------------------------------------------------------------------------------------
; main loop: check vital parameters and time-scale the inputs
;-----------------------------------------------------------------------------------------------
.fm16_mix_mainloop:			;start point for "fast" loop and "repeated" samples
	MIX_LIMITER	d6,norm		;limit remaining samples in D6 to $7fff (16 bit signed max.)

	ifne	benchmark
	move	#$00f,$dff180
	endc
	

	moveq	#0,d3
	move.b	AS_FMPos(a4),d3			;ringbuffer index
	lsl.w	#FMB_Accuracy+FMB_Taps+1,d3	;accuracy bits + taps bits + word size
	lea	FM_tab_100,a3
	add.l	d3,a3				;table for current ringbuffer position

	lea	-16(a5),a5 ; simplifies check for "less than 16 samples" to generate

	move.l	d0,d7			; current position in instrument
	add.l	a1,d7			; + 16 output samples wrt. sampling rate
	cmp.w	d7,d6			; >= end of instrument
	blo.w	.fm16_mix_underflow16	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)
	move.w	a5,d1			; 16 samples left to write in output buffer ?
	blt.w	.fm16_mix_underflow16 	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

;-----------------------------------------------------------------------------------------------
; main loop
; d1,d2,d3,d5,d7 are free
;-----------------------------------------------------------------------------------------------
.fm16_mix_fastloop:
	fm16_mix_one	1
	fm16_mix_one	2
	fm16_mix_one	3
	fm16_mix_one	4
	fm16_mix_one	5
	fm16_mix_one	6
	fm16_mix_one	7
	fm16_mix_one	8
	fm16_mix_one	9
	fm16_mix_one	10
	fm16_mix_one	11
	fm16_mix_one	12
	fm16_mix_one	13
	fm16_mix_one	14
	fm16_mix_one	15
	fm16_mix_one	16

		lea	-16(a5),a5	; no condition codes changed (=16 samples processed)
		move.l	a5,d1							          
		swap	d1		; if( remaining_output_bytes < 16 ) 0xffff else 0x0000
		or.w	d6,d1		; if( remaining_output_bytes < 16 ) -1 else remaining_input_bytes
		move.l	d0,d7		;processed bytes
		add.l	a1,d7		;last byte to be processed
		cmp.w	d7,d1		;if( D1 < D7 ) -> stop
		bgt.w	.fm16_mix_fastloop
;-----------------------------------------------------------------------------------------------

;-----------------------------------------------------------------------------------------------
; less than 16 bytes to go or sample ending
;-----------------------------------------------------------------------------------------------
.fm16_mix_underflow16:
		lea	16(a5),a5 ; restore A5 with remaining bytes to be filled (we started with -16 in fast loop)

		ifne	benchmark
		move	#$444,$dff180
		endc

.fm16_mix_underflow_loop:
		move.w	a5,d1
		ble.w	.fm16_mix_buffer_full ; done, buffer filled

	fm16_mix_one	17
		
		cmp	d0,d6		;
		lea	-1(a5),a5	; samples -1 (before jumping out)
		bhi.b	.fm16_mix_underflow_loop ; 
		; loop until end of instrument
;-----------------------------------------------------------------------------------------------

;-----------------------------------------------------------------------------------------------
; we land here with A5 having the number of remaining bytes
; decide whether to loop the current sample or to clear the rest of the buffer
;-----------------------------------------------------------------------------------------------

	; non-looping instruments require some extra treatment:
	;  - we have a look-ahead in here, so that 2 "future" samples are already stored
	;  - provide a "dummy" extension with "0"-samples
	;  - no problem with "repeating" instruments since the integer position is reset and
	;    new samples are loaded as needed
	MIX_FM_FADEOUT .fm16_mix_clear_rest,.fm16_mix_mainloop ;

	;-------------------------- "usual" loop code ------------------------------------------

	MIX_LOOPCHECK	.fm16_mix_clear_rest,.fm16_mix_PingPongLoop,1	;1st arg: bra target for one-shot samples, 2nd arg: bra target for ping-pong loop code

	bra.w	.fm16_mix_mainloop

;----------------- ping-pong loop first entry --------------------------
.fm16_mix_PingPongLoop
	MIX_PINGPONG_INIT	.fm16_mix_Ping_backwards,1

;-------- DONE: one buffer is filled, now just store parameters and leave
.fm16_mix_buffer_full:

	ifne	benchmark
	move	#$000,$dff180
	endc

	; check whether we reached the end of the instrument before leaving
	MIX_POSTCHECK	.fm16_mix_bfull_ret,1

.fm16_mix_bfull_ret:
	moveq	#1,d0
	rts

;--------------------- clear remaining bytes - TODO: filter here also ----------------
.fm16_mix_clear_rest:
		move.l	A5,d1		;remaining bytes
		moveq	#$0,d0
		subq	#1,d1		;for dbf loop
		bmi.s	.fm16_mix_ret_clear
.fm16_mix_clear_loop:
		move	d0,(a2)+
		dbf	d1,.fm16_mix_clear_loop
.fm16_mix_ret_clear
		moveq	#1,d0
		rts

; no explicit clear when empty-buffer aware routines are in place
.fm16_mix_ret_nop:
		moveq	#0,d0
		rts

;-------------- Ping-Pong Loop entry point for backwards loop -------------------
;- custom routine, repeating the normal interpolation but checking against negative offsets
.fm16_mix_Ping_backwards	;yep, do loop backwards

		;neg.w	d4		;count integer positions backwards (subx instead of addx for offsets)
					;in D6, the remaining length is kept (current position = rest until "0" is reached)
		move.l	a1,d1		;invert end position check
		neg.w	d1
		move.l	d1,a1
		
		MIX_LIMITER	d6,ping	;limit D6 to 0x7fff

		lea	-16(a5),a5 ; simplifies check for "less than 16 samples"
		moveq	#0,d5
		moveq	#0,d2
.fm16_mix_PingB_fastloop:
		move.w	a5,d1		 ; sets condition codes but not "X"
		lea	(A1,D0.l),A3	 ; see above, (needs addq #1,d1)
		blt.w	.fm16_mix_PingB_underflow16 ; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

		;if position after 16 bytes (d0+next samples) <= start of loop, then leave here
		lea	(a3,d6.w),A3
		move.w	a3,d3
		blt.w	.fm16_mix_PingB_underflow16	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

		rept	8
			move.b	(a0,d0.w),d2
			sub.l	d4,d0
			subx.w	d5,d0
			move.w	(A6,d2.w*2),d3
			swap	d3
			move.b	(a0,d0.w),d2
			sub.l	d4,d0
			subx.w	d5,d0
			move.w	(A6,d2.w*2),d3
			move.l	d3,(a2)+
		endr
		lea	-16(a5),a5	; no condition codes changed
		bra.w	.fm16_mix_PingB_fastloop

		; slow loop, A5 almost empty and/or near end of loop area
.fm16_mix_PingB_underflow16:
		lea	16(a5),a5 ; restore A5 with remaining bytes to be filled (we started with -16 in fast loop)

		ifne	benchmark
			move	#$444,$dff180
		endc

		; NOTE: X bit still set until this point
		; less than 16 bytes to be written (bad choice of SamLEN) 
		; or (more likely) near the end of the sample
.fm16_mix_PingB_underflow_loop:
		move.w	a5,d1
		ble.w	.fm16_mix_Ping_buffer_full

		move.b	(a0,d0.w),d2	; one byte at a time, happens rarely (!!)
		sub.l	d4,d0
		subx.l	d5,d0
		move.w	(A6,d2.w*2),(A2)+

		lea	-1(a5),a5	; samples -1 (before jumping out)

		move	d0,d3
		add	d6,d3
		blt.w	.fm16_mix_Ping_directionChange ; if position is at beginning of loop area, we are done here
		

		bra.b	.fm16_mix_PingB_underflow_loop
		; loop until end of sample

.fm16_mix_Ping_directionChange:
		; we were looping backwards, now go forwards
		MIX_PING_DIRECTIONCHANGE        ;change mixing direction offsets, issue interrupt (if necessary)
		bsr	.fm16_mix_preload
		bra.w	.fm16_mix_mainloop

.fm16_mix_Ping_buffer_full:
		MIX_PING_BUFFERFULL	.fm16_mix_Ping_return
.fm16_mix_Ping_return:
		moveq	#1,d0
		rts
;
; preload FM_Preload samples to AS_FMSamples buffer
; called at each start of an instrument (after a "silence")
; (not for repeating instruments, there the buffer is
;  already available)
;
; input:
;  - AS_Struct in A4
; output:
;  - D0 offset + 2
;
; notes:
;  - AS_FMPos may be at any state, sanity check is done here
.fm16_mix_preload:
		push	d1-d3
		move.b	AS_FMPos(A4),d3

		moveq	#FM_Preload-fadein-1,d1
		moveq	#0,d2
		move.l	d2,AS_FMSamples1(A4)			;clear sample buffer 
		move.l	d2,AS_FMSamples2(A4)			;clear sample buffer
.fm16_mix_preload_loop:
		and.w	#FM_Taps-1,d3				;make sure, the position is sane (assumes 2^n taps)
		move.w	(a0,d0.w*2),d2				;8 Bit input
		move.w	(A6,d2.l*2),AS_FMSamples1(a4,d3.w*2)	;Volume correction, store
		addq.w	#1,d0
		addq.w	#1,d3
		dbf	d1,.fm16_mix_preload_loop

		;D0 points to the next sample to be loaded -> compatible with other mixing routines
		;subq.w	#1,d0 ; point to last sample stored in buffer so +1 will load "next" in main loop (OBSOLETE)
		bset	#ASLoopF_FMInit,AS_CurrentFPOS+2(a4) ; Filtered Mixing initialized ?

		; fade instrument in by 1 sample (to reduce clicking)
		; longer fade-in times will require to re-calculate d0
		ifne	fadein
		sub.l	d4,d0		; go back 1 sample (so that the next sample will be loaded after the first byte)
		add.w	d4,d0		; just to make sure, in fact d4.w _should_ be zero as we check for downsampling cases
		endc

		move.b	d3,AS_FMPos(A4)
		pop	d1-d3
		rts

;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
;--------------------------- 16 Bit Filtered Down Mixing Routine (Part1 - Move Data) -----------------
;-----------------------------------------------------------------------------------------------------
; Input:
;        A4   - AudioStruct for current channel and panning
;        A2   - Output buffer for "SamLEN*2" shorts
;        D0   - number of samples (SamLEN*2)
;        A6   - Volume table
; Output:
;        (A2) - Output samples
;        D0   - 1=mixing done, 0=mixing skipped (no sample played)
;
; Used Registers: all, none saved
;        D0   - current offset towards A0 (upper word is current fractional position)
;        D4   - interpolation step size, lower word integer part, upper word fractional part
;        D5   - upper word of D4, 0 in lower word
;        A0   - current source sample, offset by (already) played bytes
;        A5   - remaining samples (except in fast loop, where it is -16)
;        A6   - Volume Table
; Notes:
;  - 4 Tap FIR interpolation filter (Windowed Sinc), tradeoff between multiplies per output sample 
;    and sound quality
;  - TODO: ping-pong backwards preload and FM mix, currently uses nearest neighbor
;  - IMPORTANT:   don't call this one directly (go to fm16_mix first), it always assumes the step 
;                 size to be >=1.0
;               - if you don't like it this way, adapt the mixing macro to check for d4.w >= 1
;
;-----------------------------------------------------------------------------------------------------
fm16down_mix:
	ifne	benchmark
	move	#$f00,$dff180
	endc

	; filtered mixing routine can`t handle downsampling, refer to regular "mix" in those cases
	; TODO: branch to nearest neighbor for ratios >= 2
;	move.w	AS_Period(a4),d6	;current period
;	beq.w	.fm16down_mix_ret_nop
;	cmp.w	mixingperiod4,d6	;must be larger than mixing period, else use regular routine
;	blt.w	mix16			;ok, period is wrong, use regular mixer	

	; init macro
	MIX_INIT	.fm16down_mix_ret_nop,0,1,1 ; set up registers (common routine to all mixers, no volume table, add offset twice)
						; third arg, if given and ==1, calculate which filtered mixing table to use,
						; based on the downsampling factor

	btst	#ASLoopF_FMInit,AS_CurrentFPOS+2(a4) 
	bne.s	.fm16down_mix_fminitialized	; Filtered Mixing initialized ?
	bsr.w	.fm16down_mix_preload		; preload samples into filter buffer
.fm16down_mix_fminitialized:

	moveq	#0,d2			; irrelevant ?!
	moveq	#0,d3			;

	btst	#ASLoopF_BackLooping,AS_CurrentFPOS+2(a4) ; D6 -> stop position for backward loop
	bne.w	.fm16down_mix_Ping_backwards

	MIX_REMBYTES	d6,d1,1		;total length - played length = remaining bytes, 2nd arg: scratch register, 3rd arg: 1=16 Bit mode, 0=8 Bit mode

;-----------------------------------------------------------------------------------------------
; main loop: check vital parameters and time-scale the inputs
;-----------------------------------------------------------------------------------------------
.fm16down_mix_mainloop:			;start point for "fast" loop and "repeated" samples
	MIX_LIMITER	d6,norm		;limit remaining samples in D6 to $7fff (16 bit signed max.)

	ifne	benchmark
	move	#$00f,$dff180
	endc
	

	moveq	#0,d3
	move.b	AS_FMPos(a4),d3			;ringbuffer index
	lsl.w	#FMB_Accuracy+FMB_Taps+1,d3	;accuracy bits + taps bits + word size
	move.l	FM_tab(pc),a3
	add.l	d3,a3				;table for current ringbuffer position

	lea	-16(a5),a5 ; simplifies check for "less than 16 samples" to generate

	move.l	d0,d7			; current position in instrument
	add.l	a1,d7			; + 16 output samples wrt. sampling rate
	cmp.w	d7,d6			; >= end of instrument
	blo.w	.fm16down_mix_underflow16	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)
	move.w	a5,d1			; 16 samples left to write in output buffer ?
	blt.w	.fm16down_mix_underflow16 	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

;-----------------------------------------------------------------------------------------------
; main loop
; d1,d2,d3,d5,d7 are free
;-----------------------------------------------------------------------------------------------
.fm16down_mix_fastloop:
	fm16down_mix_one	1,0
	fm16down_mix_one	2,0
	fm16down_mix_one	3,0
	fm16down_mix_one	4,0
	fm16down_mix_one	5,0
	fm16down_mix_one	6,0
	fm16down_mix_one	7,0
	fm16down_mix_one	8,0
	fm16down_mix_one	9,0
	fm16down_mix_one	10,0
	fm16down_mix_one	11,0
	fm16down_mix_one	12,0
	fm16down_mix_one	13,0
	fm16down_mix_one	14,0
	fm16down_mix_one	15,0
	fm16down_mix_one	16,0

		lea	-16(a5),a5	; no condition codes changed (=16 samples processed)
		move.l	a5,d1							          
		swap	d1		; if( remaining_output_bytes < 16 ) 0xffff else 0x0000
		or.w	d6,d1		; if( remaining_output_bytes < 16 ) -1 else remaining_input_bytes
		move.l	d0,d7		;processed bytes
		add.l	a1,d7		;last byte to be processed
		cmp.w	d7,d1		;if( D1 < D7 ) -> stop
		bgt.w	.fm16down_mix_fastloop
;-----------------------------------------------------------------------------------------------

;-----------------------------------------------------------------------------------------------
; less than 16 bytes to go or sample ending
;-----------------------------------------------------------------------------------------------
.fm16down_mix_underflow16:
		lea	16(a5),a5 ; restore A5 with remaining bytes to be filled (we started with -16 in fast loop)

		ifne	benchmark
		move	#$444,$dff180
		endc

.fm16down_mix_underflow_loop:
		move.w	a5,d1
		ble.w	.fm16down_mix_buffer_full ; done, buffer filled

	fm16down_mix_one	17,1
		
		cmp	d0,d6		;
		lea	-1(a5),a5	; samples -1 (before jumping out)
		bhi.w	.fm16down_mix_underflow_loop ; 
		; loop until end of instrument
;-----------------------------------------------------------------------------------------------

;-----------------------------------------------------------------------------------------------
; we land here with A5 having the number of remaining bytes
; decide whether to loop the current sample or to clear the rest of the buffer
;-----------------------------------------------------------------------------------------------

	; non-looping instruments require some extra treatment:
	;  - we have a look-ahead in here, so that 2 "future" samples are already stored
	;  - provide a "dummy" extension with "0"-samples
	;  - no problem with "repeating" instruments since the integer position is reset and
	;    new samples are loaded as needed
	MIX_FM_FADEOUT .fm16down_mix_clear_rest,.fm16down_mix_mainloop ;

	;-------------------------- "usual" loop code ------------------------------------------

	MIX_LOOPCHECK	.fm16down_mix_clear_rest,.fm16down_mix_PingPongLoop,1	;1st arg: bra target for one-shot samples, 2nd arg: bra target for ping-pong loop code

	tst.w	d0				; integer position != 0 -> one sample was missing 
	beq.s	.fm16down_mix_no_reload		; ok, fm16down_mix_one exited normally

	moveq	#0,d2
	move.w	(A0),d2				;get sample (first sample of repeat part)
 	move.w	(A6,d2.l*2),d2			;volume correct
	moveq	#0,d3
	move.b	AS_FMPos(a4),d3			;ringbuffer index
	and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
	move.w	d2,AS_FMSamples1(A4,d3.w*2)	;
	addq.w	#1,d3
	and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
	move.b	d3,AS_FMPos(A4)

.fm16down_mix_no_reload:

	bra.w	.fm16down_mix_mainloop

;----------------- ping-pong loop first entry --------------------------
.fm16down_mix_PingPongLoop
	MIX_PINGPONG_INIT	.fm16down_mix_Ping_backwards,1

;-------- DONE: one buffer is filled, now just store parameters and leave
.fm16down_mix_buffer_full:

	ifne	benchmark
	move	#$000,$dff180
	endc

	; check whether we reached the end of the instrument before leaving
	MIX_POSTCHECK	.fm16down_mix_bfull_ret,1

.fm16down_mix_bfull_ret:
	moveq	#1,d0
	rts

;--------------------- clear remaining bytes - TODO: filter here also ----------------
.fm16down_mix_clear_rest:
		move.l	A5,d1		;remaining bytes
		moveq	#$0,d0
		subq	#1,d1		;for dbf loop
		bmi.s	.fm16down_mix_ret_clear
.fm16down_mix_clear_loop:
		move	d0,(a2)+
		dbf	d1,.fm16down_mix_clear_loop
.fm16down_mix_ret_clear
		moveq	#1,d0
		rts

; no explicit clear when empty-buffer aware routines are in place
.fm16down_mix_ret_nop:
		moveq	#0,d0
		rts

;-------------- Ping-Pong Loop entry point for backwards loop -------------------
;- custom routine, repeating the normal interpolation but checking against negative offsets
.fm16down_mix_Ping_backwards	;yep, do loop backwards

		;neg.w	d4		;count integer positions backwards (subx instead of addx for offsets)
					;in D6, the remaining length is kept (current position = rest until "0" is reached)
		move.l	a1,d1		;invert end position check
		neg.w	d1
		move.l	d1,a1
		
		MIX_LIMITER	d6,ping	;limit D6 to 0x7fff

		lea	-16(a5),a5 ; simplifies check for "less than 16 samples"
		moveq	#0,d5
		moveq	#0,d2
.fm16down_mix_PingB_fastloop:
		move.w	a5,d1		 ; sets condition codes but not "X"
		lea	(A1,D0.l),A3	 ; see above, (needs addq #1,d1)
		blt.w	.fm16down_mix_PingB_underflow16 ; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

		;if position after 16 bytes (d0+next samples) <= start of loop, then leave here
		lea	(a3,d6.w),A3
		move.w	a3,d3
		blt.w	.fm16down_mix_PingB_underflow16	; IMPORTANT: DON`T JUMP ELSEWHERE, keep A5 in mind (-16)

		rept	8
			move.b	(a0,d0.w),d2
			sub.l	d4,d0
			subx.w	d5,d0
			move.w	(A6,d2.w*2),d3
			swap	d3
			move.b	(a0,d0.w),d2
			sub.l	d4,d0
			subx.w	d5,d0
			move.w	(A6,d2.w*2),d3
			move.l	d3,(a2)+
		endr
		lea	-16(a5),a5	; no condition codes changed
		bra.w	.fm16down_mix_PingB_fastloop

		; slow loop, A5 almost empty and/or near end of loop area
.fm16down_mix_PingB_underflow16:
		lea	16(a5),a5 ; restore A5 with remaining bytes to be filled (we started with -16 in fast loop)

		ifne	benchmark
			move	#$444,$dff180
		endc

		; NOTE: X bit still set until this point
		; less than 16 bytes to be written (bad choice of SamLEN) 
		; or (more likely) near the end of the sample
.fm16down_mix_PingB_underflow_loop:
		move.w	a5,d1
		ble.w	.fm16down_mix_Ping_buffer_full

		move.b	(a0,d0.w),d2	; one byte at a time, happens rarely (!!)
		sub.l	d4,d0
		subx.l	d5,d0
		move.w	(A6,d2.w*2),(A2)+

		lea	-1(a5),a5	; samples -1 (before jumping out)

		move	d0,d3
		add	d6,d3
		blt.w	.fm16down_mix_Ping_directionChange ; if position is at beginning of loop area, we are done here
		

		bra.b	.fm16down_mix_PingB_underflow_loop
		; loop until end of sample

.fm16down_mix_Ping_directionChange:
		; we were looping backwards, now go forwards
		MIX_PING_DIRECTIONCHANGE        ;change mixing direction offsets, issue interrupt (if necessary)
		bsr	.fm16down_mix_preload
		bra.w	.fm16down_mix_mainloop

.fm16down_mix_Ping_buffer_full:
		MIX_PING_BUFFERFULL	.fm16down_mix_Ping_return
.fm16down_mix_Ping_return:
		moveq	#1,d0
		rts
;
; preload FM_Preload samples to AS_FMSamples buffer
; called at each start of an instrument (after a "silence")
; (not for repeating instruments, there the buffer is
;  already available)
;
; input:
;  - AS_Struct in A4
; output:
;  - D0 offset + 2
;
; notes:
;  - AS_FMPos may be at any state, sanity check is done here
.fm16down_mix_preload:
		push	d1-d3
		move.b	AS_FMPos(A4),d3

		moveq	#FM_Preload-fadein-1,d1
		moveq	#0,d2
		move.l	d2,AS_FMSamples1(A4)			;clear sample buffer 
		move.l	d2,AS_FMSamples2(A4)			;clear sample buffer
.fm16down_mix_preload_loop:
		and.w	#FM_Taps-1,d3				;make sure, the position is sane (assumes 2^n taps)
		move.w	(a0,d0.w*2),d2				;8 Bit input
		move.w	(A6,d2.l*2),AS_FMSamples1(a4,d3.w*2)	;Volume correction, store
		addq.w	#1,d0
		addq.w	#1,d3
		dbf	d1,.fm16down_mix_preload_loop

		;D0 points to the next sample to be loaded -> compatible with other mixing routines
		;subq.w	#1,d0 ; point to last sample stored in buffer so +1 will load "next" in main loop (OBSOLETE)
		bset	#ASLoopF_FMInit,AS_CurrentFPOS+2(a4) ; Filtered Mixing initialized ?

		; fade instrument in by 1 sample (to reduce clicking)
		; longer fade-in times will require to re-calculate d0
		ifne	fadein
		sub.l	d4,d0		; go back 1 sample (so that the next sample will be loaded after the first byte)
		add.w	d4,d0		; just to make sure, in fact d4.w _should_ be zero as we check for downsampling cases
		endc

		move.b	d3,AS_FMPos(A4)
		pop	d1-d3
		rts


;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
;-- Apply high-boost filter to input buffer (in-place)                                              --
;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
; Input:
;       A0    - mixing buffer to be filtered
;       A1    - previous stage of filter, coefficient buffer
;       D7    - number of 16 bit samples in buffer (SamLen*2)
; Output:
;       (A0)+ - filtered data
; Notes: - no normalization is done here, prepare mixing buffer contents for a typical boost of factor 1.5-2
;        - coefficients for 44 kHz sampling rate
;
HighBoostFilter:
	movem.w	(A1),d0-d1	;d0 = 4*x[n-1], d1 = y[n-1]

	;filter term:
	; 
	; 1 - 0.5*z^-1
	; ------------
	; 1 - 0.5*z^-1

	subq	#1,d7
.doboost:
	move	(A0),d2		;x[n]
	asr.w	#1,d0		;x[n-1]/2=0.5*x[n-1]
	asr.w	#1,d1		;0.5*y[n-1]
	neg.w	d0		;-0.5*x[n-1]
	neg.w	d1
	add.w	d2,d0		;y[n]=x[n]-0.5*x[n-1]
	add.w	d0,d1		;y[n]=x[n]-0.5*x[n-1]-0.5*y[n-1]
	move	d2,d0		;remember x[n]
	move.w	d1,(a0)+
	dbf	d7,.doboost

	movem.w	d0-d1,(A1)	;remember last input, output
	rts

; same as above, just tuned down the high boost to match the 28 kHz rate setting
; this one does a little more high boost than needed, 14 dB at 28 kHz instead of ~12.5 dB
; ...chirp frequency response evaluation pending...
HighBoostFilter28:
	movem.w	(A1),d0-d1	;d0 = 4*x[n-1], d1 = y[n-1]

	;filter term:
	; 
	; 1 - 0.5*z^-1
	; ------------
	; 1 - 0.25*z^-1

	subq	#1,d7
.doboost:
	move	(A0),d2		;x[n]
	asr.w	#1,d0		;x[n-1]/2=0.5*x[n-1]
	asr.w	#3,d1		;0.25*y[n-1]
	neg.w	d0		;-0.5*x[n-1]
	neg.w	d1
	add.w	d2,d0		;y[n]=x[n]-0.5*x[n-1]
	add.w	d0,d1		;y[n]=x[n]-0.5*x[n-1]-0.5*y[n-1]
	move	d2,d0		;remember x[n]
	move.w	d1,(a0)+
	dbf	d7,.doboost

	movem.w	d0-d1,(A1)	;remember last input, output
	rts


;-----------------------------------------------------------------------------------------------------
;-----------------------------------------------------------------------------------------------------
;--------------------------------- Hardware Output Routines  -----------------------------------------
;-----------------------------------------------------------------------------------------------------

	ifne	PAULA14
		include	"misc/Ampout_Paula14.s"	;default: Paula 14 Bit
Ampout_DEFAULT	=	AmpOut_Paula14
Ampout_HAVEDEF	=	1
	endc

	ifne	TOCCATA
		include	misc/Toccata.i
		include	misc/Toccata_lib.i

		include "misc/Ampout_Toccata.s"	;toccata.library (Toccata,Melody,UAE)


		ifnd	Ampout_HAVEDEF
Ampout_DEFAULT	=	AmpOut_Toccata
Ampout_HAVEDEF	=	1
		endc
	endc


	ifne	AHIOUT
		include	devices/ahi.i
		include	lvo/ahi_lib.i
		include "utility/hooks.i"

		include "misc/Ampout_AHI.s"	

		ifnd	Ampout_HAVEDEF
Ampout_DEFAULT	=	AmpOut_AHI
Ampout_HAVEDEF	=	1
		endc

	endc

;*****************************************************************************
;*                        Daten und Initialisierungen                        *
;*****************************************************************************
dectable:
	dc.l	1,10,100,1000,10000,100000,1000000,10000000,100000000

int:	dc.b	`intuition.library`,0

	ifne	V37Gadtools
gad:	dc.b	`gadtools13.library`,0
       	else
gad:	dc.b	`gadtools.library`,0
	endc

dos:	dc.b	`dos.library`,0
graf	dc.b	`graphics.library`,0

NoportMessage:	dc.b	`Cannot find EaglePlayer Port`,0
	even
portname:	dc.b	`EAGLEPLAYERPORT`,0

my_easygadget:	dc.b	`   OK   `,0
my_easygadget2	dc.b	`Retry|Cancel`,0

	even
	Project0WindowTags:
Project0L:
	DC.L	WA_Left,0
Project0T:
	DC.L	WA_Top,0
Project0W:
	DC.L	WA_Width,0
Project0H:
	DC.L	WA_Height,0

	DC.L	WA_IDCMP,IDCMP_Changewindow!IDCMP_RawKey!IDCMP_Menupick!LISTVIEWIDCMP!BUTTONIDCMP!IDCMP_CLOSEWINDOW!IDCMP_REFRESHWINDOW
	DC.L	WA_Flags,WFLG_DRAGBAR!WFLG_DEPTHGADGET!WFLG_CLOSEGADGET!WFLG_SMART_REFRESH

	DC.L	WA_Title,Project0WTitle
	DC.L	WA_ScreenTitle,Project0STitle
	dc.l	wa_newlookmenus,1
	dc.l	wa_autoadjust,1

	dc.l	wa_pubscreen
Pubadr
	dc.l	0
	dc.l	WA_Zoom
	dc.l	ZoomStruct
	DC.L	TAG_DONE,0
ZoomStruct
	dc.w	0,0,150,0

Ahawidth
	dc.l	399
Ahaheight
	dc.l	80

IR:
	DC.L	GT_VisualInfo,$00000000,GTBB_recessed,0
	dc.l	TAG_DONE
MTags0:
	DC.L	GTMN_Newlookmenus,1
	dc.l	TAG_DONE

TopazName:
	DC.B	'topaz.font',0

;--------------------------- spezifische Texte und Strukturen --------------------
	ifne	PAULA14

savename:	dc.b	`14_Bit_Amplifier.cfg`,0

MyAnalyzername:		dc.b	`E_14Bit_AMP.1`,0
MyUsername:		dc.b	`14 Bit Amplifier`,0
myBuggsname:		dc.b	`Henryk Richter`,0

myinfo		dc.b	`This Amplifier supports`,10
		dc.b	`up to 32 voices in a special`,10
		dc.b	`Amiga 14 Bit mode.`,10,10
		dc.b	`It allows to change the`,10
		dc.b	`stereo bandwidth from mono`,10
		dc.b	`to surround for much better`,10
		dc.b	`sound in critical rooms and`,10
		dc.b	`especially when using head-`,10
		dc.b	`phones. Additionally you`,10
		dc.b	`can use the "pseudo stereo"`,10
		dc.b	`mode for mono samples and`,10
		dc.b	`-modules.`,10
		dc.b	`This program allows to `,10
		dc.b	`change all parameters`,10
		dc.b	`( mixing rate,  surround`,10
		dc.b	`settings, volume boost ...)`,10
		dc.b	`in realtime.`,10,10
		dc.b	`Version `
		ver
		dc.b	` `,10
		dc.b	`Written by Henryk Richter,`
		dc.b	`14 Bit-table creation and`,10
		dc.b	`calibration routines by`,10
		dc.b	`Christian Buchner`,10,0

my_easytitle:	dc.b	`14 Bit Amplifier Message`,0

Aboutmessage:	dc.b	`Eagleplayer 14 Bit 32 voices Amplifier `
		ver
		dc.b	10,10
		dc.b	`(C) 1995-2016 Henryk Richter`,10,10
		dc.b	`Notes:`,10
		dc.b	`- restricted to 68020+`,10
		dc.b	`- try Surround on fast Amigas, it rocks (!)`,10
		dc.b	`  compared to "headphone" modes in other players`,10
		dc.b	`- ECS/AGA in double scan mode (VGAOnly,DBLPal,Super72,...)`,10
		dc.b    `  required for Paula sampling rates >28 kHz`,10
		dc.b	`- Double-clicking on a row in the lister enables/disables`,10
		dc.b    `  a particular mode`,10 
		dc.b	'- Filtered mixing requires a potent machine, 68040+',10
		dc.b	0

Project0WTitle:
	dc.b	'14 Bit Amplifier',0
Project0STitle:
	DC.B	'14 Bit Surround Amplifier, (C) DEFECT Softworks',0

	else	; PAULA14
	 ifne	AHIOUT
savename:	dc.b	`AHI_Amplifier.cfg`,0

MyAnalyzername:		dc.b	`E_AHI_AMP.1`,0
MyUsername:		dc.b	`AHI Amplifier`,0
myBuggsname:		dc.b	`Henryk Richter`,0

myinfo          dc.b    `This Amplifier directs the `
                dc.b    `mixed data to AHI. It should `
                dc.b    `be used when direct support for `
		dc.b	`a specific hardware is not available. `,10
		dc.b	`Please note that AHI is an additional `,10
		dc.b	`layer between the audio and the actual output. `,10
		dc.b	`It will be therefore more demanding on the CPU. `,10
		dc.b	`Since this Amplifier shares it's code base with `,10
		dc.b	`the 14 Bit and Toccata Amplifiers, the same `,10
		dc.b	`feature set is available.`,10
                dc.b    `Version `
                ver
                dc.b    ` `,10
		dc.b	`Written by Henryk Richter,`,0

my_easytitle:	dc.b	`AHI Amplifier Message`,0

FailOpenMessage:
		dc.b	`Failed to open AHI.`,10
		dc.b	`Please check that you have the ahi.device,`,10
		dc.b	`appropriate AHI drivers and AHI audiomodes`,10
		dc.b	`installed. AHI 4.8 (on Aminet) comes with`,10
		dc.b	`an installer script that should take care`,10
		dc.b	`of the necessary steps.`,0

Aboutmessage:	dc.b	`Eagleplayer AHI Amplifier `
		ver
		dc.b	10,10
		dc.b	`(C) 1995-2016 Henryk Richter`,10,10
		dc.b	`Notes:`,10
		dc.b	`- restricted to 68020+`,10
		dc.b	`- configure an appropriate AHI mode before playing`,10
		dc.b	`- Double-clicking on a row in the lister enables/disables`,10
		dc.b    `  a particular mode`,10 
		dc.b	'- Filtered mixing requires a potent machine, 68040/40 or better',10
		dc.b	'- Highboost option in filtered mixing was tuned for the Paula ',10
		dc.b    '  lowpass and does not really benefit soundcards with proper reconstruction',10
		dc.b	'  filters.',10
		dc.b	0

Project0WTitle:
	dc.b	'AHI Amplifier',0
Project0STitle:
	DC.B	'AHI Amplifier, (C) DEFECT Softworks',0

unknownstring:	dc.b	'unnamed',0

	 else	; AHIOUT
savename:	dc.b	`Toccata_AmplifierII.cfg`,0

MyAnalyzername:		dc.b	`E_Toccata_AMP.1`,0
MyUsername:		dc.b	`Toccata Amplifier`,0
myBuggsname:		dc.b	`Henryk Richter`,0

myinfo          dc.b    `This Amplifier supports `
                dc.b    `up to 32 voices using the `
                dc.b    `toccata.library (Toccata, Melody and Prelude Soundboards). `
		dc.b	`It allows to change the`,10
		dc.b	`stereo bandwidth from mono`,10
		dc.b	`to surround for much better`,10
		dc.b	`sound in critical rooms and`,10
		dc.b	`especially when using head-`,10
		dc.b	`phones. Additionally you`,10
		dc.b	`can use the "pseudo stereo"`,10
		dc.b	`mode for mono samples and`,10
		dc.b	`-modules.`,10
                dc.b    `This program allows to `
                dc.b    `change the parameters for `
                dc.b    `surround settings, volume boost and surround delay `
                dc.b    `in realtime.`,10
                dc.b    `Version `
                ver
                dc.b    ` `,10
		dc.b	`Written by Henryk Richter,`,0

my_easytitle:	dc.b	`Toccata Amplifier Message`,0

Aboutmessage:	dc.b	`Eagleplayer Toccata Amplifier `
		ver
		dc.b	10,10
		dc.b	`(C) 1995-2016 Henryk Richter`,10,10
		dc.b	`Notes:`,10
		dc.b	`- restricted to 68020+`,10
		dc.b	`- try Surround on fast Amigas, it rocks (!)`,10
		dc.b	`  compared to "headphone" modes in other players`,10
		dc.b	`- Double-clicking on a row in the lister enables/disables`,10
		dc.b    `  a particular mode`,10 
		dc.b	'- Filtered mixing requires a potent machine, 68040+',10
		dc.b	'- Highboost option in filtered mixing was tuned for the Paula ',10
		dc.b    '  lowpass and does not benefit soundcards with proper reconstruction',10
		dc.b	'  filters.',10
		dc.b	0

Project0WTitle:
	dc.b	'Toccata Amplifier',0
Project0STitle:
	DC.B	'Toccata Surround Amplifier, (C) DEFECT Softworks',0


	 endc	; AHIOUT
	endc	; PAULA14


	ifne	BUFCHECK
Internalmessage:
	dc.b	`Internal Check Failure`,10
	dc.b	`Buffer Index: `
Internalidx:
	dc.b	`0    `,10
	dc.b	`Pos: `
Internalpos:
	dc.b	`0    `,10
	dc.b	0
	endc
	
	even

;update Project0_CNT after making changes to the gadget list: update Types, Gads and Tags

; embossed bevel boxes
Project0Bevel:
	dc.w	2
;	dc.w	4,116,300,38	;left,top,width,height
;	dc.w	4,58,300,57
	dc.w	4,58,300+XWidth,57+18
	dc.w	4,116+18,300+XWidth,38	;left,top,width,height


; gadget types
Project0GTypes:
    DC.W    LISTVIEW_KIND
    DC.W    INTEGER_KIND
    DC.W    SLIDER_KIND
    DC.W    SLIDER_KIND
    DC.W    TEXT_KIND
    DC.W    TEXT_KIND
    DC.W    SLIDER_KIND
    DC.W    SLIDER_KIND
    DC.W    CYCLE_KIND
    DC.W    BUTTON_KIND
    DC.W    CHECKBOX_KIND

Project0NGads:
NG_Listview
    DC.W    4,14,301+XWidth,48
    DC.L    LISTVIEWText,0
    DC.W    GD_LISTVIEW
    DC.L    PLACETEXT_ABOVE,0,0

    DC.W    229+XWidth,97,69,14
    DC.L    SamplefreqText,0
    DC.W    GD_Samplefreq
    DC.L    0,0,0

    DC.W    12,99,213,10
    DC.L    MixingSliderText,0
    DC.W    GD_MixingSlider
    DC.L    PLACETEXT_ABOVE,0
 	ifne	AHIOUT
    DC.L    GD_Disable
    	else
    DC.L    0
    	endc

    DC.W    66,72,182+XWidth,12
    DC.L    VBoostText,0
    DC.W    GD_VBoost
    DC.L    PLACETEXT_ABOVE,0,0

    DC.W    26,71,26,14
    DC.L    VoicesText,0
    DC.W    GD_Voices
    DC.L    PLACETEXT_ABOVE,0,0

	DC.W    257+XWidth,71,36,13
	DC.L    NullText,0
	DC.W    GD_NumVBoost
	DC.L    PLACETEXT_ABOVE,0,0

	DC.W    14,130+18,283+XWidth,9
	DC.L    Gadget60Text,0
	DC.W    GD_Playmode
	DC.L    PLACETEXT_ABOVE,0,0


	DC.W    152+(XWidth/2),141+18,145+(XWidth/2),9
	DC.L    SurlevelText,0
	DC.W    GD_Surlevel
	DC.L    PLACETEXT_LEFT,0,0

	;/* cycle for mixing mode */
	DC.W    72-8,114,226+8,14
	DC.L    MixModeText,0
	DC.W    GD_Mixmode
	DC.L    PLACETEXT_LEFT,0,0

	;/* Button in AHI Mode */
	DC.W    72-8,96,150,16
	dc.l	AHIModeText,0
	dc.w	GD_AHIMode
	dc.l	PLACETEXT_IN,0
 	ifne	AHIOUT
	    DC.L    0
       	else
	    DC.L    GD_Disable
    	endc

	;/* Button in AHI Mode */
	DC.W    22,100,16,16
	dc.l	AHIGlobalText,0
	dc.w	GD_AHIGlobal
	dc.l	PLACETEXT_ABOVE,0
 	ifne	AHIOUT
	    DC.L    0
       	else
	    DC.L    GD_Disable
    	endc


myhook					;Filterhook, um zu gro?e Fonts nicht
					;anzuzeigen
	dcb.b	MLN_SIZE,0
	dc.l	my_hook
	dc.l	0
	dc.l	0

Project0GTags:
    DC.L    GTLV_ShowSelected,0
	dc.l	GTLV_CallBack,myhook
    DC.L    TAG_DONE

    DC.L    GTIN_Number,32768
    DC.L    GTIN_MaxChars,5
    DC.L    STRINGA_Justification,GACT_STRINGCENTER
    DC.L    TAG_DONE
    DC.L    GTSL_Min,400
    DC.L    GTSL_Max,5800
    DC.L    GTSL_Level,100
    DC.L    PGA_Freedom,LORIENT_HORIZ
    DC.L    GA_Immediate,1
    DC.L    GA_RelVerify,1
    DC.L    TAG_DONE
    DC.L    GTSL_Min,10
    DC.L    GTSL_Max,100
    DC.L    GTSL_Level,100
    DC.L    PGA_Freedom,LORIENT_HORIZ
    DC.L    GA_Immediate,1
    DC.L    GA_RelVerify,1
    DC.L    TAG_DONE
    DC.L    GTTX_Text,VoicesString
    DC.L    GTTX_Border,1
    DC.L    TAG_DONE
    DC.L    GTTX_Text,NumVBoostString
    DC.L    TAG_DONE
    DC.L    GTSL_Min,-64
    DC.L    GTSL_Max,64
    DC.L    PGA_Freedom,LORIENT_HORIZ
    DC.L    GA_Immediate,1
    DC.L    GA_RelVerify,1
    DC.L    GT_Underscore,'_'
    DC.L    TAG_DONE
    DC.L    PGA_Freedom,LORIENT_HORIZ
    DC.L    GA_Immediate,1
    DC.L    GA_RelVerify,1
    DC.L    GTSL_Min,0
    DC.L    GTSL_Max,SamLEN/10
    DC.L    TAG_DONE
    DC.L    GTCY_Labels,MixModeCycleTexts
    DC.L    GTCY_Active,0
    DC.L    TAG_DONE
    ; todo: AHI Mode button attributes
    DC.L    TAG_DONE
    ; todo: AHI Global button attributes
    DC.L    TAG_DONE


NumVBoostString:
    DC.B    '10.2',0

VoicesString:
    DC.B    '16',0

LISTVIEWText:
    DC.B    'Prefs for each number of Voices',0

SamplefreqText:
    DC.B    ' ',0

MixingSliderText:
    DC.B    'Sample mixing frequency',0

VBoostText:
    DC.B    'Volume Boost'
NullText:
    dc.b	` `,0

VoicesText:
    DC.B    'Voices',0

Gadget60Text:
    DC.B    'Mono <------- _Stereo ---> Surround',0

SurlevelText:
    DC.B    'Surround delay',0

MixModeText:
    DC.B    'Mixer',0
MixMode0Text:
    DC.B    'Amiga - Nearest Neighbor',0
MixMode1Text:
    DC.B    'Polyphase Sinc',0
MixMode2Text:
    DC.B    'Highboost Polyphase Sinc',0
AHIModeText:
    DC.B    'Set AHI Mode',0
AHIGlobalText:
    DC.B    'Global',0

	even
MixModeCycleTexts:
	dc.l	MixMode0Text
	dc.l	MixMode1Text
	dc.l	MixMode2Text
	dc.l	0



Project0NewMenu0:
    DC.B    NM_TITLE,0
    DC.L    Project0MName0
    DC.L    0
    DC.W    0
    DC.L    0,0

	DC.B	NM_ITEM,0
	DC.L	Project0MName91
	DC.L	Project0MComm9
Hiercheck_Mono
	DC.W	CHECKIT!Menutoggle
	DC.L	0,0

	DC.B    NM_ITEM,0
	DC.L    NM_BARLABEL,0
	DC.W    0
	DC.L    0,0

    DC.B    NM_ITEM,0
    DC.L    Project0MName6
    DC.L    Project0MComm6
    DC.W    0
    DC.L    0,0

    DC.B    NM_ITEM,0
    DC.L    Project0MName7
    DC.L    Project0MComm7
    DC.W    0
    DC.L    0,0

    DC.B    NM_ITEM,0
    DC.L    Project0MName8
    DC.L    Project0MComm8
    DC.W    0
    DC.L    0,0

    DC.B    NM_TITLE,0
    DC.L    Project0MName61
    DC.L    0
    DC.W    0
    DC.L    0,0

	DC.B	NM_ITEM,0
	DC.L	Project0MName1
	DC.L	Project0MComm1
Hiercheck
	DC.W	CHECKIT!Menutoggle
	DC.L	0,0

	DC.B    NM_ITEM,0
	DC.L    NM_BARLABEL,0
	DC.W    0
	DC.L    0,0

    DC.B    NM_ITEM,0
    DC.L    Project0MName71
    DC.L    0
    DC.W    0
    DC.L    0,0

    DC.B    NM_ITEM,0
    DC.L    Project0MName81
    DC.L    0
    DC.W    0
    DC.L    0,0

    DC.B    NM_END,0
    DC.L    0,0
    DC.W    0
    DC.L    0,0

Project0MName0:		DC.B    'Project',0
Project0MName1		dc.b	'Global',0
Project0MComm1		dc.b	'G',0
Project0MName6:		DC.B    'About',0
Project0MComm6:		DC.B    'A',0
Project0MName7:		DC.B    'Hide',0
Project0MComm7:		DC.B    'H',0
Project0MName8:		DC.B    'Quit',0
Project0MComm8:		DC.B    'Q',0
Project0MComm9:		DC.B    'M',0

Project0MName61:
    DC.B    'Settings',0

Project0MName71:
    DC.B    'Reset to Defaults',0

Project0MName81:
    DC.B    'Save Settings',0

Project0MName91:
    DC.B    'Mono Modules Mode',0

	even
TD:
    DC.L    TAG_DONE

;--------------------------------------------------------------------------------------------

	*------- Activate ------*
EP_Menu:	dc.l	EPM_Show		;Next
		dc.w	0			;Left Edge
		dc.w	0			;Verti Pos.
		dc.w	0			;Width-Size (Negationbreite)
		dc.w	10			;Hight-Size
		dc.w	$5B
		dc.l	0
		dc.l	.ActivateIText
		dc.l	0
		dc.b	0
		ds.b	7
		dc.l	0			*Kick2.04
.ActivateIText:	dc.b	0,3
		ds.w	2
		dc.w	1
		dc.l	0			*Topaz80
		dc.l	0			*Active.MSG (filled by EP)
		dc.l	0
		dc.w	EPNr_MainWindow	*EPNr_Engine
		dc.w	LNr_Active

	*------- Show-Window ------*
EPM_Show	dc.l	0 EPM_BarLabel		;Next
		dc.w	0			;Left Edge
		dc.w	0			;Verti Pos.
		dc.w	0			;Width-Size (Negationbreite)
		dc.w	10			;Hight-Size
		dc.w	$5B
		dc.l	0
		dc.l	.ShowIText
		dc.l	0
		dc.b	0
		ds.b	7
		dc.l	0			*Kick2.04
.ShowIText:	dc.b	0,3
		ds.w	2
		dc.w	1
		dc.l	0			*Topaz80
		dc.l	0			*Active.MSG (filled by EP)
		dc.l	0
		dc.w	EPNr_EngineShowState
		dc.w	LNr_Show2

;----------------- hier nun die Daten fuer den Mixing Player ----------------

AdditiveArray	dcb.b	255,$55		;calibration array, initialized with defaults (why not $3f?)
		dc.b	$7f


LastBuffer	dc.l	0

AudioRate:	dc.l	3546895		;fuer PAL
clock:		dc.l	0		;PAL reference clock (will be unused at the end of the cleanup process)
mrate:		dc.l	0
mixingperiod:	dc.l	0		;Mixing period
referenceperiod: dc.l	0		;Mixing Period * 262144 for mixing step size calculations
mixingperiod4:	dc.w	0,0		;Mixing Period * 4 (for filtered mixing comparisons)

myAudioStruct:	dc.l	0
numStructs:	dc.w	0
numleft:	dc.w	0
numright:	dc.w	0
divtabs:	ds.l	16
LastBoost:	dc.w	0
maxchan:	dc.w	0
modenr:		dc.w	0
;curchannels	dc.w	0

	;diese 6 hier zusammen lassen
	ifne	PAULA14
buffer0:	dc.l	0
buffer1:	dc.l	0
buffer2:	dc.l	0
buffer3:	dc.l	0
	endc
cbufad:		dc.l	0	;copybufStereoT
cbufempty:	dc.l	0	;copybufemptyStereo14
	;---------------------------

SpecialBuffer1:	dc.l	0
SpecialBuffer2:	dc.l	0
mixad1:		dc.l	mix
mixad2:		dc.l	mix2
boostad:	dc.l	0	;high boost address

Vtabaddr:	dc.l	0
VtabaddrR:	dc.l	0
Vtabsize:	dc.l	0
usedmode:	dc.l	0
	ifeq	nosurroundtab

SurroundTabAddr: dc.l	0
SurroundTabSize: dc.l	0
SurroundTabAddrMid: dc.l 0

	endc

buff14          dc.l    0
buff14_2        dc.l    0
Vtab:		dc.l	0
effectvol:	dc.w	0
onplay:		dc.b	0
waitstart:	dc.b	0
prefetchL:	dc.w	0
prefetchR:	dc.w	0
unusedW:	dc.w	0	;currently unused
newmixingperiod: dc.l	0
tmpvol:		dc.l	0
tmpvol2:	dc.w	0
calcvol:	ds.w	4
fformat:	dc.b	0
UNUSEDXXX	dc.b	0
playmixrate:	dc.l	0
UPSBuffer0:	dc.l	0
UPSBuffer1:	dc.l	0
UPSBuffer2:	dc.l	0
UPSBuffer3:	dc.l	0

	ifne	BUFCHECK
Checkbuffers:	dc.l	BCHECK1,32
		dc.l	BCHECK2,32
		dc.l	BCHECK3,32
		dc.l	0,0
	endc

FM_tab:	;active filtered mixing table (dependent on up/downsampling ratio)
	dc.l	FM_tab_100

	; currently 4 scaling ratios
FM_tabs_Bits	equ	2

	; Filtered Mixing Tables
	include	"misc/Ampout_FilterTables4.i"	;Filtered Mixing tables, 4 Taps

	; Debug stuff
	ifne	mixdebug
	dcb.b	64,$80

testsample:
;	incdir	modules:Instruments/
	incdir	Temp:Instruments/
	incbin	drop2
;	incdir	work:src/
;	incbin	tp5v23pingpong.raw

;	dc.w	100
;	dc.w	120
;	dc.w	140
;	dc.w	160
;	dcb.w	64,200
	dcb.b	512,0

testsamplesize = *-testsample
;	dc.w	0

	dcb.w	64,$ff9f
	endc

;divtab2
;a set 16
;	rept	16
;		ds.b	256*a
;a set a-1
;	endr
;Vtab		ds.b	66*256			; Volume tab size for 14 Bit
;						; for 8 Bit only ? of this needed

;*********** Leertabellen (sp?ter im BSS-Hunk) ***********************
		section	1,BSS_p
	ifne	AHIOUT
AHIAMP_Tagbuffer:	ds.l	30
AHIAMP_Tagbuffer2:	ds.l	30
	endc
		rsreset
datas:
RealNumGadgets	rs.w	1
MC68020on	rs.w	1
MC68040on	rs.w	1
		rs.w	1	;unused
MyVariable	rs.l	1
Attr		rs.l	2
msgport		rs.l	1
intsave:	rs.l	1
grafsave	rs.l	1
gadsave:	rs.l	1
_dos		rs.l	1
Port:		rs.l	1
Jumpadr		rs.l	1
Menuadr:	rs.l	1
mypubscreen:	rs.l	1	;aktueller Publicscreen
fonth:		rs.l	1	;Fonthoehe
;fontatt:	rs.l	1	;Fontattr Strukturadresse (entweder die des
				;Screens oder die eigene,wenn Screenfont zu
				;gro?)
vi:		rs.l	1	;Visualinfo

mytags:		rs.l	10*2
mycontextgad:	rs.b	gg_SIZEOF	;CreateContext() braucht Zeiger auf
					;leere Gadgetstruktur !?

mygad:		rs.b	gng_SIZEOF	;Platz fuer New Gadget Struktur


Project0Gadgets:	rs.l	21
Gadgetadrs=Project0Gadgets

BufNewGad:		rs.w	4
			rs.l	2
			rs.w	1
			rs.l	3

myhoehe2:	rs.w	1
myhoehe1:	equ	myhoehe2
winwidth:	rs.w	1

Project0GList:	rs.l	1
Project0Menus:	rs.l	1
Winhandle	rs.l	1
Font:		rs.l	1
FontX:		rs.w	1
FontY:		rs.w	1
OffX:		rs.w	1
OffY:		rs.w	1
		rs.w	2

MY_Easystruct:	rs.l	1	;EASY-Request
		rs.l	1
		rs.l	1
		rs.l	1
		rs.l	1
temp1		rs.l	128
temp2		rs.l	128
temp:		rs.l	1

refreshflag	rs.b	1
counter		rs.b	1		;Frei
topazused	rs.b	1
modeid		rs.b	1

;an_structadr	rs.l	1
Mytask		rs.l	1
MySignal	rs.l	1
My2ndSignal	rs.l	1
immernull	rs.l	1
myobject	rs.l	1
merkpos		rs.l	1
currpos		rs.l	1
visible		rs.l	1
myplayertaglist	rs.l	1
maxtextwidth	rs.w	1
maxvisible	rs.w	1
SaveString	rs.l	1
myfontatt:	rs.l	1
		rs.w	2

MsgPort1:	rs.b	MP_SIZE	;extra Messageport

micros		rs.l	1
seconds		rs.l	1
merk_seconds	rs.l	1
lasttime	rs.l	1
messagesent	rs.b	1
messagesent2	rs.b	1
savemodnr	rs.l	1
counter2	rs.w	1
onsysfont	rs.b	1
exg		rs.b	1
onglobal	rs.b	1
		rs.b	1
MyRastport	rs.b	rp_SIZEOF
fontadr		rs.l	1
onmixingslider	rs.b	1
onvboostslider	rs.b	1
onPlaymodeslider	rs.b	1
onSurlevslider		rs.b	1

NumVoc		rs.l	1	;"12"
numboo		rs.l	1	;"10.1 "
		rs.w	1

					;
SelectedL:	rs.w	1		;die zwei gehoeren zusammen
SelectedW:	rs.w	1		;-------------------------
					;
Currentnumvoc	rs.w	1

List:		rs.b	LH_SIZE
Nodes:		rs.b	ML_Sizeof*MaxVoices	;max. 32 voices
	ifne	AHIOUT
AHINameList:	rs.b	LH_SIZE		;copied list of mode IDs and their names
AHINameListInit: rs.w	1		;list init flag
	endc

ConfigDatas	rs.b	0

PrefsVersion	rs.l	1		;"EP".w + Versionsnummer, wg. Veränderungen = Check

Prefslist	rs.w	my_prefs_Sizeof ;mixingrate.l
					;Vboost.w
					;Effect.w  StereoBasisbreite.b
					;          Surround Level.b
					;
					;Flags.w (Bit0=disable)

ConfigDatasEND	rs.b	0

	ifne	debug
saveusernr	rs.w	1
	endc
DatasLen:	rs.w	0
		ds.b	DatasLen

	ifne	PAULA14

	section	2,bss_c

;-------------- Puffer fuer die DMA simulierung -----------------
Buffer0a:	ds.w	SamLEN*NumChipBuffers
		ds.w	2

Buffer1a:	ds.w	SamLEN*NumChipBuffers
		ds.w	2

Buffer2a:	ds.w	SamLEN*NumChipBuffers
		ds.w	2

Buffer3a:	ds.w	SamLEN*NumChipBuffers
		ds.w	2
EndBuffer

	endc	;PAULA14


	section	3,bss_p
	ifne mixdebug
testtab:	ds.w	$10000 ; table for boost test and other tests that require a temporary buffer
	endc
	
Amp_OHWStr:		ds.l	1
Amp_OHWStatus:		ds.l	1
Amp_OHWMixrates:	ds.l	1
Amp_OHWMixrateCount:	ds.l	1

	; Toccata Data Buffers
	ifne	TOCCATA
TRawBuffer1:     ds.b    TRawSize         ;diese beiden Zeilen nicht auseinanderreissen !
TRawBuffer2:     ds.b    TRawSize         ;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
EndTRawBuffer:
	endc

MyUPSStruct	ds.b	UPS_SizeOF

;--------------------------------------- Start Clear on Start DMA ---------------------------------------
StartBSSBuffer	; all between here and EndBSSBuffer will be cleared on "start DMA"

BCHECK3		ds.l	16	; integrity check field
TempBuffer:	ds.b	SamLEN*2*2+4		;no zeroing necessary
BCHECK1		ds.l	16

TempBuffer2:	ds.b	SamLEN*2*2+4		;no zeroing necessary
BCHECK2		ds.l	16

	ifne	AHIOUT
AHIAMP_Buffer0		ds.w	SamLEN*2*2	;SamLEN*2 sample, 16 bit signed (ds.w), 2 channels (*2)
AHIAMP_Buffer1		ds.w	SamLEN*2*2
	endc

ZeroSamplebuffer: ds.b	SamLEN*2*2+2		;used in case that no mixing took place, always zero-filled

TBufLeft	ds.w	SamLEN*2		;Delayed left channel, 16 bit for surround buffering
TBufRight	ds.w	SamLEN*2		; Delayed data buffers for surround

	ifne	PAULA14
SurroundTemp:	ds.b	SamLEN*2*2*2+2		;used in revised surround routine for combination of main and delayed channel before ChipRAM copy
						;(Samlen*2)*(2 channels max.)*(2 bytes per channel), currently only in 14 Bit Paula mode
	endc

		; Data buffers for Highboost filter
HB_DelayL	ds.w	2			;just two coefficients needed in current implementation
HB_DelayR	ds.w	2


UPSBuffer0a:	ds.w	SamLEN*NumChipBuffers	;selbe Laengen und Offsets wie Buffer0a...3a
		ds.w	2
UPSBuffer1a:	ds.w	SamLEN*NumChipBuffers
		ds.w	2
UPSBuffer2a:	ds.w	SamLEN*NumChipBuffers
		ds.w	2
UPSBuffer3a:	ds.w	SamLEN*NumChipBuffers
Security	ds.b	20


EndBSSBuffer
;--------------------------------------- End Clear on Start DMA ---------------------------------------

SilenceBuffer:	ds.w	MIX_FM_FADEOUT_TIME	;see MixMacros.i

Extra_AS	ds.b	AS_Sizeof*MaxVoices

;TMP_AS		ds.b	AS_Sizeof
;TMP_SMPL	ds.b	200000


	end
